;;-*- Mode: Lisp; Package: CCL -*-
;;
;; print-class-tree.lisp
;; Copyright 1992-1994, Apple Computer, Inc.
;; Copyright 1995 Digitool, Inc.


;;;;;;;;;;;;;
;;
;; Modification History
;;
;; 10/12/93 alice - ignore non-existent classes e.g. temp-class
;; 01/02/92 bill New file.
;;

;;;;;;;;;;;;;
;;
;; Code to print class trees.
;;
;; Output for all the classes in MCL 2.0 is included as a comment
;; at the end of this file.
;;
;; All symbols are printed in the COMMON-LISP-USER package.
;; Hence, external symbols in the COMMON-LISP and CCL packages
;; print with no package prefix.
;;
;; Output is in two forms:
;;
;; 1) No multiple inheritance.
;;
;;    The class is printed on the first line and the rest of it's class
;;    precedence list is printed on the remaining lines, indented two spaces.
;;
;;    Example:
;;
;;      WINDOW
;;        VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
;;        STANDARD-OBJECT T
;;
;; 2) Multiple inheritance
;;
;;    A tree of superclasses is printed, one class per line with vertical
;;    bars ("|") to help see the structure.
;;    Any class which will appear again later in the hierarchy is
;;    printed with an asterisk ("*") in front of it.
;;    If there are classes left after the entire tree is printed,
;;    which happens with some cases of complicated multiple inheritance,
;;    the remainder of the class precedence list is printed on the
;;    last line with two asterisks ("**") in front of it.
;;    You can read the class precedence list from this tree by reading
;;    top to bottom and left to right and omitting classes that are
;;    preceded by a single asterisk.
;;
;;    Example:
;;
;;      SIMPLE-STRING
;;      | STRING
;;      | | * VECTOR
;;      | CCL::SIMPLE-1D-ARRAY
;;      | | VECTOR
;;      | | | * ARRAY
;;      | | | * SEQUENCE
;;      | | SIMPLE-ARRAY
;;      | | | ARRAY
;;      | | | | * T
;;        ** SEQUENCE T
;;
;;    This diagram can be interpreted as follows:
;;
;;    The direct superclasses of SIMPLE-STRING are STRING & CCL::SIMPLE-1D-ARRAY
;;    The direct superclass of STRING is VECTOR
;;    The direct superclasses of CCL::SIMPLE-1D-ARRAY are VECTOR & SIMPLE-ARRAY
;;    The direct superclasses of VECTOR are ARRAY & SEQUENCE
;;    The direct superclass of SIMPLE-ARRAY is ARRAY
;;    The direct superclass of ARRAY is T
;;    The remainder of the class precedence list is (SEQUENCE T)
;;
;;    The class precedence list of the class SIMPLE-STRING is:
;;    (SIMPLE-STRING STRING CCL::SIMPLE-1D-ARRAY VECTOR SIMPLE-ARRAY ARRAY
;;     SEQUENCE T)

(in-package :ccl)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (export '(print-class-tree print-all-class-trees)))

(defun print-class-tree (class &key (stream t) indent cpl-cell
                               (right-margin 80)
                               (consider-simple t))
  (when (symbolp class)
    (setq class (find-class class)))
  (when (null indent)
    (setq indent 0))
  (when (null cpl-cell)
    (setq cpl-cell (list (class-precedence-list class))))
  (let ((*package* *package*)
        (do-supers? (eq class (caar cpl-cell))))
    (in-package :cl-user)
    (flet ((print-indentation (stream indent)
             (dotimes (i indent) (format stream "| "))))
      (fresh-line stream)
      (print-indentation stream indent)
      (unless do-supers? (format stream "* "))
      (format stream "~s~%" (class-name class))
      (force-output stream)
      (when do-supers?
        (pop (car cpl-cell))
        (let ((supers (class-direct-superclasses class)))
          (labels ((simple-supers (supers)
                     (and (null (cdr supers))
                          (dolist (sup supers t)
                            (unless (simple-supers
                                     (class-direct-superclasses sup))
                              (return nil))))))
            (if (and consider-simple (simple-supers supers))
              ; At top level and all inheritance is single.
              (print-wrapped-list
               (mapcar 'class-name (car cpl-cell))
               stream
               (+ indent indent 2)
               right-margin)
              ; Not at top level or some inheritance was multiple
              (progn
                (dolist (super supers)
                  (print-class-tree
                   super 
                   :stream stream
                   :indent (1+ indent)
                   :cpl-cell cpl-cell
                   :consider-simple nil
                   :right-margin right-margin))
                (when (and consider-simple (car cpl-cell))
                  ; Had some CPL left at top level.
                  (let ((left-margin (+ indent indent 2)))
                    (format stream "~&~vt** " left-margin)
                    (print-wrapped-list
                     (mapcar 'class-name (car cpl-cell))
                     stream
                     (+ left-margin 3)
                     right-margin
                     nil)))))))))))

(defun print-wrapped-list (list &optional
                                (stream t) (left-margin 0) (right-margin 80)
                                (initial-indent t))
  (loop
    (when initial-indent
      (format stream "~&~vt" left-margin))
    (setq initial-indent t)
    (let ((col left-margin)
          (first t))
      (loop
        (let ((string (format nil "~s" (car list))))
          (if (or (< (incf col (length string)) right-margin) first)
            (progn
              (unless first
                (format stream " ")
                (incf col 1))
              (write-string string stream)
              (setq first nil)
              (when (null (setq list (cdr list)))
                (return-from print-wrapped-list nil)))
            (return)))))))
      

(defun print-all-class-trees (&optional (stream t) (right-margin 80))
  (let (classes)
    (maphash #'(lambda (name class) class (when (find-class name nil)(push name classes)))
             %find-classes%)
    (dolist (class (sort classes 'string-lessp))
      (print-class-tree class :stream stream :right-margin right-margin))))

#|
(print-all-class-trees (front-window))

CCL::ACTION-SEQUENCE-DIALOG
  DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::ACTION-SEQUENCE-ITEM
  SEQUENCE-DIALOG-ITEM TABLE-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
APPLE-MENU
  MENU CCL::MENU-ELEMENT CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
APPLEEVENT-ERROR
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
APPLICATION
  STANDARD-OBJECT T
APROPOS-DIALOG
| CCL::ARROW-DIALOG
| | * WINDOW
| CCL::STRING-DIALOG
| | DIALOG
| | | WINDOW
| | | | VIEW
| | | | | SIMPLE-VIEW
| | | | | | OUTPUT-STREAM
| | | | | | | STREAM
| | | | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | | | STANDARD-OBJECT
| | | | | | | | | | T
ARITHMETIC-ERROR
| SIMPLE-ERROR
| | SIMPLE-CONDITION
| | | * CONDITION
| | ERROR
| | | SERIOUS-CONDITION
| | | | CONDITION
| | | | | STANDARD-OBJECT
| | | | | | T
ARRAY
  T
INSPECTOR::ARRAY-INSPECTOR
| INSPECTOR::UVECTOR-INSPECTOR
| | INSPECTOR::BASICS-FIRST-MIXIN
| | | * STANDARD-OBJECT
| | INSPECTOR::INSPECTOR
| | | STANDARD-OBJECT
| | | | T
CCL::ARROW-DIALOG
  WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
INSPECTOR::BACKTRACE-INFO-PANE
| INSPECTOR::BOTTOM-LINE-MIXIN
| | * STANDARD-OBJECT
| VIEW
| | SIMPLE-VIEW
| | | OUTPUT-STREAM
| | | | STREAM
| | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | STANDARD-OBJECT
| | | | | | | T
INSPECTOR::BACKTRACE-INSPECTOR-PANE
| INSPECTOR::INSPECTOR-PANE
| | INSPECTOR::BOTTOM-LINE-MIXIN
| | | * STANDARD-OBJECT
| | VIEW
| | | SIMPLE-VIEW
| | | | OUTPUT-STREAM
| | | | | STREAM
| | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | STANDARD-OBJECT
| | | | | | | | T
INSPECTOR::BACKTRACE-VIEW
  VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
INSPECTOR::BACKTRACE-WINDOW
| INSPECTOR::UNDO-VIEW-MIXIN
| | * STANDARD-OBJECT
| INSPECTOR::BACKTRACE-VIEW
| | * VIEW
| WINDOW
| | VIEW
| | | SIMPLE-VIEW
| | | | OUTPUT-STREAM
| | | | | STREAM
| | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | STANDARD-OBJECT
| | | | | | | | T
BASE-CHARACTER
  CHARACTER T
BASE-STRING
| STRING
| | VECTOR
| | | ARRAY
| | | | * T
| | | SEQUENCE
| | | | T
CCL::BASIC-EDITABLE-TEXT-DIALOG-ITEM
| KEY-HANDLER-MIXIN
| | * STANDARD-OBJECT
| DIALOG-ITEM
| | SIMPLE-VIEW
| | | OUTPUT-STREAM
| | | | STREAM
| | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | STANDARD-OBJECT
| | | | | | | T
INSPECTOR::BASIC-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
INSPECTOR::BASICS-FIRST-MIXIN
  STANDARD-OBJECT T
BIGNUM
  INTEGER RATIONAL REAL NUMBER T
CCL::BINARY-FILE-STREAM
| CCL::BINARY-STREAM
| | * STREAM
| FILE-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
CCL::BINARY-STREAM
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
BIT
  FIXNUM INTEGER RATIONAL REAL NUMBER T
BIT-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
INSPECTOR::BOTTOM-LINE-MIXIN
  STANDARD-OBJECT T
BROADCAST-STREAM
  OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::BUFFER
  CCL::INTERNAL-STRUCTURE T
BUFFER-MARK
  T
CCL::BUFFER-STREAM
| INPUT-STREAM
| | * STREAM
| OUTPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
BUILT-IN-CLASS
  CLASS SPECIALIZER METAOBJECT STANDARD-OBJECT T
BUTTON-DIALOG-ITEM
  CCL::CONTROL-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::BYTE-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
INSPECTOR::CACHE-ENTRY
  STANDARD-OBJECT T
INSPECTOR::CACHE-ENTRY-STREAM
| TRUNCATING-STRING-STREAM
| | STRING-OUTPUT-STREAM
| | | STRING-STREAM
| | | | * STREAM
| | | OUTPUT-STREAM
| | | | STREAM
| | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | STANDARD-OBJECT
| | | | | | | T
CELL-ERROR
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
CHARACTER
  T
CHECK-BOX-DIALOG-ITEM
  CCL::CONTROL-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CLASS
  SPECIALIZER METAOBJECT STANDARD-OBJECT T
CCL::CLOSURE-DEF
  STRUCTURE-OBJECT T
INSPECTOR::CLOSURE-INSPECTOR
  INSPECTOR::FUNCTION-INSPECTOR INSPECTOR::INSPECTOR STANDARD-OBJECT T
COLOR-DIALOG
  DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::COMBINED-METHOD
  FUNCTION T
INSPECTOR::COMMAND-PANE
| INSPECTOR::BOTTOM-LINE-MIXIN
| | * STANDARD-OBJECT
| VIEW
| | SIMPLE-VIEW
| | | OUTPUT-STREAM
| | | | STREAM
| | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | STANDARD-OBJECT
| | | | | | | T
CCL::COMPILE-TIME-CLASS
  CLASS SPECIALIZER METAOBJECT STANDARD-OBJECT T
CCL::COMPILE-TIME-ERROR
| SIMPLE-ERROR
| | SIMPLE-CONDITION
| | | * CONDITION
| | ERROR
| | | SERIOUS-CONDITION
| | | | CONDITION
| | | | | STANDARD-OBJECT
| | | | | | T
CCL::COMPILE-TIME-PROGRAM-ERROR
| CCL::SIMPLE-PROGRAM-ERROR
| | SIMPLE-CONDITION
| | | * CONDITION
| | PROGRAM-ERROR
| | | ERROR
| | | | SERIOUS-CONDITION
| | | | | CONDITION
| | | | | | STANDARD-OBJECT
| | | | | | | T
FUNCTION
  T
COMPILED-LEXICAL-CLOSURE
  FUNCTION T
CCL::COMPILER-FUNCTION-OVERFLOW
  CONDITION STANDARD-OBJECT T
CCL::COMPILER-POLICY
  CCL::INTERNAL-STRUCTURE T
CCL::COMPILER-WARNING
  WARNING CONDITION STANDARD-OBJECT T
COMPLEX
  NUMBER T
COMTAB
  CCL::INTERNAL-STRUCTURE T
CONCATENATED-STREAM
  INPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CONDITION
  STANDARD-OBJECT T
CONS
  LIST SEQUENCE T
INSPECTOR::CONS-INSPECTOR
| INSPECTOR::BASICS-FIRST-MIXIN
| | * STANDARD-OBJECT
| INSPECTOR::INSPECTOR
| | STANDARD-OBJECT
| | | T
CCL::CONTROL-DIALOG-ITEM
  DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
CONTROL-ERROR
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
CCL::DA-WINDOW
  WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
DEFAULT-BUTTON-DIALOG-ITEM
  BUTTON-DIALOG-ITEM CCL::CONTROL-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW
  OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
DIALOG
  WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
DIALOG-ITEM
  SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
DIVISION-BY-ZERO
| ARITHMETIC-ERROR
| | SIMPLE-ERROR
| | | SIMPLE-CONDITION
| | | | * CONDITION
| | | ERROR
| | | | SERIOUS-CONDITION
| | | | | CONDITION
| | | | | | STANDARD-OBJECT
| | | | | | | T
CCL::DOC-OUTPUT-CLASS
| FRED-WINDOW
| | FRED-MIXIN
| | | * SIMPLE-VIEW
| | WINDOW
| | | VIEW
| | | | SIMPLE-VIEW
| | | | | OUTPUT-STREAM
| | | | | | STREAM
| | | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | | STANDARD-OBJECT
| | | | | | | | | T
DOUBLE-FLOAT
  FLOAT REAL NUMBER T
ECHO-STREAM
| TWO-WAY-STREAM
| | INPUT-STREAM
| | | * STREAM
| | OUTPUT-STREAM
| | | STREAM
| | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | STANDARD-OBJECT
| | | | | | T
EDITABLE-TEXT-DIALOG-ITEM
| FRED-DIALOG-ITEM
| | FRED-MIXIN
| | | * SIMPLE-VIEW
| | CCL::BASIC-EDITABLE-TEXT-DIALOG-ITEM
| | | KEY-HANDLER-MIXIN
| | | | * STANDARD-OBJECT
| | | DIALOG-ITEM
| | | | SIMPLE-VIEW
| | | | | OUTPUT-STREAM
| | | | | | STREAM
| | | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | | STANDARD-OBJECT
| | | | | | | | | T
CCL::ENCAPSULATION
  STRUCTURE-OBJECT T
END-OF-FILE
  STREAM-ERROR ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
CCL::ENVIRONMENT-DIALOG
  DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
ERROR
  SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
INSPECTOR::ERROR-FRAME
  STANDARD-OBJECT T
CCL::EVAL-CANT-COPE
| SIMPLE-CONDITION
| | * CONDITION
| ERROR
| | SERIOUS-CONDITION
| | | CONDITION
| | | | STANDARD-OBJECT
| | | | | T
CCL::EVAL-PROGRAM-ERROR
| CCL::SIMPLE-PROGRAM-ERROR
| | SIMPLE-CONDITION
| | | * CONDITION
| | PROGRAM-ERROR
| | | ERROR
| | | | SERIOUS-CONDITION
| | | | | CONDITION
| | | | | | STANDARD-OBJECT
| | | | | | | T
CCL::EXPORT-CONFLICT-ERROR
  PACKAGE-ERROR ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
EXTENDED-CHARACTER
  CHARACTER T
CCL::FIELD-DESCRIPTOR
  STRUCTURE-OBJECT T
FILE-ERROR
| SIMPLE-CONDITION
| | * CONDITION
| ERROR
| | SERIOUS-CONDITION
| | | CONDITION
| | | | STANDARD-OBJECT
| | | | | T
FILE-STREAM
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
FIXNUM
  INTEGER RATIONAL REAL NUMBER T
FLOAT
  REAL NUMBER T
CCL::FLOAT-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
FLOATING-POINT-OVERFLOW
| ARITHMETIC-ERROR
| | SIMPLE-ERROR
| | | SIMPLE-CONDITION
| | | | * CONDITION
| | | ERROR
| | | | SERIOUS-CONDITION
| | | | | CONDITION
| | | | | | STANDARD-OBJECT
| | | | | | | T
FLOATING-POINT-UNDERFLOW
| ARITHMETIC-ERROR
| | SIMPLE-ERROR
| | | SIMPLE-CONDITION
| | | | * CONDITION
| | | ERROR
| | | | SERIOUS-CONDITION
| | | | | CONDITION
| | | | | | STANDARD-OBJECT
| | | | | | | T
INSPECTOR::FONT-SIZE-MANAGER
  SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
INSPECTOR::FORMATTING-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
FRED-DIALOG-ITEM
| FRED-MIXIN
| | * SIMPLE-VIEW
| CCL::BASIC-EDITABLE-TEXT-DIALOG-ITEM
| | KEY-HANDLER-MIXIN
| | | * STANDARD-OBJECT
| | DIALOG-ITEM
| | | SIMPLE-VIEW
| | | | OUTPUT-STREAM
| | | | | STREAM
| | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | STANDARD-OBJECT
| | | | | | | | T
FRED-MIXIN
  SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
CCL::FRED-RECORD
  CCL::INTERNAL-STRUCTURE T
CCL::FRED-SCRAP-HANDLER
  SCRAP-HANDLER STANDARD-OBJECT T
FRED-WINDOW
| FRED-MIXIN
| | * SIMPLE-VIEW
| WINDOW
| | VIEW
| | | SIMPLE-VIEW
| | | | OUTPUT-STREAM
| | | | | STREAM
| | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | STANDARD-OBJECT
| | | | | | | | T
FUNCALLABLE-STANDARD-CLASS
  CCL::STD-CLASS CLASS SPECIALIZER METAOBJECT STANDARD-OBJECT T
CCL::FUNCALLABLE-STANDARD-OBJECT
| STANDARD-OBJECT
| | * T
| FUNCTION
| | T
FUNCTION
  T
INSPECTOR::FUNCTION-ARGS-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
INSPECTOR::FUNCTION-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
CCL::GENERAL-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
GENERIC-FUNCTION
| METAOBJECT
| | * STANDARD-OBJECT
| CCL::FUNCALLABLE-STANDARD-OBJECT
| | STANDARD-OBJECT
| | | * T
| | FUNCTION
| | | T
INSPECTOR::GF-INSPECTOR
  INSPECTOR::FUNCTION-INSPECTOR INSPECTOR::INSPECTOR STANDARD-OBJECT T
CCL::GVECTOR
  T
HASH-TABLE
  CCL::INTERNAL-STRUCTURE T
CCL::IMMEDIATE
  T
CCL::IMPORT-CONFLICT-ERROR
  PACKAGE-ERROR ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
CCL::INDENTING-STRING-OUTPUT-STREAM
| STRING-OUTPUT-STREAM
| | STRING-STREAM
| | | * STREAM
| | OUTPUT-STREAM
| | | STREAM
| | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | STANDARD-OBJECT
| | | | | | T
CCL::INPUT-BINARY-FILE-STREAM
| CCL::INPUT-BINARY-STREAM
| | CCL::BINARY-STREAM
| | | * STREAM
| | * INPUT-STREAM
| CCL::INPUT-FILE-STREAM
| | FILE-STREAM
| | | * STREAM
| | INPUT-STREAM
| | | STREAM
| | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | STANDARD-OBJECT
| | | | | | T
CCL::INPUT-BINARY-STREAM
| CCL::BINARY-STREAM
| | * STREAM
| INPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
CCL::INPUT-FILE-STREAM
| FILE-STREAM
| | * STREAM
| INPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
INPUT-STREAM
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
INSPECTOR::INSPECTOR
  STANDARD-OBJECT T
INSPECTOR::INSPECTOR-EDITOR
  WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
INSPECTOR::INSPECTOR-HISTORY-WINDOW
| INSPECTOR::INSPECTOR-WINDOW
| | INSPECTOR::UNDO-VIEW-MIXIN
| | | * STANDARD-OBJECT
| | WINDOW
| | | VIEW
| | | | SIMPLE-VIEW
| | | | | OUTPUT-STREAM
| | | | | | STREAM
| | | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | | STANDARD-OBJECT
| | | | | | | | | T
INSPECTOR::INSPECTOR-PANE
| INSPECTOR::BOTTOM-LINE-MIXIN
| | * STANDARD-OBJECT
| VIEW
| | SIMPLE-VIEW
| | | OUTPUT-STREAM
| | | | STREAM
| | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | STANDARD-OBJECT
| | | | | | | T
INSPECTOR::INSPECTOR-SCROLL-BAR
  SCROLL-BAR-DIALOG-ITEM CCL::CONTROL-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW
  OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
INSPECTOR::INSPECTOR-VIEW
  INSPECTOR::FONT-SIZE-MANAGER SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
INSPECTOR::INSPECTOR-WINDOW
| INSPECTOR::UNDO-VIEW-MIXIN
| | * STANDARD-OBJECT
| WINDOW
| | VIEW
| | | SIMPLE-VIEW
| | | | OUTPUT-STREAM
| | | | | STREAM
| | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | STANDARD-OBJECT
| | | | | | | | T
CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
INTEGER
  RATIONAL REAL NUMBER T
CCL::INTERNAL-STRUCTURE
  T
CCL::INTERPRETED-FUNCTION
  FUNCTION T
CCL::INTERPRETED-LEXICAL-CLOSURE
  FUNCTION T
CCL::INVALID-ARGUMENTS
  STYLE-WARNING CCL::COMPILER-WARNING WARNING CONDITION STANDARD-OBJECT T
CCL::INVALID-ARGUMENTS-GLOBAL
  STYLE-WARNING CCL::COMPILER-WARNING WARNING CONDITION STANDARD-OBJECT T
CCL::IO-BINARY-FILE-STREAM
| CCL::INPUT-BINARY-FILE-STREAM
| | CCL::INPUT-BINARY-STREAM
| | | * CCL::BINARY-STREAM
| | | * INPUT-STREAM
| | CCL::INPUT-FILE-STREAM
| | | * FILE-STREAM
| | | * INPUT-STREAM
| CCL::OUTPUT-BINARY-FILE-STREAM
| | CCL::OUTPUT-BINARY-STREAM
| | | CCL::BINARY-STREAM
| | | | * STREAM
| | | * OUTPUT-STREAM
| | CCL::OUTPUT-FILE-STREAM
| | | FILE-STREAM
| | | | * STREAM
| | | OUTPUT-STREAM
| | | | * STREAM
  ** INPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::IO-BINARY-STREAM
| CCL::INPUT-BINARY-STREAM
| | * CCL::BINARY-STREAM
| | * INPUT-STREAM
| CCL::OUTPUT-BINARY-STREAM
| | CCL::BINARY-STREAM
| | | * STREAM
| | OUTPUT-STREAM
| | | * STREAM
  ** INPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::IO-FILE-STREAM
| CCL::INPUT-FILE-STREAM
| | * FILE-STREAM
| | * INPUT-STREAM
| CCL::OUTPUT-FILE-STREAM
| | FILE-STREAM
| | | * STREAM
| | OUTPUT-STREAM
| | | * STREAM
  ** INPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::IVECTOR
  T
KEY-HANDLER-MIXIN
  STANDARD-OBJECT T
CCL::KEYSTROKE-ACTION-DIALOG
  DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
KEYWORD
  SYMBOL T
CCL::LEXICAL-ENVIRONMENT
  CCL::INTERNAL-STRUCTURE T
CCL::LFUN-VECTOR
  T
CCL::LISP-SCRAP-HANDLER
  SCRAP-HANDLER STANDARD-OBJECT T
CCL::LISP-WDEF-MIXIN
  STANDARD-OBJECT T
LIST
  SEQUENCE T
INSPECTOR::LIST-INSPECTOR
| INSPECTOR::BASICS-FIRST-MIXIN
| | * STANDARD-OBJECT
| INSPECTOR::INSPECTOR
| | STANDARD-OBJECT
| | | T
LISTENER
| FRED-WINDOW
| | FRED-MIXIN
| | | * SIMPLE-VIEW
| | WINDOW
| | | VIEW
| | | | SIMPLE-VIEW
| | | | | OUTPUT-STREAM
| | | | | | * STREAM
| INPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
LOGICAL-PATHNAME
  PATHNAME CCL::INTERNAL-STRUCTURE T
DOUBLE-FLOAT
  FLOAT REAL NUMBER T
LONG-METHOD-COMBINATION
  METHOD-COMBINATION METAOBJECT STANDARD-OBJECT T
CCL::LONG-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
MACPTR
  T
CCL::MACTYPE
  STRUCTURE-OBJECT T
MENU
  CCL::MENU-ELEMENT CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::MENU-ELEMENT
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
MENU-ITEM
  CCL::MENU-ELEMENT CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::MENU-OF-DEFS-DIALOG
  CCL::SELECT-DIALOG CCL::ARROW-DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
MENUBAR
  STANDARD-OBJECT T
METAOBJECT
  STANDARD-OBJECT T
METHOD
  METAOBJECT STANDARD-OBJECT T
METHOD-COMBINATION
  METAOBJECT STANDARD-OBJECT T
METHOD-FUNCTION
  FUNCTION T
INSPECTOR::METHOD-INSPECTOR
| INSPECTOR::STANDARD-OBJECT-INSPECTOR
| | INSPECTOR::OBJECT-FIRST-INSPECTOR
| | | INSPECTOR::OBJECT-FIRST-MIXIN
| | | | * STANDARD-OBJECT
| | | * INSPECTOR::INSPECTOR
| INSPECTOR::FUNCTION-INSPECTOR
| | INSPECTOR::INSPECTOR
| | | STANDARD-OBJECT
| | | | T
MINI-BUFFER
  OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::MODIFY-READ-ONLY-BUFFER
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
CCL::NO-SUCH-PACKAGE
  PACKAGE-ERROR ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
NULL
| SYMBOL
| | * T
| LIST
| | SEQUENCE
| | | T
NUMBER
  T
INSPECTOR::OBJECT-FIRST-INSPECTOR
| INSPECTOR::OBJECT-FIRST-MIXIN
| | * STANDARD-OBJECT
| INSPECTOR::INSPECTOR
| | STANDARD-OBJECT
| | | T
INSPECTOR::OBJECT-FIRST-MIXIN
  STANDARD-OBJECT T
CCL::OUTPUT-BINARY-FILE-STREAM
| CCL::OUTPUT-BINARY-STREAM
| | CCL::BINARY-STREAM
| | | * STREAM
| | * OUTPUT-STREAM
| CCL::OUTPUT-FILE-STREAM
| | FILE-STREAM
| | | * STREAM
| | OUTPUT-STREAM
| | | STREAM
| | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | STANDARD-OBJECT
| | | | | | T
CCL::OUTPUT-BINARY-STREAM
| CCL::BINARY-STREAM
| | * STREAM
| OUTPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
CCL::OUTPUT-FILE-STREAM
| FILE-STREAM
| | * STREAM
| OUTPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
OUTPUT-STREAM
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
PACKAGE
  T
PACKAGE-ERROR
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
INSPECTOR::PACKAGE-INSPECTOR
| INSPECTOR::UVECTOR-INSPECTOR
| | INSPECTOR::BASICS-FIRST-MIXIN
| | | * STANDARD-OBJECT
| | INSPECTOR::INSPECTOR
| | | STANDARD-OBJECT
| | | | T
CCL::PACKAGE-NAME-CONFLICT-ERROR
| PACKAGE-ERROR
| | * ERROR
| SIMPLE-ERROR
| | SIMPLE-CONDITION
| | | * CONDITION
| | ERROR
| | | SERIOUS-CONDITION
| | | | CONDITION
| | | | | STANDARD-OBJECT
| | | | | | T
PANE-SPLITTER
  SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
PATHNAME
  CCL::INTERNAL-STRUCTURE T
INSPECTOR::PLIST-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
CCL::POOL
  T
POP-UP-MENU
| MENU
| | CCL::MENU-ELEMENT
| | | * CCL::INSTANCE-INITIALIZE-MIXIN
| DIALOG-ITEM
| | SIMPLE-VIEW
| | | OUTPUT-STREAM
| | | | STREAM
| | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | STANDARD-OBJECT
| | | | | | | T
CCL::POP-UP-TERMINAL-IO
| CCL::TERMINAL-IO
| | INPUT-STREAM
| | | * STREAM
| | OUTPUT-STREAM
| | | STREAM
| | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | STANDARD-OBJECT
| | | | | | T
CCL::POPULATION
  T
PPRINT-DISPATCH
  STRUCTURE-OBJECT T
PRINT-NOT-READABLE
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
CCL::PRINT-OPTIONS-DIALOG
  CCL::STRING-DIALOG DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
PROGRAM-ERROR
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
RADIO-BUTTON-DIALOG-ITEM
  CCL::CONTROL-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
RANDOM-STATE
  CCL::INTERNAL-STRUCTURE T
RATIO
  RATIONAL REAL NUMBER T
RATIONAL
  REAL NUMBER T
READTABLE
  CCL::INTERNAL-STRUCTURE T
REAL
  NUMBER T
CCL::RECORD-DESCRIPTOR
  STRUCTURE-OBJECT T
INSPECTOR::RECORD-INSPECTOR
| INSPECTOR::OBJECT-FIRST-INSPECTOR
| | INSPECTOR::OBJECT-FIRST-MIXIN
| | | * STANDARD-OBJECT
| | INSPECTOR::INSPECTOR
| | | STANDARD-OBJECT
| | | | T
RESTART
  CCL::INTERNAL-STRUCTURE T
SCRAP-HANDLER
  STANDARD-OBJECT T
SCROLL-BAR-DIALOG-ITEM
  CCL::CONTROL-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::SEARCH-DIALOG
  DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::SEARCH-FILE-DIALOG
  CCL::STRING-DIALOG DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::SELECT-DIALOG
  CCL::ARROW-DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::SELECTION-STREAM
| CCL::BUFFER-STREAM
| | INPUT-STREAM
| | | * STREAM
| | OUTPUT-STREAM
| | | STREAM
| | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | STANDARD-OBJECT
| | | | | | T
SEQUENCE
  T
SEQUENCE-DIALOG-ITEM
  TABLE-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
INSPECTOR::SEQUENCE-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
SERIOUS-CONDITION
  CONDITION STANDARD-OBJECT T
SHORT-FLOAT
  FLOAT REAL NUMBER T
SHORT-METHOD-COMBINATION
  METHOD-COMBINATION METAOBJECT STANDARD-OBJECT T
SIGNED-BYTE
  INTEGER RATIONAL REAL NUMBER T
CCL::SIMPLE-1D-ARRAY
| VECTOR
| | * ARRAY
| | * SEQUENCE
| SIMPLE-ARRAY
| | ARRAY
| | | * T
  ** SEQUENCE T
SIMPLE-ARRAY
  ARRAY T
SIMPLE-BASE-STRING
| BASE-STRING
| | * STRING
| SIMPLE-STRING
| | STRING
| | | * VECTOR
| | CCL::SIMPLE-1D-ARRAY
| | | VECTOR
| | | | * ARRAY
| | | | * SEQUENCE
| | | SIMPLE-ARRAY
| | | | ARRAY
| | | | | * T
  ** SEQUENCE T
SIMPLE-BIT-VECTOR
| BIT-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
CCL::SIMPLE-BYTE-VECTOR
| CCL::BYTE-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
SIMPLE-CONDITION
  CONDITION STANDARD-OBJECT T
CCL::SIMPLE-DESTRUCTURING-ERROR
| CCL::SIMPLE-PROGRAM-ERROR
| | SIMPLE-CONDITION
| | | * CONDITION
| | PROGRAM-ERROR
| | | ERROR
| | | | SERIOUS-CONDITION
| | | | | CONDITION
| | | | | | STANDARD-OBJECT
| | | | | | | T
SIMPLE-ERROR
| SIMPLE-CONDITION
| | * CONDITION
| ERROR
| | SERIOUS-CONDITION
| | | CONDITION
| | | | STANDARD-OBJECT
| | | | | T
CCL::SIMPLE-FLOAT-VECTOR
| CCL::FLOAT-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
CCL::SIMPLE-LONG-VECTOR
| CCL::LONG-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
CCL::SIMPLE-PROGRAM-ERROR
| SIMPLE-CONDITION
| | * CONDITION
| PROGRAM-ERROR
| | ERROR
| | | SERIOUS-CONDITION
| | | | CONDITION
| | | | | STANDARD-OBJECT
| | | | | | T
CCL::SIMPLE-READER-ERROR
| STREAM-ERROR
| | * ERROR
| SIMPLE-ERROR
| | SIMPLE-CONDITION
| | | * CONDITION
| | ERROR
| | | SERIOUS-CONDITION
| | | | CONDITION
| | | | | STANDARD-OBJECT
| | | | | | T
CCL::SIMPLE-STORAGE-CONDITION
| SIMPLE-CONDITION
| | * CONDITION
| STORAGE-CONDITION
| | SERIOUS-CONDITION
| | | CONDITION
| | | | STANDARD-OBJECT
| | | | | T
SIMPLE-STRING
| STRING
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
SIMPLE-TYPE-ERROR
| SIMPLE-CONDITION
| | * CONDITION
| TYPE-ERROR
| | ERROR
| | | SERIOUS-CONDITION
| | | | CONDITION
| | | | | STANDARD-OBJECT
| | | | | | T
CCL::SIMPLE-UNSIGNED-BYTE-VECTOR
| CCL::UNSIGNED-BYTE-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
CCL::SIMPLE-UNSIGNED-LONG-VECTOR
| CCL::UNSIGNED-LONG-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
CCL::SIMPLE-UNSIGNED-WORD-VECTOR
| CCL::UNSIGNED-WORD-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
SIMPLE-VECTOR
| CCL::GENERAL-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
SIMPLE-VIEW
  OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
SIMPLE-WARNING
| SIMPLE-CONDITION
| | * CONDITION
| WARNING
| | CONDITION
| | | STANDARD-OBJECT
| | | | T
CCL::SIMPLE-WORD-VECTOR
| CCL::WORD-VECTOR
| | * VECTOR
| CCL::SIMPLE-1D-ARRAY
| | VECTOR
| | | * ARRAY
| | | * SEQUENCE
| | SIMPLE-ARRAY
| | | ARRAY
| | | | * T
  ** SEQUENCE T
DOUBLE-FLOAT
  FLOAT REAL NUMBER T
SPECIALIZER
  METAOBJECT STANDARD-OBJECT T
INSPECTOR::STACK-FRAME-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
INSPECTOR::STACK-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
INSPECTOR::STACK-INSPECTOR-VIEW
  INSPECTOR::INSPECTOR-VIEW INSPECTOR::FONT-SIZE-MANAGER SIMPLE-VIEW
  OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::STACK-OVERFLOW-CONDITION
| CCL::SIMPLE-STORAGE-CONDITION
| | SIMPLE-CONDITION
| | | * CONDITION
| | STORAGE-CONDITION
| | | SERIOUS-CONDITION
| | | | CONDITION
| | | | | STANDARD-OBJECT
| | | | | | T
STANDARD-ACCESSOR-METHOD
  STANDARD-METHOD METHOD METAOBJECT STANDARD-OBJECT T
STANDARD-CHAR
  BASE-CHARACTER CHARACTER T
STANDARD-CLASS
  CCL::STD-CLASS CLASS SPECIALIZER METAOBJECT STANDARD-OBJECT T
STANDARD-GENERIC-FUNCTION
| GENERIC-FUNCTION
| | METAOBJECT
| | | * STANDARD-OBJECT
| | CCL::FUNCALLABLE-STANDARD-OBJECT
| | | STANDARD-OBJECT
| | | | * T
| | | FUNCTION
| | | | T
STANDARD-METHOD
  METHOD METAOBJECT STANDARD-OBJECT T
CCL::STANDARD-METHOD-COMBINATION
  METHOD-COMBINATION METAOBJECT STANDARD-OBJECT T
STANDARD-OBJECT
  T
INSPECTOR::STANDARD-OBJECT-INSPECTOR
| INSPECTOR::OBJECT-FIRST-INSPECTOR
| | INSPECTOR::OBJECT-FIRST-MIXIN
| | | * STANDARD-OBJECT
| | INSPECTOR::INSPECTOR
| | | STANDARD-OBJECT
| | | | T
STANDARD-READER-METHOD
  STANDARD-ACCESSOR-METHOD STANDARD-METHOD METHOD METAOBJECT STANDARD-OBJECT T
STANDARD-WRITER-METHOD
  STANDARD-ACCESSOR-METHOD STANDARD-METHOD METHOD METAOBJECT STANDARD-OBJECT T
STATIC-TEXT-DIALOG-ITEM
  DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
CCL::STD-CLASS
  CLASS SPECIALIZER METAOBJECT STANDARD-OBJECT T
INSPECTOR::STEP-COMMAND-PANE
| INSPECTOR::COMMAND-PANE
| | INSPECTOR::BOTTOM-LINE-MIXIN
| | | * STANDARD-OBJECT
| | VIEW
| | | SIMPLE-VIEW
| | | | OUTPUT-STREAM
| | | | | STREAM
| | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | STANDARD-OBJECT
| | | | | | | | T
INSPECTOR::STEP-PANE
  VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
INSPECTOR::STEP-WINDOW
  WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
STORAGE-CONDITION
  SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
STREAM-ERROR
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
STRING
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
CCL::STRING-DIALOG
  DIALOG WINDOW VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM
  CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::STRING-INPUT-STREAM
| STRING-STREAM
| | * STREAM
| INPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
INSPECTOR::STRING-INSPECTOR
| INSPECTOR::ARRAY-INSPECTOR
| | INSPECTOR::UVECTOR-INSPECTOR
| | | INSPECTOR::BASICS-FIRST-MIXIN
| | | | * STANDARD-OBJECT
| | | INSPECTOR::INSPECTOR
| | | | STANDARD-OBJECT
| | | | | T
STRING-OUTPUT-STREAM
| STRING-STREAM
| | * STREAM
| OUTPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
STRING-STREAM
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
STRUCTURE-CLASS
  CLASS SPECIALIZER METAOBJECT STANDARD-OBJECT T
STRUCTURE-OBJECT
  T
STYLE-WARNING
  CCL::COMPILER-WARNING WARNING CONDITION STANDARD-OBJECT T
SYMBOL
  T
CCL::SYMBOL-LOCATIVE
  T
SYNONYM-STREAM
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
T
  NIL
TABLE-DIALOG-ITEM
  DIALOG-ITEM SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
CCL::TERMINAL-IO
| INPUT-STREAM
| | * STREAM
| OUTPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
CCL::TERMINAL-IO-RUBOUT-HANDLER
  STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::TEXT-SCRAP-HANDLER
  SCRAP-HANDLER STANDARD-OBJECT T
TRUNCATING-STRING-STREAM
| STRING-OUTPUT-STREAM
| | STRING-STREAM
| | | * STREAM
| | OUTPUT-STREAM
| | | STREAM
| | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | STANDARD-OBJECT
| | | | | | T
TWO-WAY-STREAM
| INPUT-STREAM
| | * STREAM
| OUTPUT-STREAM
| | STREAM
| | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | STANDARD-OBJECT
| | | | | T
TYPE-ERROR
  ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
INSPECTOR::UNBOUND-MARKER
  STANDARD-OBJECT T
UNBOUND-VARIABLE
  CELL-ERROR ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
UNDEFINED-FUNCTION
  CELL-ERROR ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
CCL::UNDEFINED-FUNCTION-CALL
| CONTROL-ERROR
| | * ERROR
| UNDEFINED-FUNCTION
| | CELL-ERROR
| | | ERROR
| | | | SERIOUS-CONDITION
| | | | | CONDITION
| | | | | | STANDARD-OBJECT
| | | | | | | T
CCL::UNDEFINED-FUNCTION-REFERENCE
  STYLE-WARNING CCL::COMPILER-WARNING WARNING CONDITION STANDARD-OBJECT T
INSPECTOR::UNDO-VIEW-MIXIN
  STANDARD-OBJECT T
CCL::UNINTERN-CONFLICT-ERROR
  PACKAGE-ERROR ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
UNSIGNED-BYTE
  INTEGER RATIONAL REAL NUMBER T
CCL::UNSIGNED-BYTE-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
CCL::UNSIGNED-LONG-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
CCL::UNSIGNED-WORD-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
CCL::USE-PACKAGE-CONFLICT-ERROR
  PACKAGE-ERROR ERROR SERIOUS-CONDITION CONDITION STANDARD-OBJECT T
INSPECTOR::USUAL-BASICS-FIRST-INSPECTOR
| INSPECTOR::BASICS-FIRST-MIXIN
| | * STANDARD-OBJECT
| INSPECTOR::USUAL-INSPECTOR
| | INSPECTOR::INSPECTOR
| | | STANDARD-OBJECT
| | | | T
INSPECTOR::USUAL-FORMATTING-INSPECTOR
| INSPECTOR::USUAL-INSPECTOR
| | * INSPECTOR::INSPECTOR
| INSPECTOR::FORMATTING-INSPECTOR
| | INSPECTOR::INSPECTOR
| | | STANDARD-OBJECT
| | | | T
INSPECTOR::USUAL-INSPECTOR
  INSPECTOR::INSPECTOR STANDARD-OBJECT T
INSPECTOR::USUAL-OBJECT-FIRST-INSPECTOR
| INSPECTOR::OBJECT-FIRST-MIXIN
| | * STANDARD-OBJECT
| INSPECTOR::USUAL-INSPECTOR
| | INSPECTOR::INSPECTOR
| | | STANDARD-OBJECT
| | | | T
INSPECTOR::UVECTOR-INSPECTOR
| INSPECTOR::BASICS-FIRST-MIXIN
| | * STANDARD-OBJECT
| INSPECTOR::INSPECTOR
| | STANDARD-OBJECT
| | | T
CCL::VAR-CHECK-ITEM
  CHECK-BOX-DIALOG-ITEM CCL::CONTROL-DIALOG-ITEM DIALOG-ITEM SIMPLE-VIEW
  OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
VECTOR
| ARRAY
| | * T
| SEQUENCE
| | T
VIEW
  SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
WARNING
  CONDITION STANDARD-OBJECT T
WINDOID
| WINDOW
| | VIEW
| | | SIMPLE-VIEW
| | | | OUTPUT-STREAM
| | | | | STREAM
| | | | | | CCL::INSTANCE-INITIALIZE-MIXIN
| | | | | | | * STANDARD-OBJECT
| CCL::LISP-WDEF-MIXIN
| | STANDARD-OBJECT
| | | T
WINDOW
  VIEW SIMPLE-VIEW OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN
  STANDARD-OBJECT T
WINDOW-MENU-ITEM
  MENU-ITEM CCL::MENU-ELEMENT CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::WINDOWS-MENU-MENU-ITEM
  MENU-ITEM CCL::MENU-ELEMENT CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::WORD-VECTOR
| VECTOR
| | ARRAY
| | | * T
| | SEQUENCE
| | | T
CCL::XP-STREAM
  OUTPUT-STREAM STREAM CCL::INSTANCE-INITIALIZE-MIXIN STANDARD-OBJECT T
CCL::XP-STRUCTURE
  CCL::INTERNAL-STRUCTURE T

|#
#|
	Change History (most recent last):
	2	12/27/94	akh	merge with d13
|# ;(do not edit past this line!!)
