;;; big-file-patch.lisp
;;; Thursday March 20, 2003 SVS
;;; Patches functions in l1-sysio to enable access to files
;;; up to 2^64 in length.

;;; USE AT YOUR OWN RISK.
;;; This code should not be considered 
;;; trustworthy until much more testing has been done. If you choose
;;; to use this code, do not use it to operate on critical files.

;;; INSTALLATION:
;;; First, load this patch file. But that in itself is not enough. You must also call
;;; activate-big-file-patch to enable access to big files. If you want to activate this patch from your
;;; init file, put this command into your init file (assuming you've already loaded
;;; the patch):
;;; (ccl::activate-big-file-patch *load-pathname*)
;;; This will ensure that your init file is closed before the patch is activated.

;;; DE-INSTALLATION:
;;; To revert file i/o functions to the old (small file) mechanism,
;;; call deactivate-big-file-patch. Again, if you do this from a file that's
;;; being loaded, do it this way:
;;; (ccl::deactivate-big-file-patch *load-pathname*)

;;; You can call activate-big-file-patch and deactivate-big-file-patch programmatically as
;;; often as you wish. This is useful for testing for bugs in the patch and speed testing.

;;; DISCUSSION (please read before using):

;;; What is this?
;;; The original MacOS did not support files longer than 2^31 (2 GB) in length.
;;; Nowadays, with HFS+, Carbon, and MacOSX that limitation has been removed and files up
;;; to 2^64 in length are supported. However, a new lowlevel API is needed to support access
;;; to these big files.

;;; The traditional MCL implementation of Common Lisp file stream i/o functions such as #'file-length,
;;; #'file-position, #'open, #'read, #'write, #'print, #'format, etc. when operated on file streams
;;; uses the old API, and thus won't work on big files. This patch is intended to remedy that
;;; situation. It makes the file stream i/o functions use the new MacOS 64-bit API.
;;; In addition, non-ANSI functions like copy-file, file-info, etc. are modified here to work on
;;; big files. (Copy-file also gets some new behavior here: It does some error-prone stuff first
;;; before it creates its output file. This enables more intelligent error recovery. It also
;;; takes a progress-object parameter so you can copy with a progress bar. See example
;;; at the end of this file.)

;;; After installing this patch, you should still be able to transparently operate on small
;;; files just as before, as well as on files larger than 2^31 in length. However, operations on
;;; small files may be somewhat slower than they were with the old small-file API. I'm still
;;; working on this but there might not be a good solution. Since it's easy to activate and
;;; deactivate this patch programmatically (see above), you can easily test the speed of file operations
;;; with and without the patch activated.

;;; What this patch won't do:
;;; Provide support for long file names. That's a separate issue.
;;; Allow Fred to open larger files (Fred uses a non-ANSI file I/O mechanism and has internal limits).
;;; Allow the Lisp loader to load big source or fasl files. Again, this doesn't use the CL file API.
;;; Provide big file access for programs that use the mac file API directly, bypassing the Common
;;;   Lisp API. Programs like CL-HTTP that use mac-file-io.lisp are in this category.

;;; We do NOT need to check to see if the file in use is on an HFS+ partition (HFS doesn't
;;;   support big files) because the new routines will supposedly work there--they'll just
;;;   throw an error if you try to make a file too big on them.

;;; We now have a routine #'volume-supports-big-files? that programmers can use to check
;;;   whether a volume supports big files.

;;; Still to do: update the routines in mac-file-io.lisp too. Does anybody really care except CL-HTTP?
;;; use #_PBCreateforkSync and friends to support forks other than data and resource. (This is not
;;;   properly a big files issue, and probably belongs in a different patch. And it looks
;;;   like the new Apple management is trying to get away from forks entirely,
;;;   so maybe it's a moot point.

;;; Documentation: http://developer.apple.com/techpubs/macosx/Carbon/Files/FileManager/File_Manager/index.html

(in-package :ccl)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (unless
    (let ((attr (gestalt #$gestaltFSAttr)))
      (and (logbitp #$gestaltHasHFSPlusAPIs attr)
           (logbitp #$gestaltFSSupports4GBVols attr)
           (logbitp #$gestaltFSSupports2TBVols attr)
           (logbitp #$gestaltFSSupportsHFSPlusVols attr)))
    (error "This version of MacOS does not seem to support HFS Plus and/or large filesystems. Big file
           support cannot work here.")))

; Best speed is when *elements-per-buffer* is around 8192, but always more than default 2048
(eval-when (:load-toplevel :execute)
  (unless (> *elements-per-buffer* 2048)
    (setf *elements-per-buffer* 8192)))

(def-accessors (fblock) %svref
  nil                                   ; 'fblock
  fblock.pb                             ; a parameter block; nil if closed.
  fblock.lastchar                       ; untyi char or nil
  fblock.dirty                          ; non-nil when dirty
  fblock.buffer                         ; macptr to buffer; nil when closed
  fblock.bufvec                         ; buffer vector; nil when closed
  fblock.bufsize                        ; size (in 8-bit bytes) of buffer
  fblock.bufidx                         ; index of next element to read/write
  fblock.bufcount                       ; # of elements in buffer
  fblock.filepos                        ; 8-bit position at last read/write
  fblock.fileeof                        ; file's logical eof.
  fblock.stream                         ; backptr to file stream
  fblock.element-type                   ; typespec
  fblock.nbits-per-element              ; # of bits per element
  fblock.elements-per-buffer            ; 512 or whatever
  fblock.minval                         ; minimum value of element type or nil: < 0 
  fblock.maxval                         ; maximum value or nil
  fblock.element-bit-offset             ; for non-arefable n-bit elements
  fblock.scriptruns			; goo for files containing fat scripts
  fblock.charpos			; character position as opposed to byte position
  fblock.endrun				; end of current run (beware untyi!)
  fblock.scriptidx			; current index into scriptruns
  fblock.chartable			; char-byte-table of current script
  )

(eval-when (:execute :compile-toplevel)
  (defmacro with-paramBlock ((fblock pb1 &optional ignore-if-closed) &body body)
    `(let* ((,pb1 (%fblock.pb  ,fblock ,ignore-if-closed)))
       (when (macptrp ,pb1)
         (locally
           (declare (type macptr ,pb1))
           ,@body)))))

(defconstant uw-hi (byte 32 32))
(defconstant uw-lo (byte 32 0))
(defvar *resource-fork-name-ptr* nil "Cached pointer to resource fork name.")
(defvar *data-fork-name-ptr* nil "Cached pointer to data fork name.")

(defvar *nocache-copies* t "True if you don't want memory caches used by copy-file. Somewhat faster this way.")

;; ---- Some utility functions

(defun data-fork-name ()
  (or *data-fork-name-ptr*
      (prog1
        (setf *data-fork-name-ptr* (#_NewPtr (record-length :hfsunistr255)))
        (#_FSGetDataForkName *data-fork-name-ptr*)
        )))

(defun resource-fork-name ()
  (or *resource-fork-name-ptr*
      (prog1
        (setf *resource-fork-name-ptr* (#_NewPtr (record-length :hfsunistr255)))
        (#_FSGetResourceForkName *resource-fork-name-ptr*)
        )))

(defun path-to-fsref (path fsref &optional errchk dont-resolve)
  "Similar in spirit to %path-to-iopb. It would be nice to use
   FSPathMakeRef here, but it apparently is only implemented in OSX"
  (let ((errnum nil))
    (with-pstrs ((path-str (if dont-resolve
                             (mac-namestring path)
                             (mac-namestring (truename path)))))
      (rlet ((fsspec :FSSpec)) ; all this rigamarole just to get an fsref
        (setf errnum (#_FSMakeFSSpec 0 0 path-str fsspec))
        (if (zerop errnum)
          (setf errnum (#_FSpMakeFSRef fsspec fsref))
          (if errchk
            (signal-file-error errnum path-str)
            errnum))))))

(defun volume-supports-big-files? (path)
  "Returns some parameters about the volume indicated in path. Be sure to include the colon
   in the path. Path can refer to a file on the volume if you wish, but only the volume portion
   is used here. Returns nil if volume doesn't exist."
  (let ((dir (mac-directory-namestring path))
        (errcode nil)
        version extended-attributes)
    (%stack-iopb (pb np)
      (%put-string np dir)
      (rlet ((vab GetVolParmsInfoBuffer))
        (%put-ptr pb vab $ioBuffer)
        (%put-long pb (record-length GetVolParmsInfoBuffer) $ioReqCount)
        
        (%put-word pb 0 $ioVRefNum)
        
        (setq errcode (#_PBHGetVolParmsSync pb))
        (when (zerop errcode)
          (setq version (pref vab GetVolParmsInfoBuffer.vMVersion))
          (when (> version 2) ; Must be > 2 if it supports HFS+ APIs directly
            ;http://developer.apple.com/techpubs/macosx/Carbon/Files/FileManager/File_Manager/
            ;   Functions/Getting_Info_bout_Volumes.html#//apple_ref/C/func/PBHGetVolParmsSync
            (setq extended-attributes (pref vab GetVolParmsInfoBuffer.vMExtendedAttributes))
            ;(print extended-attributes)
            (and 
             (logbitp #$bSupportsHFSPlusAPIs extended-attributes) ; true if supports HFS+ API directly
             (logbitp #$bSupports2TBFiles extended-attributes) ; true if supports big files
             ; discussion: we might not need to check $bSupportsHFSPlusAPIs at all.
             ;   Because if $bSupports2TBFiles is true, $bSupportsHFSPlusAPIs might be
             ;   irrelevant for purposes of this function
             )))))))

;; ---- Replacements for file size utilities

(defun utcseconds (utcptr)
  "Returns bignum seconds from a Mac :UTCDateTime pointer."
  (+ ccl::$mac-time-offset ; Because the Mac thinks Universal time starts in 1904, and Lisp thinks it starts in 1900
     (ash (pref utcptr :UTCDateTime.highSeconds) 16)
     (pref utcptr :UTCDateTime.lowSeconds)))

(defun getutcdatetime ()
  "Same as get-universal-time?"
  (rlet ((utc :UTCDateTime))
    (#_GetUTCDateTime utc 0)
    (utcseconds utc)))

(defun %catalog-info (fsref selector-bits)
  "Calls #_FSGetCatalogInfo with the given selector-bits and returns the requested fields of the
   :FSCatalogInfo record. Assumes you don't care about returning the unicode name, the fsspec, or the parentRef.
   Returns results in a list in order of most significant 1-bit in selector-bits to least."
  (rlet ((cataloginfo :FSCataloginfo))
    (setf selector-bits (logand selector-bits #$kFSCatInfoGettableInfo)) ; don't allow any reserved nonsense
    (#_FSGetCatalogInfo
     fsref
     selector-bits
     cataloginfo
     (%null-ptr) ; unicode name out
     (%null-ptr) ; fsspec
     (%null-ptr)) ; parentRef
    (flet ((bnz (v1 v2) ; bit not zero
             (not (zerop (logand v1 v2)))))
      (let ((result nil))
        (when (bnz selector-bits #$kFSCatInfoTextEncoding)
          (push (pref cataloginfo :FSCataloginfo.textEncodingHint) result))
        (when (bnz selector-bits #$kFSCatInfoNodeFlags)
          (push (pref cataloginfo :FSCataloginfo.nodeFlags) result))
        (when (bnz selector-bits #$kFSCatInfoVolume)
          (push (pref cataloginfo :FSCataloginfo.volume) result))
        (when (bnz selector-bits #$kFSCatInfoParentDirID)
          (push (pref cataloginfo :FSCataloginfo.parentDirID) result))
        (when (bnz selector-bits #$kFSCatInfoNodeID)
          (push (pref cataloginfo :FSCataloginfo.nodeID) result))
        (when (bnz selector-bits #$kFSCatInfoCreateDate)
          (push (utcseconds (pref cataloginfo :FSCataloginfo.createDate)) result))
        (when (bnz selector-bits #$kFSCatInfoContentMod)
          (push (utcseconds (pref cataloginfo :FSCataloginfo.contentModDate)) result))
        (when (bnz selector-bits #$kFSCatInfoAttrMod)
          (push (utcseconds (pref cataloginfo :FSCataloginfo.attributeModDate)) result))
        (when (bnz selector-bits #$kFSCatInfoAccessDate)
          (push (utcseconds (pref cataloginfo :FSCataloginfo.accessDate)) result))
        (when (bnz selector-bits #$kFSCatInfoBackupDate)
          (push (utcseconds (pref cataloginfo :FSCataloginfo.backupDate)) result))
        (when (bnz selector-bits #$kFSCatInfoPermissions)
          (push (pref cataloginfo :FSCataloginfo.permissions) result))
        (when (bnz selector-bits #$kFSCatInfoFinderInfo)
          (push (pref cataloginfo :FSCataloginfo.finderInfo) result))
        (when (bnz selector-bits #$kFSCatInfoFinderXInfo)
          (push (pref cataloginfo :FSCataloginfo.extFinderInfo) result))
        (when (bnz selector-bits #$kFSCatInfoValence)
          (push (pref cataloginfo :FSCataloginfo.valence) result))
        (when (bnz selector-bits #$kFSCatInfoDataSizes)
          (push (unsignedwide->integer (pref cataloginfo :FSCataloginfo.dataPhysicalSize)) result)
          (push (unsignedwide->integer (pref cataloginfo :FSCataloginfo.dataLogicalSize)) result))
        (when (bnz selector-bits #$kFSCatInfoRsrcSizes)
          (push (unsignedwide->integer (pref cataloginfo :FSCataloginfo.rsrcPhysicalSize)) result)
          (push (unsignedwide->integer (pref cataloginfo :FSCataloginfo.rsrcLogicalSize)) result))
        (when (bnz selector-bits #$kFSCatInfoSharingFlags)
          (push (pref cataloginfo :FSCataloginfo.sharingFlags) result))
        (when (bnz selector-bits #$kFSCatInfoUserPrivs)
          (push (pref cataloginfo :FSCataloginfo.userPrivileges) result))
        result))))

; This should probably replace the use of the old file-info function. This is more flexible and it handles big files.
(defun catalog-info (path selector-bits &optional dont-resolve)
  (rlet ((fsref :fsref))
    (path-to-fsref path fsref nil dont-resolve)
    (%catalog-info fsref selector-bits)))

(defun file-info-64 (path)
  "For backward compatibility, duplicate the functionality of the old file-info function for 64-bit files."
  (let ((data (catalog-info path (+ #$kFSCatInfoCreateDate
                                    #$kFSCatInfoContentMod
                                    #$kFSCatInfoDataSizes
                                    #$kFSCatInfoRsrcSizes))))
    (destructuring-bind (rsrclog rsrcphys datalog dataphys mod create) data
      (values create mod rsrclog datalog rsrcphys dataphys))))

(defun filename-from-fsref (fsref)
  "Given an fsref, return the name of the file it thinks it's referring to."
  (rlet ((cataloginfo :FSCataloginfo)
         (unistr :HFSUniStr255))
    (#_FSGetCatalogInfo
     fsref
     #$kFSCatInfoNone
     cataloginfo
     unistr ; unicode name out
     (%null-ptr) ; fsspec
     (%null-ptr)) ; parentRef
    (%get-ustring unistr)
    ))

(defun %get-ustring (fname)
  (let* ((len (pref fname :hfsunistr255.length))
         (res (make-string len :element-type 'extended-character)))
    (dotimes (i len)
      (setf (schar res i) (code-char (%get-unsigned-word fname (+ 2 i i)))))
    res))

; (decode-universal-time (car (ccl::catalog-info "foofile" #$kFSCatInfoCreateDate)))
; (decode-universal-time (car (ccl::catalog-info "foofile" #$kFSCatInfoContentMod)))

(defun %file-size-64 (fsref)
  "Returns 4 values representing file sizes of this fsref: resource-logical, resource-physical, data-logical, and data-physical."
  (%catalog-info fsref #.(+ #$kFSCatInfoDataSizes #$kFSCatInfoRsrcSizes)))

; These are slower than the old routines.
; I suspect the time sink is because of the need to call #_FSMakeFSSpec and then #_FSpMakeFSRef
; It would be interesting to try to do these functions using #_PBIterateForksSync instead.

(defun file-resource-size-64 (path &optional dont-resolve)
  (nth 0 (catalog-info path #.(+ #$kFSCatInfoDataSizes #$kFSCatInfoRsrcSizes) dont-resolve)))

(defun file-data-size-64 (path &optional dont-resolve)
  (nth 2 (catalog-info path #.(+ #$kFSCatInfoDataSizes #$kFSCatInfoRsrcSizes) dont-resolve)))

(defun file-allocated-resource-size-64 (path &optional dont-resolve)
  (nth 1 (catalog-info path #.(+ #$kFSCatInfoDataSizes #$kFSCatInfoRsrcSizes) dont-resolve)))

(defun file-allocated-data-size-64 (path &optional dont-resolve)
  (nth 3 (catalog-info path #.(+ #$kFSCatInfoDataSizes #$kFSCatInfoRsrcSizes) dont-resolve)))

;; ---- Replacements for l1-sysio.lisp routines

(defun %fopen-64 (name &optional 
                         (vrefnum 0) 
                         (permission $fsRdWrPerm) 
                         resource
                         stream
                         (element-type 'base-character)
                         (elements-per-buffer *elements-per-buffer*)
                         scriptruns)
  (declare (ignore vrefnum)) ; nobody's using this any more
  (let* ((eof nil)
         (subtype nil)
         (element-size nil)
         (min nil)
         (max nil)
         (bit-offset nil)
         (buffer-octet-size  elements-per-buffer))
    ; need ppc types here too. ok.
    (let ((maybe-subtype (element-type-subtype element-type)))
      (if (eq element-type 'extended-character)
        (setq element-size 16 subtype maybe-subtype) ;$v_xstr)
        (if (subtypep element-type 'character)  ; not really cltl2 but what the foo. Nobody wants 16 bit char files.
          (setq element-size 8 subtype (element-type-subtype 'base-character)) ; $v_sstr)
          (let* ((signed (list 'signed-byte 0))
                 (unsigned (list 'unsigned-byte 0)))
            (declare (dynamic-extent signed unsigned))
            (dotimes (i 32 (report-bad-arg element-type '(or character (signed-byte 32) (unsigned-byte 32))))
              (let* ((j (1+ i)))
                (declare (fixnum j))
                (rplaca (cdr signed) j)
                (unless (= j 1)
                  (if (subtypep element-type signed)
                    (setq element-size j
                          subtype (if (memq j '(8 16 32)) maybe-subtype))))
                (rplaca (cdr unsigned) j)
                (if (subtypep element-type unsigned)
                  (setq element-size j
                        min 0
                        subtype (if (memq j '(1 8 16 32)) maybe-subtype))))
              (if element-size (return)))
            (if min                         ; was unsigned
              (setq max (1- (ash 1 element-size)))
              (setq min (- (ash 1 (1- element-size)))
                    max (1- (ash 1 (1- element-size)))))
            (do* ((m (ash elements-per-buffer 3) (+ m 8)))
                 ((eq element-size (gcd m element-size)) (setq buffer-octet-size (ash m -3)))
              (declare (fixnum m)))
            (unless subtype (setq bit-offset 0))))))
    (setq elements-per-buffer (%octets-to-elements buffer-octet-size element-size))
    ;(print (list elements-per-buffer buffer-octet-size))    
    ;;(with-pstrs ((pname name))
    ;(rlet ((fsref :fsref)) ; nope
    (let ((fsref (make-record (:FSRef :clear t)))) ; gotta keep this thing around, safe and sound
      ;;(fname :hfsunistr255))
      (let ((fname (if resource (resource-fork-name) (data-fork-name))))
        ;;(#_FSGetResourceForkName fname)
        ;;(#_FSGetDataForkName fname))
        (path-to-fsref name fsref nil t)
        ;;(%stack-block ((paramBlock 52 :clear t))
        (rlet ((paramBlock :FSForkIOParam))
          ;;(setf (rref paramblock :hparamblockrec.ioNamePtr) pname
          ;;      (rref paramblock :hparamblockrec.ioVrefNum) vrefnum
          ;;      (rref paramblock :hparamblockrec.ioversnum) 0
          ;;      (rref paramblock :hparamblockrec.ioPermssn) permission
          ;;      (rref paramblock :hparamblockrec.ioMisc) (%null-ptr))
          (setf (pref paramblock :FSForkIOParam.ref) fsref
                (pref paramblock :FSForkIOParam.forkNameLength) (pref fname :hfsunistr255.length)
                (pref paramblock :FSForkIOParam.forkName)       (pref fname :hfsunistr255.unicode)
                (pref paramblock :FSForkIOParam.permissions) permission)
          (let ((res
                 (#_PBOpenForkSync paramblock)
                 ))
            (unless (eql #$noErr res)
              (signal-file-error res name)))
          ;;(#_PBGetEOFSync paramBlock)
          (#_PBGetForkSizeSync paramblock)
          ; unsignedwide->integer would be a bit faster here.
          ;(setf eof (+ (ash (pref paramblock :FSForkIOParam.positionOffset.hi) 32)
          ;             (pref paramblock :FSForkIOParam.positionOffset.lo)))
          (setf eof (unsignedwide->integer (pref paramblock :FSForkIOParam.positionOffset)))
          ;;(setq eof (%get-long paramBlock $ioLEOF))
          (let* ((bufvector nil)
                 (buffer nil))
            (if subtype
              (progn
                (multiple-value-setq (bufvector buffer)
                  (%make-heap-ivector subtype
                                      buffer-octet-size
                                      elements-per-buffer)))
              (setq buffer (#_NewPtr buffer-octet-size)))
            (let* ((pb (make-record (:FSForkIOParam :clear t)))
                   (fblock (%istruct
                            'fblock
                            pb
                            nil
                            nil
                            buffer
                            bufvector
                            buffer-octet-size
                            0
                            0
                            0
                            (%octets-to-elements eof element-size)
                            stream
                            element-type
                            element-size
                            elements-per-buffer
                            min
                            max
                            bit-offset
                            scriptruns
                            0
                            0 
                            -1
                            nil)))
              ;(unless (eql 0 vrefnum)
              (setf 
               ;;(rref pb :hparamblockrec.ioVrefNum) ;; hope we don't need this anymore
               ;;(%stack-iopb (vpb np)
               ;;  (setf (rref vpb :hparamblockrec.iovrefnum) vrefnum
               ;;        (rref vpb :hparamblockrec.ionameptr) pname
               ;;        (rref vpb :hparamblockrec.iovolindex) -1)
               ;;  (#_PBHGetVInfoSync vpb)
               ;;  (rref vpb :hparamblockrec.iovrefnum))
               ;;(rref pb :hparamblockrec.iocompletion) (%null-ptr)
               (rref pb :FSForkIOParam.iocompletion) (%null-ptr)
               
               ;;(rref pb :hparamblockrec.iorefnum) (rref paramBlock :hparamblockrec.ioRefnum)
               (rref pb :FSForkIOParam.forkRefNum) (rref paramBlock :FSForkIOParam.forkRefNum)
               
               (pref pb :FSForkIOParam.ref) fsref ; or (pref paramBlock :FSForkIOParam.ref) ; the fsref
               
               ;;(rref pb :hparamblockrec.ioresult) #$noErr
               (rref pb :FSForkIOParam.ioresult) #$noErr
               
               ;;(rref pb :hparamblockrec.ioposmode) #$fsFromStart
               (rref pb :FSForkIOParam.positionMode) #$fsFromStart
               
               ;(rref pb :hparamblockrec.ioreqcount) buffer-octet-size
               ;;(rref pb :hparamblockrec.iobuffer) buffer
               (rref pb :FSForkIOParam.buffer) buffer
               
               ;;(rref pb :hparamblockrec.ioactcount) 0)
               (rref pb :FSForkIOParam.actualCount) 0)
              ;(print (filename-from-fsref (pref pb :FSForkIOParam.ref)))
              (when (> eof 0)
                (%fbRead fblock 0))
              fblock)))))))

; This has stopped working as of MCL 4.3.1 and Carbonlib 1.3.1. It's returning the wrong volume #
#|
(defun volume-from-forkref (n)
  (rlet ((pb :FSForkCBInfoParam)
         (fi :FSForkInfo))
    (setf (pref pb :FSForkCBInfoParam.desiredRefNum) n
          (pref pb :FSForkCBInfoParam.volumeRefNum) 0
          (pref pb :FSForkCBInfoParam.ref) (%null-ptr)
          (pref pb :FSForkCBInfoParam.ForkInfo) fi
          (pref pb :FSForkCBInfoParam.ForkName) (%null-ptr)
          )
    (#_PBGetForkCBInfoSync pb)
    (break)
    ;(pref pb :FSForkCBInfoParam.volumeRefNum) ; always returns 0
    (pref (pref pb :FSForkCBInfoParam.ForkInfo) :FSForkInfo.volume) ; this works
    ))
|#

; try again
(defun volume-from-fsref (fsref)
  "Given an fsref, return the volume reference number for the volume it's on."
  (rlet ((catalogInfo :FSCatalogInfo))
    (errchk (#_FSGetCatalogInfo fsref #$kFSCatInfoVolume catalogInfo (%null-ptr) (%null-ptr) (%null-ptr)))
    (pref cataloginfo :FSCatalogInfo.volume)))

; This may be dangerous. The parameter block for 64-bit files is a :FSForkIOParam, which
;   is NOT what #_PBFlushVolSync is expecting. But it may be similar enough that the
;   "reserved" fields in a :FSForkIOParam are what #_PBFlushVolSync expects.
; Let's not take the chance. Best I can tell, the proper reserved field of a :FSForkIOParam
;   [where ioVRefNum used to be] always contains 0. So let's be sure by calling volume-from-fsref
;   instead and use a separate pb.
(defun %fclose-64 (fblock &optional flush-only)
  (with-paramblock (fblock pb1 t)
    ;(print (filename-from-fsref (pref pb1 :FSForkIOParam.ref)))
    (rlet ((vpb :VolumeParam))
      (let ((vnum ;(volume-from-forkref (pref pb1 :FSForkIOParam.forkrefnum)) ; broken
             (volume-from-fsref (pref pb1 :FSForkIOParam.ref))
             ))
        (setf (pref vpb :VolumeParam.ioNamePtr) (%null-ptr)
              (pref vpb :VolumeParam.ioVRefNum) vnum)
        (unwind-protect
          (%fbFlush fblock)  ;; may fail if disk full
          (unless flush-only
            (errchk (#_PBCloseForkSync pb1)))
          ; http://developer.apple.com/techpubs/macosx/Carbon/Files/FileManager/File_Manager/DataTypes/FSRef.html
          ; says "There is no need to call the File Manager to dispose an FSRef." So it's probably okay.
          ;(when (macptrp (rref pb1 :FSForkIOParam.ref))
          ;  (#_DisposePtr (rref pb1 :FSForkIOParam.ref))) ; dump the fsref
          (if flush-only
            (progn
              (errchk (#_PBFlushForkSync pb1))
              (if (eq flush-only 'finish-output)
                (errchk (#_PBFlushVolSync vpb))))
            (progn
              (errchk (#_PBFlushVolSync vpb))
              (#_DisposePtr pb1)
              (setf (fblock.pb fblock) nil)
              (let* ((bufvector (fblock.bufvec fblock)))
                (if bufvector
                  (%dispose-heap-ivector bufvector)
                  (#_DisposePtr (fblock.buffer fblock))))
              (setf (fblock.buffer fblock) nil
                    (fblock.bufvec fblock) nil)))
          nil)))))

(defun %fsize-64 (fblock &optional new-size)
  (if (null new-size)
    (let* ((eof (fblock.fileeof fblock))
           (eof1 (if (fblock.dirty fblock)
                   (+ (%octets-to-elements (fblock.filepos fblock) (fblock.nbits-per-element fblock))
                      (max (fblock.bufcount fblock)
                           (fblock.bufidx fblock)))
                   0)))
      (when (fblock.lastchar fblock)
        (decf eof1))
      (max eof eof1))
    (let* ((pb (%fblock.pb fblock))
           (nbits-per-element (fblock.nbits-per-element fblock))
           (offset (- new-size (%octets-to-elements (fblock.filepos fblock) nbits-per-element)))
           (8-bit-size (%elements-to-octets new-size nbits-per-element)))
      (if (and (>= offset 0) (< offset (fblock.bufsize fblock)))
        (setf (fblock.bufcount fblock) offset))
      ; Note: even though we're messing with positionOffset here -- it's necessary for #_PBSetForkSizeSync --
      ;    it won't hurt anything because positionOffset gets reset to the current file position
      ;    in %fbread and %fbflush.
      (setf (pref pb :FSForkIOParam.positionOffset.hi) (ldb uw-hi 8-bit-size)
            (pref pb :FSForkIOParam.positionOffset.lo) (ldb uw-lo 8-bit-size))
      (setf (fblock.fileeof fblock) new-size)
      ; assumes positionMode has already been set to #$fsfromstart
      ;(errchk (#_PBAllocateForkSync pb)) ; errors (of course) when there's no allocation necessary!
      (errchk (#_PBSetForkSizeSync pb))
      new-size)))

#+EXPERIMENTAL
(defun %fsize-64 (fblock &optional new-size)
  ; An attempt to speed this up
  (if (null new-size)
    (let* ((eof (fblock.fileeof fblock))
           (eof1 (if (fblock.dirty fblock)
                   (+ (%octets-to-elements (fblock.filepos fblock) (fblock.nbits-per-element fblock))
                      (max (fblock.bufcount fblock)
                           (fblock.bufidx fblock)))
                   0)))
      (when (fblock.lastchar fblock)
        (decf eof1))
      (max eof eof1))
    (let* ((pb (%fblock.pb fblock))
           (nbits-per-element (fblock.nbits-per-element fblock))
           (offset (- new-size (%octets-to-elements (fblock.filepos fblock) nbits-per-element)))
           (8-bit-size (%elements-to-octets new-size nbits-per-element)))
      (if (and (>= offset 0) (< offset (fblock.bufsize fblock)))
        (setf (fblock.bufcount fblock) offset))
      ; Note: even though we're messing with positionOffset here -- it's necessary for #_PBSetForkSizeSync --
      ;    it won't hurt anything because positionOffset gets reset to the current file position
      ;    in %fbread and %fbflush. (Well actually, positionOffset has nothing to do with current file
      ;    position. That's stored in a hidden field somewhere.)
      (setf (pref pb :FSForkIOParam.positionOffset.hi) (ldb uw-hi 8-bit-size)
            (pref pb :FSForkIOParam.positionOffset.lo) (ldb uw-lo 8-bit-size))
      (setf (fblock.fileeof fblock) new-size)
      ;(setf alloc-amount (- 8-bit-size .... ; ah what the heck. fuhggetaboutit
      (setf (pref pb :FSForkIOParam.allocationAmount.hi) (ldb uw-hi 8-bit-size)
            (pref pb :FSForkIOParam.allocationAmount.lo) (ldb uw-lo 8-bit-size))
      ; assumes positionMode has already been set to #$fsfromstart
      (setf (pref pb :FSForkIOParam.allocationFlags) #$kFSAllocContiguousMask ) ; #$kFSAllocDefaultFlags)
      (#_PBAllocateForkSync pb) ; see if this speeds things up in OSX (other way is fast in OS9)
      ; Boy is this wrong. You get errors if you do allocate and don't setforksize.
      ;(#_PBSetForkSizeSync pb)
      new-size)))

(defun %fbRead-64 (fblock base)
  (let* ((element-size (fblock.nbits-per-element fblock))
         (pb (%fblock.pb fblock)))
    (declare (type macptr pb))
    
    (setf (fblock.filepos fblock) base)         ; 8-bit
    (setf (rref pb :FSForkIOParam.positionOffset.hi) (ldb uw-hi base)
          (rref pb :FSForkIOParam.positionOffset.lo) (ldb uw-lo base)
          (rref pb :FSForkIOParam.requestCount) (fblock.bufsize fblock)
          (rref pb :FSForkIOParam.positionMode) #$fsFromStart)
    (let* ((err (#_PBReadForkSync pb)))
      (unless (eql #$noErr err)
        (unless (eql #$eofErr err)
          (signal-file-error err))))
    (setf (fblock.bufidx fblock) 0)
    (unless (fblock.bufvec fblock)
      (setf (fblock.element-bit-offset fblock) 0))
    (setf (fblock.bufcount fblock) (%octets-to-elements (rref pb :FSForkIOParam.actualCount) element-size))))

; Flush buffer if it's dirty.
(defun %fbFlush-64 (fblock)
  (when (fblock.dirty fblock)
    (setf (fblock.dirty fblock) nil)
    (let* ((element-size (fblock.nbits-per-element fblock))
           (pb (%fblock.pb fblock))
           (8-bit-filepos (fblock.filepos fblock))
           (filepos (%octets-to-elements 8-bit-filepos (fblock.nbits-per-element fblock)))
           (elements-written (fblock.bufidx fblock))
           (elements-read (fblock.bufcount fblock))
           (new-eof (+ filepos (max elements-written elements-read))))
      (declare (type macptr pb))
      (setf (rref pb :FSForkIOParam.positionOffset.hi) (ldb uw-hi 8-bit-filepos)
            (rref pb :FSForkIOParam.positionOffset.lo) (ldb uw-lo 8-bit-filepos)
            (rref pb :FSForkIOParam.positionMode) #$fsFromStart
            (rref pb :FSForkIOParam.requestCount) (%elements-to-octets (max elements-written elements-read) element-size)) 
      (when (> new-eof (fblock.fileeof fblock))
        (setf (fblock.fileeof fblock) new-eof))
      (errchk (#_PBWriteForkSync pb)))))

; These are some hooks that you can specialize for a user-provided object. If you pass an object
;   in the :progress-object parameter of #'copy-file, these methods get called automatically. Presumably,
;   you would use them to track the progress of a copy operation.
(defmethod initialize-copy-progress ((me t) n)
  (declare (ignore n))
  "Initializes a progress object with the value n, presumably the length of a progress bar."
  nil)

(defmethod increment-copy-progress ((me t) x)
  (declare (ignore x))
  "Increment progress object by value x."
  nil)

(defmethod finish-copy-progress ((me t))
  "Tell progress object we're done with it."
  t)

(defun %copy-fork-64 (srcpb dstpb resource-p blocksz &optional (progress-object nil))
  ; Now assumes srcpb has already been successfully opened for reading.
  (let ((successful-dest -99)
        (increment-skip 0))
    (unwind-protect
      (progn
        (setf (pref dstpb :FSForkIOParam.permissions) #$fsWrPerm)
        (if resource-p
          (setf (pref dstpb :FSForkIOParam.forkName)       (pref (resource-fork-name) :hfsunistr255.unicode)
                (pref dstpb :FSForkIOParam.forkNameLength) (pref (resource-fork-name) :hfsunistr255.length))
          (setf (pref dstpb :FSForkIOParam.forkName)       (pref (data-fork-name) :hfsunistr255.unicode)
                (pref dstpb :FSForkIOParam.forkNameLength) (pref (data-fork-name) :hfsunistr255.length)))
        (errchk (setq successful-dest (#_PBOpenForkSync dstpb)))
        
        (errchk (#_PBGetForkSizeSync srcpb))
        (let ((reslen ;(+ (ash (pref srcpb :FSForkIOParam.positionOffset.hi) 32)
               ;   (pref srcpb :FSForkIOParam.positionOffset.lo))))
               (unsignedwide->integer (pref srcpb :FSForkIOParam.positionOffset)))) ; this is faster
          (setf (pref dstpb :FSForkIOParam.positionOffset.hi) (ldb uw-hi reslen)
                (pref dstpb :FSForkIOParam.positionOffset.lo) (ldb uw-lo reslen))
          (errchk (#_PBSetForkSizeSync dstpb))
          
          (setf (rref srcpb :FSForkIOParam.positionOffset.hi) 0
                (rref srcpb :FSForkIOParam.positionOffset.lo) 0
                (rref dstpb :FSForkIOParam.positionOffset.hi) 0 
                (rref dstpb :FSForkIOParam.positionOffset.lo) 0
                
                (rref srcpb :FSForkIOParam.positionMode) #$fsFromStart
                (rref dstpb :FSForkIOParam.positionMode) #$fsFromStart
                
                (rref srcpb :FSForkIOParam.requestCount) blocksz
                (rref dstpb :FSForkIOParam.requestCount) blocksz)
          (errchk (#_PBSetForkPositionSync srcpb)) ; Carbon is weird. You didn't have to do this before.
          (errchk (#_PBSetForkPositionSync dstpb)) ; Bottom line: positionOffset is not necessarily the same thing as position.
          (setf (rref srcpb :FSForkIOParam.positionMode) #$fsAtMark ; Ditto
                (rref dstpb :FSForkIOParam.positionMode) #$fsAtMark)
          (when *nocache-copies*
            (setf (rref srcpb :FSForkIOParam.positionMode) (+ #$fsAtMark #$noCacheMask)) ; this speeds up the copy
            ;(setf (rref dstpb :FSForkIOParam.positionMode) (+ #$fsAtMark #$noCacheMask)) ; this slows it down. Dramatically.
            ; I guess the lesson is "Always cache your writes, but don't cache your reads unless necessary."
            )
          (while (> reslen blocksz)
            (errchk (#_PBReadForkSync srcpb))
            (errchk (#_PBWriteForkSync dstpb))
            (when (and progress-object (> (incf increment-skip) 9))
              (increment-copy-progress progress-object (* blocksz increment-skip))
              (setf increment-skip 0))
            (setq reslen (- reslen blocksz)))
          (unless (zerop reslen)
            (setf (rref srcpb :FSForkIOParam.requestCount) reslen)
            (setf (rref dstpb :FSForkIOParam.requestCount) reslen)
            (errchk (#_PBReadForkSync srcpb))
            (errchk (#_PBWriteForkSync dstpb)))))
      (when (zerop successful-dest) (errchk (#_PBCloseForkSync dstpb))))))

; Carbon version of copy-file. Supports 64-bit file forks.
; Ultimately, this should probably be rewritten in terms of #_PBHCopyFileSync
(defun copy-file-64 (old-path new-path &key (if-exists :error) fork (blocksz *elements-per-buffer*)
                                (finder-comment-p t) ; This way, if you don't specify it, it does the right thing.
                                (progress-object nil)
                                &aux true-new-path dfp rfp srcvol dstvol file-sizes bytestocopy
                                both-forks-p finder-info-p)
  (declare (optimize (debug 3) (speed 0)))
  ; Now ensures that it can open the source pb before it creates (or overwrites!) the dest file.
  (unless fork (setq fork :both))
  (unless (memq fork '(:both :data :resource))
    (report-bad-arg fork '(member :both :data :resource)))
  (setq both-forks-p (eq fork :both))
  (when (directory-pathname-p old-path)
    (signal-file-error $xnocopydir old-path))
  (setq new-path (merge-pathnames new-path old-path))
  (rlet ((old-fsref :fsref)
         (new-fsref :fsref))
    ;;(dname :hfsunistr255)
    ;; (rname :hfsunistr255))
    (let ((rname (resource-fork-name))
          (dname (data-fork-name)))
      ;;(#_FSGetResourceForkName rname)
      ;;(#_FSGetDataForkName dname)
      
      (%stack-iopb (opb onp)
        (%stack-iopb (npb nnp)
          (flet ((ensure-destination-file (dstpb)
                   ; Create the destination file, set up its finder info, and set up its
                   ;   parameter block (the input dstpb) for the copy.
                   (unless true-new-path
                     (setq true-new-path 
                           (or (and (not both-forks-p)
                                    (probe-file new-path))
                               (progn
                                 (setq finder-info-p t)
                                 (create-file new-path :if-exists if-exists))))
                     (when (not true-new-path) (error "Couldn't find/create a valid destination file."))
                     (path-to-fsref (mac-namestring true-new-path) new-fsref nil t)
                     (%path-to-iopb true-new-path npb :errchk)
                     (setq srcvol (%get-word opb $iovrefnum))
                     (setq dstvol (%get-word npb $ioVRefNum))
                     (%put-word opb dstvol $ioVRefNum) ;;c
                     (%put-long opb (%get-long npb $ioDirID) $ioDirID) ;;c
                     (%put-ptr opb (%get-ptr npb $ioFileName) $ioFileName)         ;;c
                     (when finder-info-p  ;;c 
                       (file-errchk (#_PBHSetFInfoSync opb) new-path))  ;;c ; set most file info of new to be same as old 
                     ; (%put-word dstpb dstvol $iovrefnum)
                     (setf (pref dstpb :FSForkIOParam.ref) new-fsref
                           (pref dstpb :FSForkIOParam.permissions) #$fsWrPerm
                           )
                     )))
            
            (%path-to-iopb old-path opb :errchk)
            (setq old-path (%path-from-iopb opb))   ; truename 
            (%put-word opb 0 $ioFDirIndex)
            (file-errchk (#_PBHGetFInfoSync opb) old-path)
            (setq dfp (not (%null-ptr-p (%get-ptr opb $ioFlPyLen)))) ; there's probably a cleaner way to do this in Carbon
            (setq rfp (not (%null-ptr-p (%get-ptr opb $ioFlRPyLen))))
            
            ;(with-pstrs ((srcname (mac-namestring old-path))
            ;             )
            (path-to-fsref (mac-namestring old-path) old-fsref nil t)
            (unwind-protect
              (progn
                (%stack-block ((buf blocksz))
                  (rlet ((srcpb :FSForkIOParam)
                         (dstpb :FSForkIOParam))
                    ;(%put-word srcpb srcvol $iovrefnum)
                    ;(%put-ptr srcpb srcname $iofilename)
                    ;(%put-ptr srcpb buf $iobuffer)
                    ;(%put-ptr dstpb buf $iobuffer)
                    (setf (rref srcpb :FSForkIOParam.buffer) buf
                          (rref dstpb :FSForkIOParam.buffer) buf)
                    (setf (pref srcpb :FSForkIOParam.ref) old-fsref
                          (pref srcpb :FSForkIOParam.permissions) #$fsRdPerm)
                    (setf file-sizes (%file-size-64 old-fsref)
                          bytestocopy 0)
                    (cond (both-forks-p
                           (incf bytestocopy (+ (car file-sizes) (nth 2 file-sizes))))
                          ((eq fork :data)
                           (incf bytestocopy (nth 2 file-sizes)))
                          ((eq fork :resource)
                           (incf bytestocopy (car file-sizes))))
                    
                    (when progress-object (initialize-copy-progress progress-object bytestocopy)) ; in case user cares about progress
                    (when (memq fork '(:data :both))      ;data fork
                      (setf (pref srcpb :FSForkIOParam.forkName)       (pref dname :hfsunistr255.unicode)
                            (pref srcpb :FSForkIOParam.forkNameLength) (pref dname :hfsunistr255.length))
                      ; buffer address for _Open & _OpenRf
                      ; (setf (pref srcpb :HParamBlockRec.ioMisc) (%null-ptr))
                      (errchk (#_PBOpenForkSync srcpb))
                      (unwind-protect
                        (progn
                          
                          (ensure-destination-file dstpb) ; do this AFTER the source is opened successfully
                          (when dfp ; don't copy if the fork is zero-length, but do make a new zero-length destination
                            ;(with-pstrs ((dstname (mac-namestring true-new-path)))
                            
                            ;(%put-ptr dstpb dstname $iofilename)
                            
                            (%copy-fork-64 srcpb dstpb nil blocksz progress-object)))
                        (errchk (#_PBCloseForkSync srcpb))))
                    (when (memq fork '(:resource :both))  ; resource fork
                      (setf (pref srcpb :FSForkIOParam.forkName)       (pref rname :hfsunistr255.unicode)
                            (pref srcpb :FSForkIOParam.forkNameLength) (pref rname :hfsunistr255.length))
                      ; buffer address for _Open & _OpenRf
                      ;(setf (pref srcpb :HParamBlockRec.ioMisc) (%null-ptr))
                      (errchk (#_PBOpenForkSync srcpb))
                      (unwind-protect
                        (progn
                          
                          (ensure-destination-file dstpb) ; do this AFTER the source is opened successfully
                          (when rfp ; don't copy if the fork is zero-length, but do make a new zero-length destination
                            ;(with-pstrs ((dstname (mac-namestring true-new-path)))
                            
                            ;(%put-ptr dstpb dstname $iofilename)
                            
                            (%copy-fork-64 srcpb dstpb t blocksz progress-object)))
                        (errchk (#_PBCloseForkSync srcpb))))
                    (when finder-info-p
                      (file-errchk (#_PBHSetFInfoSync opb) new-path)) ; to get the dates right
                    ;(%put-ptr dstpb (%null-ptr) $iofilename)
                    ;(%put-word dstpb -1 $iovolindex)
                    ;(#_PBFlushVolSync :errchk dstpb)             
                    ; Now do the same trick we used in %fclose-64 with a special vpb
                    (rlet ((vpb :VolumeParam))
                      (let ((vnum ;(volume-from-forkref (pref dstpb :FSForkIOParam.forkrefnum)))) ; broken. Don't know why
                             (volume-from-fsref (pref dstpb :FSForkIOParam.ref)))) ; correct
                        ;(break)
                        (setf (pref vpb :VolumeParam.ioNamePtr) (%null-ptr)
                              (pref vpb :VolumeParam.ioVRefNum) vnum ; this works
                              ; (pref vpb :VolumeParam.ioVRefNum) dstvol ; This works fine too
                              )
                        (errchk (#_PBFlushVolSync vpb))))
                    ))
                (when finder-comment-p
                  (set-finder-comment new-path (get-finder-comment old-path))))
              (when progress-object (finish-copy-progress progress-object)))
            (values new-path old-path true-new-path)))))))

(defun remember-old-sysio-functions ()
  (unless (fboundp '%fopen-32) ; we must never do this twice
    (setf (symbol-function '%fopen-32) #'%fopen)
    (setf (symbol-function '%fclose-32) #'%fclose)
    (setf (symbol-function '%fbRead-32) #'%fbRead)
    (setf (symbol-function '%fbFlush-32) #'%fbFlush)
    (setf (symbol-function '%fsize-32) #'%fsize)
    (setf (symbol-function 'copy-file-32) #'copy-file)
    (setf (symbol-function 'file-info-32) #'file-info)
    (setf (symbol-function 'file-resource-size-32) #'file-resource-size)
    (setf (symbol-function 'file-data-size-32) #'file-data-size)
    (setf (symbol-function 'file-allocated-resource-size-32) #'file-allocated-resource-size)
    (setf (symbol-function 'file-allocated-data-size-32) #'file-allocated-data-size)
    ))

; Gotta play this game or this file won't compile properly--compiler can't even find traps
;   correctly if file access mechanism is in flux.
(defun install-new-sysio-64 ()
  (remember-old-sysio-functions)
  (let ((*warn-if-redefine-kernel* nil)
        (*warn-if-redefine* nil))
    (setf (symbol-function '%fopen) #'%fopen-64)
    (setf (symbol-function '%fclose) #'%fclose-64)
    (setf (symbol-function '%fbRead) #'%fbRead-64)
    (setf (symbol-function '%fbFlush) #'%fbFlush-64)
    (setf (symbol-function '%fsize) #'%fsize-64)
    (setf (symbol-function 'copy-file) #'copy-file-64)
    (setf (symbol-function 'file-info) #'file-info-64)
    ; I'm going to leave the following as separately-named functions (with "-64" suffix) for now
    ;   because even though they subsume the functionality of the old functions, they're
    ;   only about half as fast. And since 99% of the time the old ones work fine, let's
    ;   leave them alone.
    ;(setf (symbol-function 'file-resource-size) #'file-resource-size-64)
    ;(setf (symbol-function 'file-data-size) #'file-data-size-64)
    ;(setf (symbol-function 'file-allocated-resource-size) #'file-allocated-resource-size-64)
    ;(setf (symbol-function 'file-allocated-data-size) #'file-allocated-data-size-64)
    t
    ))

(defun revert-sysio ()
  "Run this to revert the sysio functions back to their old 32-bit versions. Useful for debugging."
  (unless (fboundp '%fopen-32)
    (error "Cannot revert the sysio functions. They must not have been properly remembered."))
  (let ((*warn-if-redefine-kernel* nil)
        (*warn-if-redefine* nil))
    (setf (symbol-function '%fopen)    (symbol-function '%fopen-32))
    (setf (symbol-function '%fclose)   (symbol-function '%fclose-32))
    (setf (symbol-function '%fbRead)   (symbol-function '%fbRead-32))
    (setf (symbol-function '%fbFlush)  (symbol-function '%fbFlush-32))
    (setf (symbol-function '%fsize)    (symbol-function '%fsize-32))
    (setf (symbol-function 'copy-file) (symbol-function 'copy-file-32))
    (setf (symbol-function 'file-info) (symbol-function 'file-info-32))
    (setf (symbol-function 'file-resource-size) (symbol-function 'file-resource-size-32))
    (setf (symbol-function 'file-data-size) (symbol-function 'file-data-size-32))
    (setf (symbol-function 'file-allocated-resource-size) (symbol-function 'file-allocated-resource-size-32))
    (setf (symbol-function 'file-allocated-data-size) (symbol-function 'file-allocated-data-size-32))
    t
    ))

(defun activate-big-file-patch (&optional filename)
  "Waits for filename not to be open, then closes all open files and installs new 64-bit file IO
   mechanism. Filename parameter is designed to be passed *load-pathname*.
   Note that all conventional file access mechanisms for 'small' files (those less than 2^31 in length)
   should still work just fine after the big-file-patch is activated. The reason we have this
   dramatic 'activation' mechanism is that
   a) We want to keep this as a patch until it's thoroughly debugged, and
   b) Because this patch affects file i/o, we can't just blithely redefine file i/o mechanisms while
      the patch itself is being loaded from a file. Duh. Therefore, we load a bunch of new routines
      and then 'activate' them by assigning the code from the new routines to the old symbols.
   If this activation causes problems, you can always call #'deactivate-big-file-patch"
  (process-run-function "updating for 64-bit files"
                        #'(lambda (thisfile)
                            (loop while 
                                  (and thisfile
                                       (member (namestring thisfile) *open-file-streams* :test 'string-equal :key 
                                               #'(lambda (thing) (namestring (ccl::stream-filename thing)))))
                                  do nil)
                            (without-interrupts
                             (mapcar #'close *open-file-streams*) ; gotta do this because
                             ; you won't be able to access stuff in open files unless you re-open them
                             ; under the new code. All the files MCL tends to keep open will be
                             ; reopened if necessary.
                             (install-new-sysio-64)))
                        filename
                        ))

(defun deactivate-big-file-patch (&optional filename)
  "Turn off the big file patch. Useful for debugging.
   See comments for activate-big-file-patch."
  (process-run-function "reverting to 32-bit files"
                        #'(lambda (thisfile)
                            (loop while 
                                  (and thisfile
                                       (member (namestring thisfile) *open-file-streams* :test 'string-equal :key 
                                               #'(lambda (thing) (namestring (ccl::stream-filename thing)))))
                                  do nil)
                            (without-interrupts
                             (mapcar #'close *open-file-streams*) ; gotta do this because
                             ; you won't be able to access stuff in open files unless you re-open them
                             ; under the new code. All the files MCL tends to keep open will be
                             ; reopened if necessary.
                             (revert-sysio)))
                        filename
                        ))

(defun clear-forkname-caches ()
  (setf *resource-fork-name-ptr* nil
        *data-fork-name-ptr* nil))

(eval-when (:load-toplevel :execute)
  (pushnew 'clear-forkname-caches *restore-lisp-functions*))


#| Test

(with-open-file (stream "Firewire 30gb:bigfoo" :direction :output :if-exists :overwrite)
  (file-length stream 500000)
  (file-position stream 450000)
  (write-string "Foo!" stream))

; Test writing to a specific place in a big honking file. This takes forever in OSX (both native and classic,
;  in any version of MCL).
;  Seems to be the allocation itself--it's like it's trying to zero the new chunk of allocated space, or
;  something. Runs fast in OS9. 
; Friday December 13, 2002 Yep--We tried it with C in OSX. Still takes forever. Perhaps because OSX is multiuser, it's
;   a security thing to zero out newly-allocated blocks???
(with-open-file (stream "Firewire 30gb:bigfoo" :direction :output :if-exists :overwrite)
  (file-length stream 5000000000) ; can comment this out and still works but takes e-v-e-n l-o-n-g-e-r. Hours.
  (file-position stream 4500000000)  ; because this writes one character at a time to extend the file. Yech.
  (write-string "Foo!" stream)
  )

; Test readback from that same place
(with-open-file (stream "Firewire 30gb:bigfoo" :direction :input)
  (let ((str (make-string 4)))
    (file-position stream 4500000000)
    (read-sequence str stream)
    str))

|#

#|
; This stuff requires progress-indicator.lisp

(defmethod initialize-copy-progress ((me ccl::task) n)
  "Initializes a progress object with the value n, presumably the length of a progress bar."
  (show-progress me :denominator n))

(defmethod increment-copy-progress ((me ccl::task) x)
  "Increment progress object by value x."
  (show-progress me :step x
                 :note (format nil "Bytes: ~:D" (task-numerator me))))

(defmethod finish-copy-progress ((me ccl::task))
  "Tell progress object we're done with it."
  (end-progress me))

(defun copy-file-with-progress (src dest &rest rest)
  (let ((progress-object (make-progress (format nil "Copying ~A" (file-namestring src))
                       :numerator 0
                       :denominator 1 ; because we don't know yet
                       :process *current-process*
                       )))
    (apply 'copy-file src dest :progress-object progress-object rest)))
; (copy-file-with-progress "disk1:foo" "disk2:")

|#