; -*- package: plasm -*-

;=============================================================================
;
;                    This is the file eval-def.lsp
;
;=============================================================================

(provide 'eval_def)

(inpackage plasm)


;*********************************************************************************
;Questo modulo e' il responsabile della valutazione delle definizioni delle
;funzioni definibili in una sessione di PLASM.
;Quando una parte di uno script e' riconosciuta come una definizione, la funzione
;EVAL-DEFIN viene chiamata sull'albero sintattico ad essa corrispondente.
;Esistono diverse regole con le quali una defizione viene riconosciuta dal parser:
;infatti alcune parti in una defizione sono optzionali e quindi esistono diverse
;regole per catturare tutti i possibili casi di presenza ed assenza di queste.
;Nell'albero sintattico la radice dell'albero contiene una informazione sia per
;quale nonterminale l'albero rappresenta, e sia  quale fra le regole (produzio-
;ni) della grammatica e' stata utilizzata.
;*********************************************************************************




;----------------------------------------------------------------------------------
;Il corpo di una funzione utente va intepretata in modo da costruire una forma lisp
;equivalente alle chiamate successive delle funzioni dell'interprete. Per far cio'
;e' necessario sospendere le riduzioni di funzioni e costruire delle forme lisp
;appropriate. La sospensione avviene se il valore della variabile 
;**DEFINITION-EVALUATING** e' T.

(setf **DEFINITION-EVALUATING** NIL)

(defun DEFINITION-P ()
  (declare (special **DEFINITION-EVALUATING**))
  **DEFINITION-EVALUATING**)


(defun RESET-**DEFINITION-EVALUATING** ()
  (declare (special **DEFINITION-EVALUATING**))
  (setf **DEFINITION-EVALUATING** nil))

(defun SET-**DEFINITION-EVALUATING** ()
  (declare (special **DEFINITION-EVALUATING**))
  (setf **DEFINITION-EVALUATING** t)) ;So we can stop the evalution of an 
                                      ;expression, because it's not the right
                                      ;moment to get its evaluation!

;----------------------------------------------------------------------------------


(setf **TOKEN-FOUND-IN-DEFINITION** nil)

(defun RESET-**TOKEN-FOUND-IN-DEFINITION** ()
  (declare (special **TOKEN-FOUND-IN-DEFINITION**))
  (setf **TOKEN-FOUND-IN-DEFINITION** nil))

(setf **LOCAL-FUNS-FOUND-IN-BODY** nil)

(defun RESET-**LOCAL-FUNS-FOUND-IN-BODY** ()
  (declare (special **LOCAL-FUNS-FOUND-IN-BODY**))
  (setf **LOCAL-FUNS-FOUND-IN-BODY** nil))


;------------------------------------------------------------------------------


(defun RESET-ENVIRONMENT-DEFINITION ()
  (RESET-**DEFINITION-EVALUATING**))


(defun SET-ENVIRONMENT-DEFINITION ()
  (RESET-**TOKEN-FOUND-IN-DEFINITION**)
  (RESET-**LOCAL-FUNS-FOUND-IN-BODY**)
  (SET-**DEFINITION-EVALUATING**))


;------------------------------------------------------------------------------
;Quando una funzione utente F viene interpretata si pone il problema di vedere quali
;sono le funzioni utilizzate nel suo corpo. Infatti, e' necessario immagazzinare 
;nella slot called-function della funzione F il nome di tutte le funzioni da cui essa
;dipende. Le funzioni primitive non vanno considerate perche' l'utente non puo'
;ridefinirle.  


(setf **TAKE-MEMORY** NIL)


(defun PROCESSED-NAME (name)
  (declare (special **TOKEN-FOUND-IN-DEFINITION**
                    **DEFAULT-FENV**
                    **TAKE-MEMORY**
                    **LOCAL-FUNS-FOUND-IN-BODY**))
  (if (and (symbolp name)
           (not (gethash name **DEFAULT-FENV**)))
    (if (not (member name **TOKEN-FOUND-IN-DEFINITION**))
      (setf **TOKEN-FOUND-IN-DEFINITION** 
            (cons name **TOKEN-FOUND-IN-DEFINITION**))))
  (if (member name **TAKE-MEMORY**)
    (setf **LOCAL-FUNS-FOUND-IN-BODY** 
          (cons name **LOCAL-FUNS-FOUND-IN-BODY**)))
  name)

;-----------------------------------------------------------------------------------
; these functions are not more used (do a fine check)

;(defun TAKE-MEMORY (names)
;  (declare (special **TAKE-MEMORY**))
;  (setf **TAKE-MEMORY** names))

;(defun NO-TAKE-MEMORY ()
;  (declare (special **TAKE-MEMORY**))
;  (setf **TAKE-MEMORY** nil))

;-----------------------------------------------------------------------------------


(defun MAKE-AUTO-EVAL-FORM (body-form)
  (declare (special function-name local-function-names))
  (cond
   ((atomic-p body-form)
    body-form)
   ((equal function-name body-form) `',body-form)
   ((member body-form local-function-names)
    `(take-me ,body-form))
   ((eq nil body-form) '())  ;nil is a symbol !!!
   ((eq t body-form) t)  ;t is a symbol for true!!!
   ((symbolp body-form)
    (if (existing-real-function-p body-form)
      `',body-form
      `(set-and-or-return-value ',body-form)))
   (T
    (cons 'list
          (mapcar #'make-auto-eval-form body-form)))))



(defmacro TAKE-ME (arg)
  `(if (not-value-p ,arg)
     (setf ,arg  (eval ,arg))
     ,arg))


(defun NOT-VALUE-P (arg)
  "this function test if the TAKE-ME argument comes from a just evaluated
   Plasm function. If this is the case then TAKE-ME function must only
   return its parameter, otherwise it must evaluated the parameter (the name
   used in a WHERE END plasm function definition" 
  (and (consp arg)
       (symbolp (first arg))
       (fboundp (first arg))))




(defun ATOMIC-P (obj)
  (declare (special flat-parameter-names local-function-names)) 
                          ;parameter and local names
  (or (and (atom obj) (not (symbolp obj)))
      (and (consp obj) (equal 'reduce-function (first obj)))
      (and (boundp 'flat-parameter-names) (member obj flat-parameter-names))))



(defun EXISTING-REAL-FUNCTION-P (name)
  (let ((fun (bound-function-p name)))
    (and fun
         (plasm-function-formal-pars fun))))


;-----------------------------------------------------------------------------------



(defun EVAL-DEFIN (tree)
  (declare (special **TOKEN-FOUND-IN-DEFINITION**))

  (set-environment-definition) ; switch on definition checking 
  
  (let* ((sons (pick-up-sons tree))
         
         (function-name ; the function name is always his second son
          (break-if-no-good-function-name (second sons)))
         
         
         (old-fun (bound-function-p function-name))
                  
         (parameter-names (pick-up-params-names
                           (is-there-this-son 'PARAMS sons)))                 
         (flat-parameter-names (apply #'append parameter-names))
         parameters
         
         (local-function-names (pick-up-all-names-of-a-trisls-where-end 
                                (is-there-this-son 'EXTESTA sons)))
         local-functions
         
         (default-names (pick-up-all-names-of-a-trisls-default-end 
                         (is-there-this-son 'DEFASTA sons)))
         default-parameters 

         expres-body 
         
         new-function)
    (declare (special function-name flat-parameter-names local-function-names))
    
    ; this function breaks the program if there are collision
    ; between names

    (check-names flat-parameter-names default-names
                 local-function-names function-name)
    

    (setf parameters (eval-params (is-there-this-son 'PARAMS sons)))
    
    (setf default-parameters (eval-defasta (is-there-this-son 'DEFASTA sons)
                                           default-names))
    
    (setf local-functions (eval-extesta (is-there-this-son 'EXTESTA sons)
                                        local-function-names))
    
    (setf expres-body 
          (make-auto-eval-form 
           (eval-expres ; there is always inside
            (is-there-this-son 'EXPRES sons))))

    (if (and (member function-name **TOKEN-FOUND-IN-DEFINITION**)
             (not flat-parameter-names))
      (warning-for-recursive-function function-name))

    (setf new-function
          (make-function function-name parameter-names local-functions 
                         flat-parameter-names local-function-names default-parameters
                         parameters expres-body
                         parameter-names
                         ))
          


    (setf (plasm-function-calling-function new-function)
          (if old-fun ;is there a so named fun ?
            (plasm-function-calling-function old-fun)
            (you-can-set-if-you-want-on-fun function-name)))

    (set-called-function new-function 
                         function-name flat-parameter-names local-function-names)

    (if (not flat-parameter-names)
      (BREAK-IF-IT-CAN-LOOP-FOR-EVER
       (plasm-function-called-function new-function)
       (plasm-function-calling-function new-function)
       function-name))

    (update-calling-functions function-name old-fun new-function)

    (insert-in-user-fenv new-function)

    (reset-environment-definition)  ; switch off definition checking 
    
    function-name))

;------------------------------------------------------------------------------------

(defun MAKE-FUNCTION (function-name parameter-names local-functions 
                      flat-parameter-names local-function-names default-parameters
                      parameters expres-body 
                      plasm-used-parameters) 
  (let ((formal-pars-symbols (use-radix-and-make-indexed-symbols 
                              'arg
                              (length parameter-names) ; number of grouped parameters
                              )))
    (make-plasm-function 
     :name function-name
     :formal-pars formal-pars-symbols
     :used-formal-pars plasm-used-parameters
     :body `(progn
                            
              ,@(apply #'append 
                       (mapcar #'param-in-let parameter-names formal-pars-symbols))
              
              (let
                (                 
                 ,@(apply #'append (mapcar #'make-pairs-list-for-selecting 
                                           parameter-names 
                                           formal-pars-symbols))

                 ,@(make-let-for-local-function local-functions)
                 
                 )
                ,@(declare-special (append flat-parameter-names 
                                           local-function-names))
                ,@(set-all-default default-parameters)
                
                ,@(apply #'append (mapcar #'total-check-type-forms parameters))
                
                ,expres-body))
     :where-name local-function-names)))



;------------------------------------------------------------------------


(defun USE-RADIX-AND-MAKE-INDEXED-SYMBOLS (radix num)
  (let (name-list)
    (dotimes (i num (reverse name-list))
      (setq name-list (cons
                       (intern (format nil "~A~A" radix i) 'plasm)
                       name-list)))))
                              

;------------------------------------------------------------------------

(defun UPDATE-CALLING-FUNCTIONS (function-name old-fun new-function)
  
  (if old-fun
    
    (progn
      (REMOVE-F-FROM-CALLING-FUNCTION-SLOT-OF-ALL-NAMES
       function-name
       (GIVE-LIST-OF-DEAD-CALLED-FUNCTION
        (plasm-function-called-function old-fun)
        (plasm-function-called-function new-function)))
      
      (INSERT-F-IN-CALLING-FUNCTIONS-SLOT-OF-ALL-NAMES
       function-name
       (give-list-of-born-called-function
        (plasm-function-called-function old-fun)
        (plasm-function-called-function new-function))))
    
    (INSERT-F-IN-CALLING-FUNCTIONS-SLOT-OF-ALL-NAMES
     function-name 
     (plasm-function-called-function new-function))
    
    )
  
  ;questa forma e' eseguita solo se stiamo
  ;interpretando una ridefizione di function-name: infatti
  ;solo in questo caso puo' essere gia' stato calcolato un
  ;qualche valore utilizzando function-name!!!
  
  (if old-fun    
    (RESET-VALUE-OF-FUNs-AND-OF-ALL-their-CALLING-FUNCTIONS 
     (plasm-function-calling-function new-function))))



;--------------------------------------------------------------------------------


(defun SET-CALLED-FUNCTION (new-function function-name parameter-names 
                            ;specifier-names
                                         local-function-names)
  (declare (special **TOKEN-FOUND-IN-DEFINITION**))
  (setf (plasm-function-called-function new-function)
        (set-difference 
         **TOKEN-FOUND-IN-DEFINITION**
         (append (list function-name)
                 parameter-names
                 ;specifier-names
                 local-function-names))))


;-------------------------------------------------------------------------------
;Se esistono delle collisioni fra nomi di parametri specificatori o funzioni locali,
;oppure esistono dichiarazioni di default per parametri non dichiarati, questa funzione
; interrompe l'interpretazione con un messaggio sul tipo di inconveniente incontrato.


(defun CHECK-NAMES (parameters-list default-list where-list function-name)
  (cond
   ((duplicates? parameters-list)
    (plasm-error 101 parameters-list)) ;more parameters with the same name
      
   ((duplicates? default-list)
    (plasm-error 103 default-list)) ;more ''default'' with the same name
   
   ((not (multiple-member default-list parameters-list))
    (plasm-error 106 default-list parameters-list))      
    ;default name(s) is (are) not matching with params
  
   
   ((duplicates? where-list)
    (plasm-error 104 where-list))
   
   ((duplicates? (append parameters-list 
                         where-list
                         (list function-name)))
    (plasm-error 105 parameters-list 
                     where-list
                     (list function-name))) ;more ''where'' with the same name
    ( T )))

;------------------------------------
;Restituisce T se esistono in list- due simboli uguali


(defun DUPLICATES? (list-) 
   (cond
    ((null list-) nil)
    ((= 1 (length list-)) nil)
    ((member (first list-) (rest list-)) T)
    (T (duplicates? (rest list-)))))



(defun MULTIPLE-MEMBER (list1 list2)
  (null (set-difference list1 list2)))


(defun DECLARE-SPECIAL (name-list)
  (if name-list
    `((declare (special ,@name-list)))))
   


(defun TOTAL-CHECK-TYPE-FORMS (list-)
  ;esempio di list- : (((seq) isseq) ((i j) isintpos))
  (if list-
    (apply #'append
           (mapcar #'multiple-type-form list-))))



(defun MULTIPLE-TYPE-FORM (pair)
  (let ((form (check-type-form (first pair) (second pair))))
    (if form
      (list form)
      nil)))







(defun CHECK-TYPE-FORM (var type)
  ; (not (and a b)) == (or (not a) (not b))
  (declare (special function-name))
  (if (equal type ''tt)
    nil
  (let ((type-form (if (symbolp type) `',type type)))
    `(type-test-function ',function-name ,type-form (list  ,@var)))))


(defun TYPE-TEST-FUNCTION (function-name type-test-fun value-list)
  (cond 
   ((not (bound-function-p type-test-fun))
    (plasm-error 601 function-name))
   (T (if (not (every #'(lambda (var)
                          (reduce-function type-test-fun var))
                         value-list))
        (plasm-error 602 function-name value-list)))))



(defun SET-ALL-DEFAULT (list-defaults)
  (if list-defaults
    (mapcar #'set-default list-defaults)))


(defun SET-DEFAULT (name-value)
  (let* ((name (first name-value))
         (value (second name-value)))
    `(set-if-default ,name ,value)))



(defmacro SET-IF-DEFAULT (name value-form)
  `(if (eq 'df ,name)
     (setf ,name  ,value-form)))




(defun CHECK-LENGTH (arg len fun-name)
  (if (and (consp arg) (/= len (length arg)))
    (plasm-error 202 arg len fun-name)))







(defun PARAM-IN-LET (param-names on-symbol)
  (declare (special function-name))
  (if (second param-names) ; param-names is a list with a length greater than 1 ?
    `((check-length ,on-symbol ,(length param-names) ',function-name))))





(defun MAKE-PAIRS-LIST-FOR-SELECTING (list-names on-symbol)
  (if (= 1 (length list-names))
    `((,(first list-names) ,on-symbol))
    (let (form)
      (dotimes (i (length list-names) form)
        (setf form
              (append form
                      (list
                       `(,(nth i list-names) (nth ,i ,on-symbol)))))))))

;------------------------------------------------------------------------------


(defun REACHABLE-NODES (to-visit)
  (let (just-visited
        node)
    (loop 
      (setq node (first to-visit))
      (if node
        (progn
          (setq to-visit (cdr to-visit))
          (if (member node just-visited)
            nil
            (progn
              (setq just-visited (cons node just-visited))
              (setq to-visit (insert (take-sons node) to-visit)))))
        (return just-visited)))))



(defun INSERT (set-a set-b) (union set-a set-b))

(defun TAKE-SONS (node)
  (declare (special GRAPH))
  (second (assoc node GRAPH)))



(defun LOCAL-CHECK (function-name graph)
  (declare (special graph))
  (mapcar
   #'(lambda (local-name-called)
       (let ((name (first local-name-called))
             (called (second local-name-called)))
         (if (or (member name called)
                 (member name (reachable-nodes called)))
           (break-definition-for-local function-name name))))
   graph))




(defun MAKE-LET-FOR-LOCAL-FUNCTION (local-functions)
  (declare (special function-name))
  (local-check 
   function-name
   (mapcar #'(lambda (pl-fun)
               (list (first pl-fun) (inside-locals (second pl-fun))))
           local-functions))
  (mapcar #'(lambda (pl-fun) 
              `(,(first pl-fun) 
                ,(if (and (consp (second pl-fun))
                         (eq 'quote (first (second pl-fun))))
                  `,(second pl-fun)
                  `',(second pl-fun))))
          local-functions))





(defun INSIDE-LOCALS (arg)
  (declare (special local-function-names))
  (cond
   ((atom arg) (if (member arg local-function-names)
                 (list arg)))
   (T (apply #'append (mapcar #'inside-locals arg)))))




;------------------------------------------------------------------------------

;  PARASTA -> ( LISTDEC )

(defun EVAL-PARASTA (tree)
  (if tree
    (eval-listdec (pick-up-son-num 2 tree))))


;------------------------------------------------------------------------------
;    LISTDEC -> ADECLAR ; LISTDEC
;    LISTDEC -> ADECLAR

(defun EVAL-LISTDEC (tree)
  (let ((sons (pick-up-sons tree)))
    (if (= 1 (length sons))
      ;we've got  LISTDEC ->  ADECLAR
      (eval-adeclar (first sons))
      ;we've got   LISTDEC -> ADECLAR ; LISTDEC
      (append (eval-adeclar (first sons))
              (eval-listdec (third sons))))))


;------------------------------------------------------------------------------
;     ADECLAR -> TOKENS :: TYPE

(defun EVAL-ADECLAR (tree)
  (list (list (eval-tokens (pick-up-son-num 1 tree))
              (eval-type (pick-up-son-num 4 tree)))))


;------------------------------------------------------------------------------

; TYPE -> EXPRES 



(defun EVAL-TYPE (tree)
  (let* ((son (pick-up-son-num 1 tree))
         (type-form (eval-expres son)))
    (make-auto-eval-form type-form)))


;--------------------------------------------------------------------------
;  TOKENS -> TOKENID , TOKENS
;  TOKENS -> TOKENID

(defun EVAL-TOKENS (tree)
  (let ((sons (pick-up-sons tree)))
    (if (= 1 (length sons))
      ;we've got TOKENS -> TOKENID
      (list (eval-tokenid (first sons)))
      ;we've got TOKENS -> TOKENID , TOKENS
      (cons (eval-tokenid (first sons))
            (eval-tokens (third sons))))))




;---------------------------------------------------------------------------
;   DEFASTA -> #\1 TRISLS #\2
;
;Nota che #\1 e #\2 sono le sostituzioni fatte da LEX per le parole chiave
;  DEFAULT e END !!!!!!


(defun EVAL-DEFASTA (tree default-names)
  (if tree
    (eval-trisls-default-end (pick-up-son-num 2 tree) default-names)))


;---------------------------------------------------------------------------
;  TRISLS -> TOKENID = EXPRES
;  TRISLS -> TOKENID = EXPRES , TRISLS


(defun EVAL-TRISLS-DEFAULT-END (tree default-names)
  (let* ((names default-names)
         (num-names (length names))
         (trisls-tree tree)
         bodies)
    (dotimes (i num-names)
     (setf bodies (append bodies
                          (list
                           (let ((body-form (eval-expres
                                             (pick-up-son-num 3 trisls-tree))))
                             (if (atomic-p body-form)
                               body-form
                               (make-auto-eval-form body-form))))))
     (if (< (1+ i) num-names)
       (setf trisls-tree
             ; TRISLS -> TOKENID = EXPRES , TRISLS
             (pick-up-son-num 5 trisls-tree))))
    ;At this point we have got a list of default names and a matching
    ;list of their value.
    (mapcar #'list names bodies)))


;--------------------------------------------------------------------------

(defun PICK-UP-ALL-NAMES-OF-A-TRISLS-WHERE-END (where-end-tree)
  (if where-end-tree
    (PICK-UP-ALL-NAMES-OF-A-TRISLS-WHERE-END1 (pick-up-son-num 2 where-end-tree))))

(defun PICK-UP-ALL-NAMES-OF-A-TRISLS-WHERE-END1 (tree)
  (let ((sons (pick-up-sons tree)))
    (if (= 3 (length sons))
      ;we've got TRISLS -> TOKENID = EXPRES
      (list (eval-tokenid (first sons)))
      ;we've got TRISLS -> TOKENID = EXPRES , TRISLS
      (cons (eval-tokenid (first sons))
            (pick-up-all-names-of-a-trisls-where-end1 (fifth sons))))))


(defun PICK-UP-ALL-NAMES-OF-A-TRISLS-DEFAULT-END (tree)
  (PICK-UP-ALL-NAMES-OF-A-TRISLS-WHERE-END tree))



;--------------------------------------------------------------------------
;    EXSTESTA -> #\3 TRISLS #\2

(defun EVAL-EXTESTA (tree local-function-names)
  (if tree
    (eval-trisls-where-end (pick-up-son-num 2 tree) local-function-names)))


;--------------------------------------------------------------------------
;     DEFLIST -> DEFIN
;     DEFLIST -> DEFIN DEFLIST
;


(defun EVAL-TRISLS-WHERE-END (tree local-function-names)
  (let* ((num-names (length local-function-names))
         (trisls-tree tree)
         bodies)
    (dotimes (i num-names)
     (setf bodies (append bodies
                          (list 
                           (make-auto-eval-form 
                            (eval-expres
                             (pick-up-son-num 3 trisls-tree))))))
     (if (< (1+ i) num-names)
       (setf trisls-tree
             ; TRISLS -> TOKENID = EXPRES , TRISLS
             (pick-up-son-num 5 trisls-tree))))
    ;At this point we have got a list of local function names and a matching
    ;list of their bodies.

    (mapcar #'list local-function-names bodies)))




;---------------------------------------------------------------------------
;Una definizione, da qualunque regola e' stata generata, ha sempre come nome
;l'interpretazioe del suo secondo figlio (vedi grammatica!). La categoria
;sintattica di questo figlio e' TOKENID.

(defun PICK-UP-NAME-OF-DEF (tree)
  (eval-tokenid (pick-up-son-num 2 tree)))


;--------------------------------------------------------------------------------
; sons e' una lista di alberi sintattici. restituisce il primo albero sintattico
; di questa lista il cui nodo padre e' son-type se questo esiste, nil altrimenti.

(defun IS-THERE-THIS-SON (son-type sons)
  (let ((sons- sons) ; sons is a list of syntax tree: one of this can be a leaf
                     ; too. A leaf is a terminal, that is a character.
        son)
    (loop
      (if (null sons-) (return))
      (setf son (first sons-))
      (setf sons- (rest sons-))
      (if (not (characterp son))
          (if (member son-type (first son)) ; a syntax tree is a pair list:
                                            ; the root and the sons. The root is
                                            ; a pair: the rule number and the 
                                            ; non-terminal name.
            (return son))))))


;--------------------------------------------------------------------------------
; It returns the the root of a syntax tree.

(defun GIVE-ROOT (tree)
  (first tree))


;--------------------------------------------------------------------------------

;  PARAMS -> PARASTA PARAMS

;  PARAMS -> PARASTA


(defun EVAL-PARAMS (tree)
  (if tree (eval-params1 tree)))

(defun EVAL-PARAMS1 (tree)
  (if (= 1 (length (pick-up-sons tree))) 

    ; we've got the rule  PARAMS -> PARASTA
    (list (eval-parasta (pick-up-son-num 1 tree)))

    ; we've got the rule  PARAMS -> PARASTA PARAMS
    (cons (eval-parasta (pick-up-son-num 1 tree))
          (eval-params1 (pick-up-son-num 2 tree)))))

;--------------------------------------------------------------------------------


(defun PICK-UP-PARAMS-NAMES (tree)
  (if tree (pick-up-params-names1 tree)))

(defun PICK-UP-PARAMS-NAMES1 (tree)
  (if (= 1 (length (pick-up-sons tree))) 

    ; we've got the rule  PARAMS -> PARASTA
    (list (take-parameter-names-from-listdec 
       (pick-up-son-num 2 (pick-up-son-num 1 tree))))

    ; we've the rules
    ; PARAMS -> PARASTA
    ;     and
    ; PARASTA -> #\( LISTDEC #\)
    (cons (take-parameter-names-from-listdec 
             (pick-up-son-num 2 (pick-up-son-num 1 tree)))
            (pick-up-params-names1 (pick-up-son-num 2 tree)))))

;--------------------------------------------------------------------------------

(defun PICK-UP-PARAMETERS-NAMES (tree)
  (cond
   ((not tree) nil)
   (t (take-parameter-names-from-listdec 
       (pick-up-son-num 2 tree)))))



;--------------------------------------------------------------------
(defun TAKE-PARAMETER-NAMES-FROM-LISTDEC (tree)
  (if (= 1 (length (pick-up-sons tree)))

    ; produzione --->  (LISTDEC ( ADECLAR))
    (take-parameter-names-from-adeclar (pick-up-son-num 1 tree))

    ; produzione --->  (LISTDEC ( ADECLAR #\; LISTDEC))
    (append 
     (take-parameter-names-from-adeclar (pick-up-son-num 1 tree))
     (take-parameter-names-from-listdec (pick-up-son-num 3 tree)))))


(defun TAKE-PARAMETER-NAMES-FROM-ADECLAR (tree)
  ; produzione --->  (ADECLAR ( TOKENS #\: #\: TYPE))
  (eval-tokens (pick-up-son-num 1 tree)))

;--------------------------------------------------------------------------


(defun PICK-UP-DEFAULT-NAMES (pairs-list)
  (cond
   ((not pairs-list) nil)
   (t (mapcar #'first pairs-list))))


;------------------------------------------------------------------------------
;It calls a plasm error if name is member of **ALL-NO-GOOD-NAMES** or is the
;name of a primitive function.

(setf **ALL-NO-GOOD-NAMES** '( df
                              ))    ; Put here all names that cannot be valid
                                    ; for an user function (other than primitive
                                    ; function name).
 
(defun BREAK-IF-NO-GOOD-FUNCTION-NAME (tree)
  (declare (special **DEFAULT-FENV** **ALL-NO-GOOD-NAMES**))
  (let ((name (take-leaf-tokenid-tree tree)))
    (if (or 
         (eq nil name)
         (eq t name)
         (not (symbolp name))
         (gethash name **DEFAULT-FENV**)
         (member name **ALL-NO-GOOD-NAMES**)
         (let ((stringa (symbol-name name)))
           (or (is-sel stringa)
               (is-skeleton stringa))))
      (plasm-error 301 name)
      name)))


;------------------------------------------------------------------------------
;If there is a function of 0-arity which calling-function and called-function 
;set are not with an empty intersection, then a never ending loop can begin.
;If it's the case, a plasm error must be called.


(defun WARNING-IF-IT-CAN-LOOP-FOR-EVER (called calling funname)
  (if (eval 
       (cons 'or 
             (mapcar #'function-without-par-p (intersection called calling))))
    (format t "~%~%WARNING: In the defining function ~A appear as called function ~A ~
                 and as calling functions ~A . There could be a never ending loop !!!!"
            funname  called calling)))



(defun WARNING-FOR-RECURSIVE-FUNCTION (function-name)
  (format t "~%~%WARNING~%~%~
             The definition of ~A is correct only if ~
             it is a recursive function." function-name))




(defun BREAK-IF-IT-CAN-LOOP-FOR-EVER (called calling funname)
  (let* ((inter (intersection called calling))
         (test (eval (cons 'or (mapcar #'function-without-par-p inter)))))
  (if test
    (plasm-error 
     0
     (format 
      t 
      "~%~%The value named ~A needs value[s] of ~A ~
       and it(they) needs(need) value of ~A!.~%~%~
      If values are functions please use definition with arguemnts."
      funname 
      (let ((stringa (format nil "~A" inter)))
        (subseq stringa 1 (1- (length stringa))))
      funname)))))




(defun BREAK-DEFINITION-FOR-LOCAL (function-name name)
  (plasm-error 0
               (format 
                t 
                "~%~%DEFINITION ERROR:~%~
                 The local object ~A given ~
                 inside the function ~A is defined in terms of itself.~%~%"
                name
                function-name)))


;-------------------------------------------------------------------------
