; -*- package: plasm -*-
;==========================================================================
;
;                           struct.LSP File
;
;==========================================================================


(provide 'struct)

(inpackage plasm)


;==================================================================
; A plasm function is a lisp structure 

(defstruct plasm-function
  name
  function-description
  formal-pars
  used-formal-pars
  body
  where-name
  calling-function
  called-function
  value 
  pol-value
  motion-labels
  file)


;==================================================================
;
; All function working on **TEMP-FENV** hash table are not more used.
; They should be cancelled. Nevertheless they could be called without
; side effects by old branches of code. They will be removed as soon as
; possible.




;--------------------------------------------------------------------------
;BOUND-FUNCTION function returns the binding of his argument in fenv, his 
;second argument. If there is not any binding in fenv it will search in
;**USER-FENV** and then in **DEFAULT-FENV**. If no binding is found an
;error will occur. 


(defun BOUND-FUNCTION (fname &key (call-error T))
    (BOUND-FUNCTION-IN-USER-AND-DEFAULT-FENV fname :call-error call-error))



(defun BOUND-FUNCTION-P (fname)
  (cond ((functionp fname) fname)
        ((symbolp fname)
          (funcall #'bound-function fname :call-error nil))))




;------------------------------------------------------------------------
;This function checks if a symbol is a function name actually present in
; **USER-FENV** and **DEFAULT-FENV**.




(defun BOUND-FUNCTION-IN-USER-FENV (fname &key (call-error T))
  (declare (special **USER-FENV**))
  (cond
   ((gethash fname **USER-FENV**))
   (call-error (plasm-error 1 fname))))

(defun BOUND-FUNCTION-IN-USER-AND-DEFAULT-FENV (fname &key (call-error T))
  (declare (special **USER-FENV** **DEFAULT-FENV**))
  (cond
   ((or (gethash fname **USER-FENV**)
        (gethash fname **DEFAULT-FENV**)))
   (call-error (plasm-error 1 fname))))


(defun BOUND-FUNCTION-IN-USER-AND-DEFAULT-FENV-P (fname)
  (if (symbolp fname)
    (funcall #'bound-function-in-user-and-default-fenv fname :call-error nil)))





(defun PICK-UP-PARAM (fenv-element)
  (plasm-function-formal-pars fenv-element))


(defun PICK-UP-BODY (fenv-element)
  (plasm-function-body fenv-element))


(defun FUNCTION-WITHOUT-PAR-P (name)
  (if (bound-function-p  name)
    (null (pick-up-param (bound-function name)))))

;--------------------------------------------------------------------------



(defun INSERT-IN-USER-FENV (new-user-function)
  (declare (special **USER-FENV**))
  (setf (gethash (plasm-function-name new-user-function) **USER-FENV**)
        new-user-function))



(defun REMOVE-FROM-USER-FENV (fname) 
  (declare (special **USER-FENV**))
  (let ((fenv-element (gethash fname **USER-FENV**)))
    (if fenv-element
        (remhash fname **USER-FENV**))))
        



(defun INSERT-IN-DEFAULT-FENV (new-user-function)
  (declare (special **DEFAULT-FENV**))
  (setf (gethash (plasm-function-name new-user-function) **DEFAULT-FENV**)
        new-user-function))



        


(defun SET-AND-OR-RETURN-VALUE (funname &optional (set? nil))
  (let ((fun-structure (bound-function funname)))
    (cond
     ((plasm-function-formal-pars fun-structure)
      (cond
       (set? (cond
             ((plasm-function-value fun-structure))
             ((set-anonymous-function-value fun-structure))))
       (funname)))
     ((plasm-function-value fun-structure)) ; a value was just calculated 
     
     (T 
      (setf (plasm-function-value fun-structure)
            (eval (plasm-function-body fun-structure)))))))




;************************************************************************

;This functions cut lattice of model when user is changing a definition


(defun RESET-VALUE-OF-FUNS-AND-OF-ALL-THEIR-CALLING-FUNCTIONS (funs)
  (let ((not-yet-visited funs)
        (just-visited nil)
        (new-to-visit nil))
    (loop
      (if (null not-yet-visited) (return))
      (setf just-visited (cons (first not-yet-visited)
                               just-visited))
      (setf new-to-visit (reset-values-of-fun-and-return-its-calling-function
                          (first not-yet-visited)))
      (setf not-yet-visited (rest not-yet-visited))
      (setf not-yet-visited
            (append not-yet-visited
                    (set-difference
                     new-to-visit
                     (append just-visited not-yet-visited)))))))


(defun RESET-VALUES-OF-FUN-AND-RETURN-ITS-CALLING-FUNCTION (fun)
  (let ((funstructure (bound-function fun)))
    (cond
     ((not (plasm-function-formal-pars funstructure))
      (setf (plasm-function-value funstructure) nil)
      (setf (plasm-function-pol-value funstructure) nil)
      (setf (plasm-function-file funstructure) nil)))
    (plasm-function-calling-function  funstructure)))





;+++++++++++ 
; You must notice that RESET-VALUE-OF-FUNS can't loop for ever. We always check
;collisions of symbols for every function while updating either Called-Function 
; or Calling-Function slot.
;+++++++++++



;---------------------------------------------------------------
;pone value come valore della slot FILE della funzione PLASM fun-name

(defun SET-AS-FILE-SLOT-VALUE-OF-FUNCTION (fun-name value)
  (setf (plasm-function-file (bound-function fun-name)) value))




;------------------------------------------------------------------------------
;It returns symbols in the new-list argument that it does not find in old-list

(defun GIVE-LIST-OF-BORN-CALLED-FUNCTION (oldlist newlist)
  (set-difference newlist oldlist))

;------------------------------------------------------------------------------
;It returns symbols in the oldlist argument no more in newlist argument

(defun GIVE-LIST-OF-DEAD-CALLED-FUNCTION (oldlist newlist)
  (set-difference oldlist newlist))


;------------------------------------------------------------------------------
;When the user redefine a function f some names can be no more live in the body.
;Then we must cancel from Calling-Function slot of died name the symbol f.

(defun REMOVE-F-FROM-CALLING-FUNCTION-SLOT-OF-ALL-NAMES (f all-names)
  (mapcar #'(lambda (g) (f-is-not-more-calling-function-of-g f g))
          all-names))


(defun F-IS-NOT-MORE-CALLING-FUNCTION-OF-G (f g)
  (let ((funstructure (bound-function g :call-error nil)))
    (if funstructure
      (setf (plasm-function-calling-function funstructure)
            (delete f (plasm-function-calling-function funstructure)))
      (f-is-not-more-calling-function-of-g-not-yet-defined f g))))


(defun F-IS-NOT-MORE-CALLING-FUNCTION-OF-G-NOT-YET-DEFINED (f g)
  (declare (special *SET-WHEN-YOU-CAN*))
  (let ((pairlist (gethash g *SET-WHEN-YOU-CAN*)))
    (setf (second pairlist) (delete f (second pairlist)))
    (if (not (second pairlist))
      (setf (gethash g *SET-WHEN-YOU-CAN*) nil))))

;------------------------------------------------------------------------------
;When the user define or redefine a function f the name f must appear in the 
;Calling-Functions slot of the names not local in the body of f definition.

(defun INSERT-F-IN-CALLING-FUNCTIONS-SLOT-OF-ALL-NAMES (f all-names)
  (mapcar #'(lambda (g) (f-is-calling-function-of-g f g))
          all-names))


(defun F-IS-CALLING-FUNCTION-OF-G (f g)
  (let ((funstructure (bound-function g :call-error nil)))
    (if funstructure
      (if (member f (plasm-function-calling-function funstructure))
        nil
        (setf (plasm-function-calling-function funstructure)
              (cons f (plasm-function-calling-function funstructure))))
      (wait-for-g-definition g f))))


(setf *SET-WHEN-YOU-CAN* (make-hash-table))


(defun RESET-*SET-WHEN-YOU-CAN* ()
  (declare (special *SET-WHEN-YOU-CAN*))
  (clrhash *SET-WHEN-YOU-CAN*))



(defun WAIT-FOR-G-DEFINITION (called-fun calling-fun)
  (declare (special *SET-WHEN-YOU-CAN*))
  (let ((exist (gethash called-fun *SET-WHEN-YOU-CAN*)))
    (if exist
      (setf (caadr exist) calling-fun)
      (setf (gethash called-fun *SET-WHEN-YOU-CAN*) 
            (list called-fun (list calling-fun))))))


(defun YOU-CAN-SET-IF-YOU-WANT-ON-FUN (fun)
  (declare (special *SET-WHEN-YOU-CAN*))
  (let ((suspend-calling (gethash fun *SET-WHEN-YOU-CAN*)))
    (if suspend-calling
      (remhash fun *SET-WHEN-YOU-CAN*))
    (second suspend-calling)))
