;;; -*- package: boole -*-
;-------------------------------------------------------------------------------
;           Universita' di Roma "La Sapienza"
;       Dipartimento di Informatica e Sistemistica
;            Via Buonarroti 12 - 00185 Roma
;
;                      CAD Group                 
; 
; Project: SimpleXn
; Identification: Package Boole                           
; Date: 20-12-90
; Note: Polyhedra Environment, Normalization, Pruning, Selection
;       Optional convex-intersection test in set-s-form
;       Optional reconstruction of adjacencies in set-W-rep     
;       This package can use Pascal triangulation routines
;       P-cell-to-s-cell and P-posneg-cell-to-s-list-list modified
;             
; Version: 1.0
;              
; Author: F. Bernardini
;-------------------------------------------------------------------------------

(provide 'boole) 

(defpackage boole)                                         ;module name

(inpackage boole)      ;package name

(export '(boole-help
	  init-pol-table
	  reset-pol-table
	  print-pol-table
	  save-pol-table
	  read-pol-table
	  print-pol
	  print-W-rep
	  print-B-expr
	  print-P-form
	  print-s-form
	  print-box
	  pol
	  add-pol
	  rem-pol
	  set-P-form
	  set-s-form
	  set-W-rep
	  set-W-rep-ad))
			
(require 'kernel)
(require 'io)
(require 'triang)           ;triangulation or triangulation-p
(require 'dvrlib)

(use-package 'kernel)
(use-package 'io)
(use-package 'triangulation)
(use-package 'dvrlib)

;-------------------------------------------------------------------------------
; The pol-table data structure
;-------------------------------------------------------------------------------

(defstruct pol-rec
  (W-rep nil)
  (B-expr '())
  (P-form '())
  (s-form '())
  (shortname nil)
  (sign '+)
  (box nil))


(defvar **pol-table '()
  "associative list: each item is a  pol-cons, 
that is the pair pol-name.pol-rec")


;-------------------------------------------------------------------------------
; Functions to manipulate the pol-table
;-------------------------------------------------------------------------------

(defun exist-pol-cons? (pol-name)
  (if (find-pol-cons pol-name) t nil))


(defun find-pol-cons (pol-name)
  (assoc pol-name **pol-table))


(defun put-pol-cons (pol-name pol-rec)
  (setq **pol-table (acons pol-name pol-rec **pol-table)))


(defun get-pol-name (pol-cons)
  (car pol-cons))


(defun get-pol-rec (pol-cons)
  (cdr pol-cons))


(defun find-W-rep (pol-name)
  (pol-rec-W-rep (get-pol-rec (find-pol-cons pol-name))))


(defun find-B-expr (pol-name)
  (pol-rec-B-expr (get-pol-rec (find-pol-cons pol-name))))


(defun find-P-form (pol-name)
  (pol-rec-P-form (get-pol-rec (find-pol-cons pol-name))))


(defun find-s-form (pol-name)
  (pol-rec-s-form (get-pol-rec (find-pol-cons pol-name))))


(defun print-pol-rec1 (pol-rec)
    (format t "B-Expr : ~A~%" (pol-rec-B-expr pol-rec))
    (format t "P-Form : ~A~%" (if (pol-rec-P-form pol-rec) t nil))
    (format t "s-Form : ~A~%" (if (pol-rec-s-form pol-rec) t nil))
    (format t "W-rep  : ~A~%" (if (pol-rec-W-rep pol-rec) t nil))
    (format t "Sign   : ~A~%" (pol-rec-sign pol-rec))
    t)


(defun print-P-form1 (P-form)
  (format t "~%")
  (dolist (P-cell P-form)
    (dolist (P-term (first P-cell))
      (format t "~A " (string-downcase P-term)))
    (dolist (P-term (second P-cell))
      (format t "~A " (string-upcase P-term)))
    (format t "~%"))
  t)


(defun print-s-form1 (s-form)
  (format t "~%")
  (dolist (s-cell s-form)
    (dolist (s-term (first s-cell))
      (format t "~A~A " 
	      (string-downcase (first s-term)) 
	      (subseq (string (cdr s-term)) 1)))
    (dolist (s-term (second s-cell))
      (format t "~A~A " 
	      (string-upcase (first s-term)) 
	      (subseq (string (cdr s-term)) 1)))
    (format t "~%"))
  t)


(defun add-W-rep (name pol)
  "Add a pol-cons to the table, with W-rep = pol"
  (if (not (polyhedron? pol))
    (errmsg 3)
    (progn (put-pol-cons name (make-pol-rec     
			  :W-rep pol
			  :B-expr name
			  :shortname (gen-shortname)
			  :sign (get-pol-sign pol)
			  :box (eval-pol-box pol)))
	   t)))


(defun add-B-expr (name expr)
  "Add a pol-cons to the table, with B-expr = expr"
  (if (not (valid-expr? expr)) 
    (errmsg 4)                                              
    (progn (put-pol-cons name (make-pol-rec                             
		   :B-expr expr
		   :shortname (gen-shortname)))
	   t)))


(defun valid-expr? (expr)
  (cond ((symbolp expr) (exist-pol-cons? expr))
	((listp expr) 
	 (case (first expr)
	   (/ (and (= (length expr) 2) 
		   (valid-expr? (second expr))))
	   ((+ - *) (and (> (length expr) 2)
			 (every #'valid-expr? (cdr expr))))))
	(t nil)))


;-------------------------------------------------------------------------------
; User interface functions
;-------------------------------------------------------------------------------

; Note: the following functions return t in no error occured
; during their evaluation, otherwise they return nil


(defun init-pol-table ()
  "Initialize the polyhedron table"
  (setq **pol-table '())
  t)


(defun reset-pol-table ()
  "Reset the polyhedron table"
  (dolist (pol-cons **pol-table)
    (setf (pol-rec-P-form (get-pol-rec pol-cons)) nil)
    (setf (pol-rec-s-form (get-pol-rec pol-cons)) nil)
    (unless (symbolp (pol-rec-B-expr (get-pol-rec pol-cons)))
      (setf (pol-rec-W-rep (get-pol-rec pol-cons)) nil)))
  t)


(defun print-pol-table ()
  "Print the polyhedron table"
  (dolist (pol-cons **pol-table)
    (format t "~%Name   : ~A~%" (get-pol-name pol-cons))
    (print-pol-rec1 (get-pol-rec pol-cons))
    (format t "~%"))
  t)


(defun save-pol-table ()
  (let ((filename (prompt-for-a-new-file)))
    (with-open-file (outfile filename
			     :direction :output
			     :if-exists :overwrite
			     :if-does-not-exist :create)
      (prin1 **pol-table outfile))
      t))
  

(defun read-pol-table ()
  (let ((filename (prompt-for-an-existing-file)))
    (with-open-file (infile filename
			    :direction :input)
      (setf **pol-table (read infile)))
    t))


(defmacro print-pol (name)
  "Print the pol-rec named name"
  `(let ((pol-cons (find-pol-cons ',name)))
     (if pol-cons
       (print-pol-rec1 (get-pol-rec pol-cons))
       (errmsg 1))))


(defmacro print-W-rep (name)
  "Print the polyhedron named name"
  `(let ((pol-cons (find-pol-cons ',name)))
    (if pol-cons
      (printpol (pol-rec-W-rep (get-pol-rec pol-cons)))
      (errmsg 1))))


(defmacro print-B-expr (name)
  "Print the boolean expression named name"
  `(let ((pol-cons (find-pol-cons ',name)))
    (if pol-cons
      (format t "~A~%~%" (pol-rec-B-expr (get-pol-rec pol-cons)))
      (errmsg 1))))


(defmacro print-P-form (name)
  "Print the P-form named name"
  `(let ((pol-cons (find-pol-cons ',name)))
    (if pol-cons
      (print-P-form1 (pol-rec-P-form (get-pol-rec pol-cons)))
      (errmsg 1))))


(defmacro print-s-form (name)
  "Print the s-form named name"
  `(let ((pol-cons (find-pol-cons ',name)))
    (if pol-cons
      (print-s-form1 (pol-rec-s-form (get-pol-rec pol-cons)))
      (errmsg 1))))


(defmacro print-box (name)
  "Print the bounding box of pol named name"
  `(let ((pol-cons (find-pol-cons ',name)))
    (if pol-cons
      (format t "~A~%~%" (pol-rec-box (get-pol-rec pol-cons)))
      (errmsg 1))))


(defmacro pol (name)
  "Return the W-rep of a pol-cons named name"
  `(find-W-rep ',name))


(defmacro add-pol (name obj)
  "Add a pol-cons to the table (obj can be a W-rep or a B-expr)"
  `(if (exist-pol-cons? ',name)
     (errmsg 2)
     (let ((temp ,obj))
       (if (listp temp) 
	 (add-B-expr ',name temp)
	 (add-W-rep ',name temp)))))


(defmacro rem-pol (name)
  "Remove a pol-cons from the table"
  `(if (exist-pol-cons? ',name)
     (progn 
       (setq **pol-table (remove (find-pol-cons ',name) **pol-table))
       t)
     (errmsg 1)))


(defmacro set-P-form (name)
  "Evaluate the B-expr named name and set 
P-form and sign fields; during the evaluation set the 
P-form e sign fields of every sub-expression evaluated"
  `(let ((pol-cons (find-pol-cons ',name)))
     (if pol-cons
       (let ((pol-rec (get-pol-rec pol-cons)))
	 (unless (pol-rec-P-form pol-rec) 
	   (multiple-value-bind 
	     (sign P-form) (B-expr-to-P-form (pol-rec-B-expr pol-rec))
	     (setf (pol-rec-P-form pol-rec) P-form)
	     (setf (pol-rec-sign pol-rec) sign)))
	 t)
       (errmsg 1))))


(defmacro set-s-form (name &optional (cit-flag t))
  "Transform a P-form in a s-form. If P-form
is nil, then it is evaluated. If the optional
parameter is nil, then the convex-intersection test
is not performed."
  `(let ((pol-cons (find-pol-cons ',name)))
     (if pol-cons
       (let ((pol-rec (get-pol-rec pol-cons)))
	 (unless (pol-rec-P-form pol-rec) (set-P-form ,name))
	 (unless (pol-rec-s-form pol-rec)
	   (setf (pol-rec-s-form pol-rec) 
		 (P-form-to-s-form (pol-rec-P-form pol-rec) ,cit-flag)))
	 t)
       (errmsg 1))))


(defmacro set-W-rep (name &optional (build-ad-flag t))
  "Transform a s-form in a W-rep. If s-form and/or 
P-form are nil, then they are evaluated. If the optional
parameter is nil, then w-rep adjacencies are not calculated.
Nota Bene: only intersection currently supported."
  `(let ((pol-cons (find-pol-cons ',name)))
     (if pol-cons
       (let ((pol-rec (get-pol-rec pol-cons)))
	 (unless (pol-rec-W-rep pol-rec)
	   (unless (pol-rec-s-form pol-rec) (set-s-form ,name))
	   (setf (pol-rec-W-rep pol-rec) 
		 (s-form-to-W-rep (pol-rec-s-form pol-rec) ,build-ad-flag)))
	 t)
       (errmsg 1))))


(defmacro set-W-rep-ad (name)
  "Set w-rep adjacencies"
  `(let ((pol-cons (find-pol-cons ',name)))
     (if pol-cons
       (let ((pol-rec (get-pol-rec pol-cons)))
	 (if (pol-rec-W-rep pol-rec)
	   (progn
	     (setf (pol-rec-W-rep pol-rec)
		   (build-polyhedron (pol-rec-W-rep pol-rec)))
	     t)
	   (errmsg 5)))
       (errmsg 1))))


(defun boole-help ()
  (print "

list of functions available: 

  Note: the following functions return t in no error occured
  during their evaluation, otherwise they return nil


init-pol-table ()
   Initialize the polyhedron table 

reset-pol-table ()
   Reset the polyhedron table 

print-pol-table ()
   Print the polyhedron table 

save-pol-table ()  

read-pol-table ()

print-pol (name)
   Print the pol-rec named name 

print-W-rep (name)
   Print the polyhedron named name 

print-B-expr (name)
   Print the boolean expression named name 

print-P-form (name)
   Print the P-form named name 

print-s-form (name)
   Print the s-form named name 

print-box (name)
   Print the bounding box of pol named name 

pol (name)
   Return the W-rep of a pol-cons named name 

add-pol (name obj)
   Add a pol-cons to the table (obj can be a W-rep or a B-expr) 

rem-pol (name)
   Remove a pol-cons from the table 

set-P-form (name)
   Evaluate the B-expr named name and set 

set-s-form (name &optional (cit-flag t))
   Transform a P-form in a s-form. If P-form
is nil, then it is evaluated. If the optional
parameter is nil, then the convex-intersection test
is not performed. 

et-W-rep (name &optional (build-ad-flag t))
   Transform a s-form in a W-rep. If s-form and/or 
P-form are nil, then they are evaluated. If the optional
parameter is nil, then w-rep adjacencies are not calculated.
Nota Bene: only intersection currently supported. 

set-W-rep-ad (name)
   Set w-rep adjacencies 
")
  (values))


;-------------------------------------------------------------------------------
; Shortname generator
;-------------------------------------------------------------------------------

(defvar **next-shortname (string #\a))


(defun init-shortname () (setq **next-shortname (string #\a)))


(defun gen-shortname ()
  "Generate a new short name for a polyhedron"
  (let ((shortname (copy-seq **next-shortname))
	(pos (position-if #'not-z **next-shortname :from-end t)))
    (if pos 
      (setf (char **next-shortname pos)
	    (succ-char (char **next-shortname pos)))
      (setf **next-shortname (succ-string **next-shortname)))
    (intern shortname)))


(defun not-z (ch) (char/= ch #\z))


(defun succ-char (ch) (code-char (1+ (char-code ch))))


(defun succ-string (str) 
  (make-string (1+ (length str)) :initial-element #\a))


;-------------------------------------------------------------------------------
; Utilities
;-------------------------------------------------------------------------------

(defun eval-pol-box (pol)
  (let* ((vl (mapcar #'get-coord (get-vlist pol)))
	 (min (first vl))
	 (max (first vl)))
    (dolist (v vl (list min max))
      (setq min (mapcar #'min v min))
      (setq max (mapcar #'max v max)))))


(defun errmsg (n)
  (let ((str (case n
	       (1 "Error: name is not in pol table")
	       (2 "Error: name is already in pol table")
	       (3 "Error: not a valid polyhedron")
	       (4 "Error: not a valid expression")
	       (5 "Error: W-rep not available"))))
    (format t str)
    nil))


;-------------------------------------------------------------------------------
; Normalization and Pruning. The tree structure with the bounding boxes
;-------------------------------------------------------------------------------

(defstruct node                                          ;a node of the tree
  (op nil :type symbol)                                  ;operator or operand
  (box nil :type list))                                  ;bounding box


(defvar emptybox '(nil nil))                        ;empty b. box


(defvar univbox '(t t))                             ;full space b. box


(defvar empty (make-node :op 'empty :box emptybox)) ;empty node


(defvar univ (make-node :op 'univ :box univbox))    ;full space node


(defun tree-box (tree)                                 ;bounding box of a 
  (if (node-p tree) (node-box tree) (node-box (first tree))))  ;subtree


(defun box* (b1 b2)                                    ;box intersection
  (cond ((empty-box? b1) emptybox)
	((empty-box? b2) emptybox)
	((univ-box? b1) b2)
	((univ-box? b2) b1)
	(t (list (mapcar #'max (first b1) (first b2)) 
		 (mapcar #'min (second b1) (second b2))))))


(defun box+ (b1 b2)                                    ;box union
  (cond ((univ-box? b1) univbox)
	((univ-box? b2) univbox)
	((empty-box? b1) b2)
	((empty-box? b2) b1)
	(t (list (mapcar #'min (first b1) (first b2)) 
		 (mapcar #'max (second b1) (second b2))))))


(defun empty-box? (b)                                  ;empty box test
  (cond ((equal b emptybox) t)
	((equal b univbox) nil)
	(t (let ((dimensions (mapcar #'- (second b) (first b))))
	     (or (some #'minusp dimensions)
		 (every #'(lambda (x) (< x **epsilon)) 
			dimensions))))))


(defun univ-box? (b)                                   ;full space box test
  (equal b univbox))


(defun inters-box? (b1 b2)                              ;box intersection test
  (cond ((empty-box? b1) nil)
	((empty-box? b2) nil)
	((univ-box? b1) t)
	((univ-box? b2) t)
	(t (not (empty-box? (box* b1 b2))))))


(defun a-tree (op L R)                                  ;recalculate a tree
  (let* ((boxL (tree-box L))                            ; starting with an op.
	 (boxR (tree-box R)))                           ; and the child subtrees
    (case op
      (* (if (inters-box? boxL boxR) 
	   (list (make-node :op '* :box (box* boxL boxR)) L R) 
	   empty))
      (- (if (inters-box? boxL boxR) 
	   (list (make-node :op '- :box boxL) L R) 
	   L))
      (+ (if (inters-box? boxL boxR) 
	   (list (make-node :op '+ :box (box+ boxL boxR)) L R) 
	   (list (make-node :op '& :box (box+ boxL boxR)) L R)))
      (& (list (make-node :op '& :box (box+ boxL boxR)) L R))
      (/ (list (make-node :op '/ :box boxL) L empty)))))


;-------------------------------------------------------------------------------
; Transformation: B-expr -> P-form
;-------------------------------------------------------------------------------

; P-form = (P-cell ..)
; P-cell = (P-poscell P-negcell)
; P-poscell = (pol-name ..)
; P-negcell = (pol-name ..)


(defconstant empty-P-form '(((empty) nil)))


(defun B-expr-to-P-form (expr)                             ;main function
  (tree-to-P-form 
   (move-neg-pol 
    (normalize 
     (B-expr-to-tree 
      (remove-neg 
       (B-expr-to-binary expr)))))))


(defun B-expr-to-binary (expr)
  (cond 
   ((symbolp expr) expr)
   ((equalp (first expr) '/) expr)
   (t (reduce #'(lambda (x y)
		  (list (first expr) 
			(B-expr-to-binary x) 
			(B-expr-to-binary y)))
	      (cdr expr)))))


(defun remove-neg (expr)                                   ;remove complement
   (if (symbolp expr)                                      ; moving them to 
     expr                                                  ; the root
     (remove-one-neg (list (first expr) 
			 (remove-neg (second expr)) 
			 (remove-neg (third expr))))))


(defun remove-one-neg (expr)                                ;recursively used by 
  (let* ((opT (first expr))                                 ; remove-neg
	 (L (second expr))
	 (listL (listp L))
	 (opL (if listL (first L) nil))
	 (R (third expr))
	 (listR (listp R))
	 (opR (if listR (first R) nil))
	 (LL (if listL (second L) nil))               
	 (RL (if listR (second R) nil)))
    
    (case opT
      (/ (cond ((equal opL '/) LL)
	       (t expr)))
      
      (* (cond ((and (equal opL '/) 
		     (equal opR '/)) (list '/ (list '+ LL RL)))
	       ((equal opL '/) (list '- R LL))
	       ((equal opR '/) (list '- L RL))
	       (t expr)))
      
      (- (cond ((and (equal opL '/) 
		     (equal opR '/)) (list '- RL LL))
	       ((equal opL '/) (list '/ (list '+ LL R)))
	       ((equal opR '/) (list '* L RL))
	       (t expr)))
      
      (+ (cond ((and (equal opL '/) 
		     (equal opR '/)) (list '/ (list '* LL RL)))
	       ((equal opL '/) (list '/ (list '- LL R)))
	       ((equal opR '/) (list '/ (list '- RL L)))
	       (t expr))))))


(defun B-expr-to-tree (expr)                               ;build the tree with
  (if (symbolp expr)                                       ; bounding boxes 
    (let ((pol-cons (find-pol-cons expr)))                 ; assoc. to the 
      (if pol-cons                                         ; B-expr
	(let* ((pol-rec (get-pol-rec pol-cons))
	       (subtree (cond ((pol-rec-W-rep pol-rec) 
			       (make-node :op expr :box (pol-rec-box pol-rec)))
			      ((pol-rec-P-form pol-rec) 
			       (P-form-to-tree (pol-rec-P-form pol-rec) 
					       (pol-rec-sign pol-rec)))
			      (t (B-expr-to-tree (pol-rec-B-expr pol-rec))))))
	  (multiple-value-bind (sign P-form) (tree-to-P-form subtree)
	    (setf (pol-rec-P-form pol-rec) P-form)
	    (setf (pol-rec-sign pol-rec) sign)
	    subtree))
	(errmsg 1)))
    (case (first expr)
      (+ (a-tree '+ (B-expr-to-tree (second expr)) 
		    (B-expr-to-tree (third expr))))
      (* (a-tree '* (B-expr-to-tree (second expr)) 
		    (B-expr-to-tree (third expr))))
      (- (a-tree '- (B-expr-to-tree (second expr)) 
		    (B-expr-to-tree (third expr))))
      (/ (a-tree '/ (B-expr-to-tree (second expr)) empty)))))


(defun tree-to-P-form (tree)
  (cond ((node-p tree) (values '+ (pos-tree-to-P-form tree)))
	((equal (node-op (first tree)) '/) 
	 (values '- (pos-tree-to-P-form (second tree))))
	(t (values '+ (pos-tree-to-P-form tree)))))


(defun pos-tree-to-P-form (tree)
  (if (node-p tree)
    (list (list (list (node-op tree)) '()))
    (case (node-op (first tree))
      ((* -) (list (tree-to-P-cell tree)))
      (& (cons (tree-to-P-cell (second tree)) 
	       (pos-tree-to-P-form (third tree)))))))


(defun tree-to-P-cell (tree)
  (if (node-p tree)
    (list (list (node-op tree)) '())
    (let ((buf (tree-to-P-cell (second tree)))
	  (this (node-op (third tree))))
      (case (node-op (first tree))
	(* (list (cons this (first buf)) (second buf)))
	(- (list (first buf ) (cons this (second buf))))))))


;-------------------------------------------------------------------------------
; Normalization algorithm

(defun P-form-to-tree (P-form sign)
  (let ((postree (if (= (length P-form) 1)
		   (P-cell-to-tree (first P-form))
		   (a-tree '& 
			   (P-cell-to-tree (first P-form)) 
			   (P-form-to-tree (cdr P-form))))))
    (if (equal sign '-)
      (a-tree '/ postree empty)
      postree)))


(defun P-cell-to-tree (P-cell)
  (let* ((poscell (first P-cell))
	 (negcell (second P-cell))
	 (posterm (first poscell))
	 (negterm (first negcell)))
    (cond ((and (null posterm) (null negterm)) empty)
	  ((and (null posterm) (null (cdr negcell))) (a-tree '- univ negterm))
	  ((and (null negterm) (null (cdr poscell))) posterm)
	  ((null posterm) 
	   (a-tree '- (P-cell-to-tree (list nil (cdr negcell))) negterm))
	  ((null negterm) 
	   (a-tree '* (P-cell-to-tree (list (cdr poscell) nil)) posterm))
	  ((string> posterm negterm)
	   (a-tree '* (P-cell-to-tree (list (cdr poscell) nil)) posterm))
	  (t (a-tree '- (P-cell-to-tree (list nil (cdr negcell))) negterm)))))


;-------------------------------------------------------------------------------
; Normalization algorithm
;-------------------------------------------------------------------------------

(defun normalize (tree)
  (if (node-p tree) 
    tree
    (do* ((temp (a-tree (node-op (first tree))
			(normalize (second tree))
			(normalize (third tree)))
		(a-tree (node-op (first newtree))
			(normalize (second newtree))
			(normalize (third newtree))))
	  (newtree  (do-subst temp) (do-subst temp)))
	 ((or (node-p newtree) (equal newtree temp)) newtree))))


(defun do-subst (tree)
  (if (node-p tree) 
    tree
    (do* ((temp tree  newtree)
	  (newtree  (subst-one temp) (subst-one temp)))
	 ((or (node-p newtree) (eq newtree temp)) newtree))))


(defun subst-one (tree)
  (if (node-p tree) 
    tree
    (let* ((opT (node-op (first tree)))
	   (L (second tree))
	   (listL (listp L))
	   (opL (if listL (node-op (first L)) (node-op L)))
	   (R (third tree))
	   (listR (listp R))
	   (opR (if listR (node-op (first R)) (node-op R)))
	   (LL (if listL (second L) nil))               
	   (LR (if listL (third L) nil))          
	   (RL (if listR (second R) nil))              
	   (RR (if listR (third R) nil))      
	   (opLR (if (node-p LR) (node-op LR) nil)))
      
      (case opT
	(/ tree)
	
	(* (cond ((equal opL 'empty) empty)                      ;L is a leaf
		 ((equal opR 'empty) empty)                      ;R is a leaf
		 ((equal opL 'univ) R)                           ;L is a leaf 
		 ((equal opR 'univ) L)                           ;R is a leaf 
		 ((equal opL '&) (a-tree '& (a-tree '* LL R) 
					    (a-tree '* LR R)))
		 (listR (case opR
			  (& (a-tree '& (a-tree '* L RL) (a-tree'* L RR)))
			  (* (a-tree '* (a-tree '* L RL) RR))
			  (- (a-tree '- (a-tree '* L RL) RR))
			  (t tree)))
		 (listL (case opL                                 ;R is a leaf 
			  (* (if (node-p LR)
			       (cond ((string= opLR opR) (a-tree '* LL R))
				     ((string> opLR opR) 
				      (a-tree '* (a-tree '* LL R) LR))
				     (t tree))
			       tree))
			  (- (if (node-p LR)
			       (cond ((string= opLR opR) empty)
				     ((string> opLR opR) 
				      (a-tree '- (a-tree '* LL R) LR))
				     (t tree))
			       tree))
			  (t tree)))
		 (t (cond ((string> opL opR) (a-tree '* R L))    ;L and R leaves
			  ((string= opL opR) L)
			  (t tree)))))
	
	(- (cond ((equal opL 'empty) empty)
		 ((equal opR 'empty) L)
		 ((equal opL 'univ) tree)
		 ((equal opR 'univ) empty)
		 ((equal opL '&) (a-tree '& (a-tree '- LL R) (a-tree '- LR R)))
		 (listR (case opR
			  (+ (a-tree '- (a-tree '- L RL) RR))
			  (& (a-tree '- (a-tree '- L RL) RR))
			  (* (a-tree '& 
				     (a-tree '- (a-tree '- L RR) RL) 
				     (a-tree '&  
					     (a-tree '*  (a-tree '- L RL) RR) 
					     (a-tree '-  (a-tree '* L RL) RR))))    
			  (- (a-tree '& 
				     (a-tree '- (a-tree '* L RR) RL) 
				     (a-tree '&  
					     (a-tree '-  (a-tree '- L RL) RR) 
					     (a-tree '*  (a-tree '* L RL) RR))))
			  (t tree)))
		 (listL (case opL
			  (* (if (node-p LR)
			       (cond ((string= opLR opR) empty)
				     ((string> opLR opR) 
				      (a-tree '* (a-tree '- LL R) LR))
				     (t tree))
			       tree))
			  (- (if (node-p LR)
			       (cond ((string= opLR opR) (a-tree '- LL R))
				     ((string> opLR opR) 
				      (a-tree '- (a-tree '- LL R) LR))
				     (t tree))
			       tree))
			  (t tree)))
		 (t (cond ((string> opL opR) (a-tree '* (a-tree '- univ R) L))
			  ((string= opL opR) empty)
			  (t tree)))))
	
	(+ (cond ((equal opL 'empty) R)
		 ((equal opR 'empty) L)
		 ((equal opL 'univ) univ )
		 ((equal opR 'univ) univ )
		 ((and (node-p L) (node-p R) (string= opL opR)) L)
		 (t (a-tree '& 
			    (a-tree '* L R) 
			    (a-tree '& (a-tree '- L R) (a-tree '- R L))))))
	
	(& (cond ((equal opL 'empty) R)
		 ((equal opR 'empty) L)
		 ((equal opL '&) (a-tree '& LL (a-tree '&  LR R)))
		 (t tree)))))))

;-------------------------------------------------------------------------------
; Re-normalization algorithm
;-------------------------------------------------------------------------------


(defun move-neg-pol (tree)                                ;move negative terms
  (if (node-p tree)                                       ; toward the leaves
    tree
    (case (node-op (first tree))
      (/ (a-tree '/ (move-neg-pol (second tree)) empty))
      (& (a-tree '& 
		 (move-neg-pol-in-minterm (second tree)) 
		 (move-neg-pol (third tree))))
      ((* -) (move-neg-pol-in-minterm tree)))))
  

(defun move-neg-pol-in-minterm (tree)                      ;tree is a "minterm"
  (if (or (node-p tree) (node-p (second tree)))            ; that is a subtree
    tree                                                   ; that contains only 
    (move-one-neg-pol (a-tree (node-op (first tree))       ; * and -
		       (move-neg-pol-in-minterm (second tree)) (third tree)))))


(defun move-one-neg-pol (tree)
  (let* ((opT (node-op (first tree)))
	 (L (second tree))
	 (opL (node-op (first L)))
	 (LL (second L))
	 (LR (third L))
	 (R (third tree)))
    (if (and (equal opT '*) (equal opL '-))
      (if (equal LL univ)
	(a-tree '- R LR)
	(a-tree '- (a-tree '* LL R) LR))
      tree)))


;-------------------------------------------------------------------------------
; Selection. Transformation: P-form -> s-form
;-------------------------------------------------------------------------------

; s-form = (s-cell ..)
; s-cell = (s-poscell s-negcell)
; s-poscell = (s-term ..)
; s-negcell = (s-term ..)
; s-term = (pol-name . sid)


(defconstant empty-s-form '((((empty . s0)) nil)))


(defun P-form-to-s-form (P-form &optional (cit-flag t))    ;main function
  (if (equalp P-form empty-P-form)                         ;if empty P-form
    empty-s-form                                           ; return empty s-form
    (progn
      (format t "~%Number of P-constituents: ~D" (length P-form))      
      (format t "~%Selecting P-constituent...")
      (let* ((i 0)
	     (s-form (mapcan #'(lambda (P-cell)
				 (format t "~D  " (setf i (1+ i)))
				 (P-cell-to-s-cell P-cell cit-flag))
			     P-form)))
	(if s-form s-form empty-s-form)))))


(defun P-cell-to-s-cell (P-cell &optional (cit-flag t))
  (let ((conv (get-P-poscell-vert (first P-cell))))         ;conv=list of lists 
    (if (and cit-flag                                       ; of vertices
	     (not (convex-intersection-test conv))) 
      (values)
      (s-list-list-to-s-cell 
       (list
	(P-posneg-cell-to-s-list-list (first P-cell) conv)
	(P-posneg-cell-to-s-list-list (second P-cell) conv))))))


(defun P-posneg-cell-to-s-list-list (P-posneg-cell conv)
  (let ((before ())
	(after conv))
    (mapcar 
     #'(lambda (P-term) 
	 (prog1
	   (P-term-to-s-list P-term (append before (cdr after)))
	   (setf before (append before (list (first after))))
	   (setf after (cdr after))))
     P-posneg-cell)))


(defun P-term-to-s-list (P-term conv)  
  (let ((res '()))
    (dolist (s (get-sid-vert-list (find-W-rep P-term)) res)
      (when (convex-intersection-test (cons (second s) conv))
	(push (cons P-term s) res)))))


(defun s-list-list-to-s-cell (s-list-list)
  (mapcar #'(lambda (s-poscell)
	      (add-s-negcell s-poscell (second s-list-list)))
	  (remove-if #'empty-cell (product (first s-list-list)))))


(defun empty-cell (s-poscell)
  (not (convex-intersection-test (mapcar #'third s-poscell))))


(defun add-s-negcell (s-poscell s-neg-list)
  (list (mapcar #'(lambda (s-term)
		    (cons (first s-term) (second s-term)))
		s-poscell)
	(let ((res '()))
	  (dolist (s-list s-neg-list)
	    (dolist (s-term s-list)
	      (unless (empty-intersection s-term s-poscell)
		(push (cons (first s-term) (second s-term)) res))))
	  res)))


(defun empty-intersection (s-term s-poscell)
  (not (convex-intersection-test 
	(cons (third s-term) (mapcar #'third s-poscell)))))


(defun product (listoflists)
    (reduce #'b-product 
	    (mapcar #'(lambda (l) (mapcar #'list l)) listoflists)))


(defun b-product (l1 l2)
  (let ((res '()))
    (dolist (e1 l1)
      (dolist (e2 l2)
	(push (append e1 e2) res)))
    res))


;-------------------------------------------------------------------------------
; Decomposition. Transformation: s-form -> pol
;-------------------------------------------------------------------------------

(defun s-form-to-svertcoord-list-list (s-form)
  (mapcar #'(lambda (s-cell) 
	      (mapcar 
	       #'(lambda (s-term)
		   (get-svertcoord (cdr s-term) (find-W-rep (car s-term))))
	       (first s-cell)))
	  s-form))


(defun s-form-to-W-rep (s-form build-ad-flag)
  (if (equalp s-form empty-s-form)                   ;if empty s-form
    (make-empty-polyhedron)                          ; return empty polyhedron
    (svertcoord-list-list-to-pol
     (s-form-to-svertcoord-list-list s-form) build-ad-flag)))


;-------------------------------------------------------------------------------
; Utilities
;-------------------------------------------------------------------------------

(defun get-P-poscell-vert (P-poscell)
  (mapcar #'get-P-term-vert P-poscell))


(defun get-P-term-vert (P-term)
  (mapcar #'get-coord (get-vlist (find-W-rep P-term))))


(defun get-sid-vert-list (pol)
  (mapcar #'(lambda (s)
	      (list (get-id s) (get-vertcoord s pol)))
	  (get-slist pol)))
