; -*- package: triangulation -*-
;-------------------------------------------------------------------------------
;           Universita' di Roma "La Sapienza"
;       Dipartimento di Informatica e Sistemistica
;            Via Buonarroti 12 - 00185 Roma
;
;                      CAD Group                 
; 
; Project: SimpleXn
; Identification: Package Triangulation                           
; Date: 27-12-90
; Note: Simplicial Decomposition, Von Hohenbalken Algorithm
;       with optional build-adjacencies flag
;             
; Version: 1.0
;              
; Author: F. Bernardini
;-------------------------------------------------------------------------------

(provide 'triang)                              ;module name


(defpackage triangulation)
(inpackage triangulation)          ;package name

(export '(convex-list-to-pol
          build-ad-pol
          svertcoord-list-list-to-pol
          convex-intersection-test
          convex-intersection))
                        
(require 'kernel)
(require 'matlib)
(require 'pl)
(require 'intersec)

(use-package 'kernel)
(use-package 'matlib)
(use-package 'pl)
(use-package 'intersect)


;-------------------------------------------------------------------------------
; Von Hohenbalken algorithms
;-------------------------------------------------------------------------------

(defvar B nil)                                       ;global variables used
(defvar U nil)                                       ; by corral


(defun corral (y r Q)
"Attention: this function uses global variables B and U.
Returns x, the nearest-to-y point in the convex with  
vertices (B union r); B will contain the set of vertices 
of the smallest-dimension simplex containing x; U will 
contain the convex coordinates of x in B"
  (do* ((nB (length B) 
            (length B))
        (t1 (mat-diff (mat-get-columns Q B) 
                      (mat-get-columns Q (make-list nB :initial-element r)))
            (mat-diff (mat-get-columns Q B) 
                      (mat-get-columns Q (make-list nB :initial-element r))))
        (t2 (vet-diff y (mat-get-a-column Q r)))
        (W (if (> nB 0) (mat-vet-solve t1 t2) nil)
           (if (> nB 0) (mat-vet-solve t1 t2) nil)))
       ((every #'(lambda (ww) (> ww **epsilon)) W)
        (push r B)
        (setf U (concatenate 'vector (list (- 1 (reduce #'+ W))) W))
        (mat-vet-product (mat-get-columns Q B) U))
    (let* ((t3 (reduce #'max 
                       (map '(vector real) 
                            #'(lambda (ww uu) (- 1 (/ ww uu))) W U)))
           (t4 (map '(vector real) 
                    #'(lambda (ww uu) 
                        (> (+ uu (/ (- ww uu) t3)) **epsilon))
                    W U)))
      (setf B (remove nil 
                      (map 'list 
                           #'(lambda (tt bb) (if tt bb nil)) 
                           t4 B)))
      (setf U (remove nil 
                      (map 'vector 
                           #'(lambda (tt uu) (if tt uu nil)) 
                           t4 W))))))


(defun near (y Q)
  "Returns the nearest-to-y point in the convex Q"
  (setf B nil)
  (setf U nil)
  (do* ((r 0 t2)
        (x (corral y r Q) 
           (corral y r Q))
        (g (vet-diff y x) 
           (vet-diff y x))
        (t2 (vet-max-position (vet-mat-product g Q))
            (vet-max-position (vet-mat-product g Q))))
       ((<= (abs (vet-product 
                  g 
                  (vet-diff x (mat-get-a-column Q t2)))) 
            **epsilon) x)))


(defun dimension (Q)
  "Returns the dimension of conv Q."
  (setf B nil)
  (setf U nil)
  (do* ((n (array-dimension Q 1))
        (y (mat-vet-product 
            Q 
            (let ((t1 (make-array (list n) 
                                  :element-type 'real))
                  (t2 0))
              (dotimes (j n) 
                (setf (svref t1 j) (log (+ 2 j))) 
                (setf t2 (+ t2 (svref t1 j))))
              (dotimes (j n t1) 
                (setf (svref t1 j) (/ (svref t1 j) t2))))))
        (r (vet-max-position (vet-mat-product y Q)) 
           (vet-max-position (vet-mat-product g Q)))
        (x (corral y r Q) 
           (corral y r Q))
        (g (vet-diff y x) (vet-diff y x)))
       ((every #'(lambda (gg) (<= (abs gg) **epsilon)) g) 
        (1- (length B)))))


(defun vertices (Q S)
  "Return the list of those points in S,
a subset of Q, that are vertices of conv Q"
  (setf B nil)
  (setf U nil)
  (do* ((n (array-dimension Q 1))
        (p (mat-vet-product 
            Q 
            (let ((t1 (make-array (list n) 
                                  :element-type 'real)))
              (dotimes (j n) (setf (svref t1 j) (/ n)))
              t1)))
        (V nil)
        (ST S)
        (h (first S) (first ST)))
       ((null h) (sort V #'<))
    (setf B nil)
    (setf ST (cdr ST))
    (do* ((y (mat-get-a-column Q h))
          (g (vet-diff y p))
          (t1 (vet-mat-product g Q) (vet-mat-product g Q))
          (m (reduce #'max t1) (reduce #'max t1))
          (k (vet-positions m t1) (vet-positions m t1))
          (nk (length k) (length k))
          (l (first k) (first k))
          (stop nil))
         (stop)
      (when (and (= nk 1) (member l S))
        (setf V (adjoin l V))
        (setf ST (remove l ST)))
      (if (or (> nk 1) (not (member h V)))
        (progn (setf g (vet-diff y (corral y (nextitem h k) Q)))
               (setf stop 
                     (every #'(lambda (gg) (<= (abs gg) **epsilon)) g)))
        (setf stop t)))))


(defun edges (Q)
  "Returns the adjacency matrix of vertices:
Two vertices are adjacent if there exists an edge
connecting them."
  (setf B nil)
  (setf U nil)
  (do* ((nQ (array-dimension Q 1))
        (A (make-array (list nQ nQ)
                       :element-type 'integer
                       :initial-element 0))
        (h 0 (1+ h))
        (V (mat-copy Q) (mat-copy Q))
        (y)
        (g))
       ((= h nQ) A)
    (mat-set-a-column V h (vet-scalar-product 
                           (mat-sum-columns Q) 
                           (/ nQ)))
    (setf y  (mat-get-a-column Q h))
    (setf g (vet-diff y (mat-get-a-column V h)))
    (setf B nil)
    (do* ((t1 (vet-mat-product g Q) (vet-mat-product g Q))
          (r (progn (setf (aref t1 h) (1- (reduce #'min t1)))
                    (vet-max-position t1))
             (progn (setf (aref t1 h) (1- (reduce #'min t1)))
                    (vet-max-position t1))))
         ((> (vet-product g (vet-diff y (mat-get-a-column Q r))) 
             0.00001))
      (setf g (vet-diff y (mat-set-a-column V h (corral y r Q)))))
    (let* ((t4 (mat-diff 
                (mat-get-columns Q (make-list nQ :initial-element h)) 
                V))
           (Z (vertices (mat-sum
                         V
                         (mat-vet-mult
                          t4
                          (vet-divide
                           (vet-mat-product 
                            g (mat-diff 
                               (mat-get-columns V 
                                                (make-list
                                                 nQ 
                                                 :initial-element h)) 
                               V))
                           (vet-mat-product g t4))))
                        (coerce 
                         (let ((S (make-array (list (- nQ h)) 
                                              :element-type 'integer)))
                           (dotimes (j (- nQ h) S) 
                             (setf (svref S j) (+ j h)))) 
                         'list))))
      (dotimes (j nQ)
        (when (member j Z) (setf (aref A h j) 1))))))
        
          
(defun permuter (A)
  "Returns a vertices permutation that
optimizes the face extracting procedure performance"
  (do* ((nA (array-dimension A 0))
        (AT (mat-sum A (mat-trans A)))
        (n (vet-interval nA))
        (r (make-array (list nA) 
                       :element-type 'integer 
                       :initial-element 1))
        (s (make-array (list nA) 
                       :element-type 'real 
                       :initial-element 1))
        (p (vet-grade-up (mat-sum-columns AT))
           (vet-grade-up (vet-mult r (mat-vet-product AT s))))
        (h 0 (1+ h)))
       ((= h nA)
        (values 
         (dotimes (i nA AT) 
           (dotimes (j nA) 
             (when (>= i j) (setf (aref AT i j) 0)))) n))
    (setf (svref r h) 0)
    (setf (svref s h) (* h 0.00001))
    (let ((nn (vet-copy n))
          (ATC (mat-copy AT)))
      (dotimes (i nA)
        (setf (svref n i) (svref nn (svref p i)))
        (dotimes (j nA) 
          (setf (aref AT i j) (aref ATC (svref p i) (svref p j))))))))
 

(defun filter (S f M W c dim)
  "Build  a face or reject it."
  (setf B S)
  (setf U nil)
  (let* ((t1 (mat-sum-columns 
              (mat-scalar-product 
               (mat-get-columns M (cons f S)) (/ (1+ (length S))))))
         (y (vet-sum t1 (vet-scalar-product (vet-diff c t1) 0.001)))
         (x (corral y f M))
         (hp (vet-diff y x))
         (V (vet-mat-product hp M))
         (r (vet-max-position V))
         (t2 (vet-product hp x))
         (t3 (vet-product hp (vet-diff x (mat-get-a-column M r))))
         (G S))
    (dotimes (j (array-dimension M 1))
      (when (and (< (- (svref V j) t2) **epsilon) (not (member j S)))
        (setf G (pushlast j G))))
    (unless (or (= (svref W (first S)) 1) (< (abs t3) **epsilon))
      (if (= (length S) dim)
        (setf G (list f))
        (progn 
          (setf B S)
          (do* ((t1 (mat-sum-columns 
                     (mat-scalar-product 
                      (mat-get-columns M B) (/ (length B)))))
                (y (vet-sum t1 (vet-scalar-product (vet-diff x t1) 0.001)))
                (hp (vet-diff y (corral y r M)) (vet-diff y (corral y r M)))
                (stop nil))
               ((every #'(lambda (hhpp) 
                           (< (abs hhpp) **epsilon)) hp)
                (setf G (list f)))
            (setf V (vet-mat-product hp M))
            (setf r (vet-max-position V))
            (setf stop (> (vet-product 
                           hp 
                           (vet-diff y (mat-get-a-column M r))) 
                          **epsilon))))))
    G))


(defun faces (M A d)
  "Extract faces of every order."
  (let* ((n (array-dimension M 1))                     ;number of vertices
         (W (map '(vector integer)                     ;simple vertices
                 #'(lambda (x) (if (= x d) 1 0))
                 (mat-sum-columns (mat-sum A (mat-trans A)))))
         (c (mat-vet-product 
             M 
             (let ((t1 (make-array (list n) :element-type 'real))
                   (t2 0))
               (dotimes (j n) 
                 (setf (svref t1 j) (log (+ j n))) 
                 (setf t2 (+ t2 (svref t1 j))))
               (dotimes (j n t1) 
                 (setf (svref t1 j) (/ (svref t1 j) t2))))))
         (Z ()))
    (dotimes (i n)                                     ;1-faces
      (dotimes (j n)
        (when (= (aref A i j) 1) (setf Z (pushlast (list i j) Z)))))
    (setf Z (list Z))
    (do* ((Q (first Z) P)                              ;cycle on dimensions
          (P () ())
          (h 2 (1+ h)))
         ((>= h d) Z)
      (dolist (G Q)                                    ;cycle on (h-1)-faces
        (do* ((S (subseq G 0 h))                       ;cycle on vert. neighbors
              (R (let ((t1 ())
                       (s1 (first S))
                       (sh (first (last S))))
                   (dotimes (j n t1)
                     (when (and (> j sh) (= (aref A s1 j) 1)) 
                       (setf t1 (pushlast j t1))))))
              (R (set-difference R G) (set-difference R G)))
             ((null R))
          (unless
            (some #'(lambda (U) (subsetp (cons (first R) S) U)) P)  
            (setf G (filter S (first R) M W c d))
            (when (> (length G) 1) (setf P (pushlast G P))))))
      (setf Z (pushlast P Z)))))
         

(defun rename-faces (F P)
  "Rebuild the original denomination of vertices, changed by permuter."
  (mapcar 
   #'(lambda (y) 
       (mapcar 
        #'(lambda (x) 
            (mapcar 
             #'(lambda (i) (svref P i)) x)) y)) F))


(defun simplicial-complex (scheme)
  "Build the simplicial complex associated with the scheme"
  (let ((Z ())
        (d (1+ (length scheme)))
        (x (caaar scheme)))
    (dolist (face (first (last scheme)) Z)
      (unless (member x face)
        (if (= (length face) d)
          (push (cons x face) Z)
          (setf Z (append 
                   (mapcar #'(lambda (s) (cons x s))
                           (simplicial-complex 
                            (extract-scheme face (butlast scheme)))) 
                   Z)))))))
            

(defun extract-scheme (face scheme)
  "Extract the sub-scheme of a face."
  (mapcar
   #'(lambda (j-faces)
       (let ((sel-faces ()))
         (dolist (f j-faces sel-faces)
           (when (subsetp f face) (push f sel-faces)))))
   scheme))


(defun convex-scheme (vcoordlist d) 
  "Build the scheme of a convex"
  (let* ((V (vcoordlist-to-vertmat vcoordlist)))
    (multiple-value-bind (AP P) (permuter (edges V))        ; else extract edges
      (rename-faces 
       (faces (mat-get-columns V P) AP d)                   ; and all the faces
       P))))


;-------------------------------------------------------------------------------
; Polyhedron reconstruction
;-------------------------------------------------------------------------------

(defun convex-list-to-vlist (convex-list)
  "Build vlist and the translation table."
  (let ((vlist ())
        (vtable ())
        (i 0))
    (dolist (convex convex-list)
      (let ((vtable-slot ()))
        (dolist (vertex convex)
          (let ((j (exist-vertex? vertex vlist)))
            (unless j
              (setf j i)
              (setf i (1+ i))
              (setf vlist 
                    (pushlast 
                     (make-vertex (make-vertex-symbol j) vertex) 
                     vlist)))
            (setf vtable-slot (pushlast j vtable-slot))))
        (setf vtable (pushlast (coerce vtable-slot 'vector) vtable))))
    (values vlist vtable)))


(defun exist-vertex? (vertex vlist)
  (position vertex (mapcar #'second vlist) :test #'same-vertex))


(defun same-vertex (v1 v2)
  (every #'(lambda (c1 c2) (< (abs (- c1 c2)) **epsilon)) v1 v2))


(defun translate-vid (scheme vtable-slot)
  "Rename vertices"
  (mapcar
   #'(lambda (s-faces)
       (mapcar
        #'(lambda (face)
            (mapcar #'(lambda (vid) (svref vtable-slot vid)) face))
        s-faces))
   scheme))


(defun sc-to-slist (sc)
  "Transform a simplicial complex in a slist"
  (let ((i -1)
        (d+1 (length (first sc))))
    (mapcar #'(lambda (s) 
                (make-simplex 
                 (make-simplex-symbol (setq i (1+ i)))
                 '+
                 (mapcar #'(lambda (c) 
                             (make-vertex-symbol c)) 
                         (sort (copy-seq s) #'<))
                 (make-list d+1 :initial-element 0)))
            sc)))


(defun extract-faces (sc faces)
  "Extract faces from a simplicial complex"
  (let* ((d+1 (length (first sc)))
         (s-faces (mapcan
                   #'(lambda (vl)
                       (let ((fl ()))
                       (dotimes (i d+1 fl)
                         (push (remove-pos i vl) fl))))
                   (copy-seq sc))))
    (mapcar
     #'(lambda (f)
         (let ((fl ()))
           (dolist (s-face s-faces fl)
             (when (subsetp s-face f)
               (push s-face fl)))))
     faces)))


(defun svertcoord-list-list-to-pol (svertcoord-list-list &optional (build-ad-flag t))
  (format t "~%Number of convexes : ~D" (length svertcoord-list-list))
  (format t "~%Calculating convexes...")  
  (let* ((i 0)
         (convex-list (mapcar #'(lambda (svl)
                                  (format t "~D  " (setf i (1+ i)))
                                  (convex-intersection svl))
                              svertcoord-list-list))
         (dim-list (mapcar 
                    #'dimension 
                    (mapcar #'vcoordlist-to-vertmat convex-list)))
         (max-dim (reduce #'max dim-list)))
    (if (= max-dim 1)                                        ;if dim = 1
      (segment-list-to-pol convex-list)                      ; a 1-polyhedron
      (convex-list-to-pol                                    ;transform in a pol
       (remove nil (mapcar #'(lambda (c d)                   ;regularization
                               (if (= d max-dim) c nil))
                           convex-list dim-list))
       max-dim build-ad-flag))))


(defun segment-list-to-pol (convex-list)
  (multiple-value-bind
    (vlist vtable) (convex-list-to-vlist convex-list)
    (let* ((sc (mapcar 
                #'(lambda (v) (list (svref v 0) 
                                    (svref v 1)))
                vtable)))
      (format t "~%Reconstructing polyhedron...~%")
      (build-ad-pol vlist (sc-to-slist sc)))))


(defun convex-list-to-pol (convex-list d &optional (build-ad-flag t))
  (format t "~%Calculating schemes of convexes...")
  (let ((scheme-list (convex-list-to-scheme-list convex-list d)))
    (format t "~%Triangulating convexes...")
    (convex-scheme-list-to-pol convex-list scheme-list build-ad-flag)))


(defun convex-list-to-scheme-list (convex-list d)
  (let ((i 0)) 
    (mapcar #'(lambda (c)
                (format t "~D  " (setf i (1+ i)))
                (convex-scheme c d))
            convex-list)))


(defun convex-scheme-list-to-pol (convex-list scheme-list  &optional (build-ad-flag t))
  "Transform a list of convexes in a polyhedron." 
  (multiple-value-bind
    (vlist vtable) (convex-list-to-vlist convex-list)
    (let* ((sc ())
           (scheme-table 
            (map 'vector #'translate-vid scheme-list vtable))
           (face-table (map 'vector #'lastitem scheme-table)))
      (do* ((others (coerce 
                     (vet-interval (array-dimension scheme-table 0)) 
                     'list))
            (next (list 0))
            (i (or (car next) (car others)) (or (car next) (car others)))
            (sc-i ())
            (fc-i ()))
           ((null i))
        (setf others (remove i others))
        (setf next (cdr next))
        (setf sc-i (simplicial-complex (svref scheme-table i)))
        (setf fc-i (extract-faces sc-i (svref face-table i)))
        (setf sc (append sc sc-i))
        (dolist (j others)
          (let ((h 0))
            (dolist (f1 (svref face-table i))
              (let ((k 0))
                (dolist (f2 (svref face-table j))
                  (when (subsetp f1 f2)
                    (setf next (adjoin j next))
                    (let ((newfaces 
                           (append 
                            (remove-pos k (lastitem (svref scheme-table j)))
                            (nth h fc-i))))
                      (setf (first (last (svref scheme-table j))) newfaces)
                      (setf (svref face-table j) newfaces)))
                  (setq k (1+ k))))
              (setq h (1+ h)))))
        (format t "~D  " i))
      (format t "~%Reconstructing polyhedron...~%")
      (if build-ad-flag
        (build-ad-pol vlist (sc-to-slist sc))
        (make-polyhedron vlist (sc-to-slist sc))))))


(defun build-ad-pol (vlist slist)
  "Build a polyhedron, reconstructing the adjacencies 
and signs of simplices."
  (let* ((d+1 (length (get-vert (first slist))))
         (ns (length slist))
         (faces (make-array (list ns)
                            :initial-contents 
                            (mapcar
                             #'(lambda (vl)
                                 (let ((fl ()))
                                   (dotimes (i d+1 (reverse fl))
                                     (push (remove-pos i vl) fl))))
                             (mapcar #'get-vert slist))))
         (admat (make-array (list ns d+1)
                            :element-type 'integer
                            :initial-element -1))
         (signvet (make-array (list ns)
                              :element-type 'integer
                              :initial-element 0)))
    (do* ((others (coerce (vet-interval ns) 'list))
          (next (list 0))
          (i (or (car next) (car others)) (or (car next) (car others))))
         ((null i))
      (setf others (remove i others))
      (setf next (cdr next))
      (when (zerop (svref signvet i)) (setf (svref signvet i) 1))
      (dolist (j others)
        (let ((h 0))
          (dolist (f1 (svref faces i))
            (let ((k 0))
              (dolist (f2 (svref faces j))
                (when (equalp f1 f2)
                  (setf next (adjoin j next))
                  (if (minusp (aref admat i h))
                    (setf (aref admat i h) j)
                    (error-handler 0 "build-ad-pol"))
                  (if (minusp (aref admat j k))
                    (setf (aref admat j k) i)
                    (error-handler 0 "build-ad-pol"))
                  (if (zerop (svref signvet j))
                    (setf (svref signvet j) 
                          (- (* (expt -1 h) (expt -1 k) (svref signvet i))))))
                (setq k (1+ k))))
            (setq h (1+ h))))))
    (make-polyhedron
     vlist
     (let ((i -1))
       (mapcar #'(lambda (s)
                   (setf i (1+ i))
                   (make-simplex
                    (get-id s)
                    (if (plusp (svref signvet i)) '+ '-)
                    (get-vert s)
                    (let ((sad ()))
                      (dotimes (j d+1 sad)
                        (let ((ad (aref admat i j)))
                          (setf sad 
                                (pushlast 
                                 (if (minusp ad) 
                                   0 
                                   (make-simplex-symbol ad)) sad)))))))
               slist)))))
          

;-------------------------------------------------------------------------------
; Utilities
;-------------------------------------------------------------------------------

(defun nextitem (item lista)
  (let ((l (length lista))
        (p (position item lista)))
    (if p (nth (mod (1+ p) l) lista) (nth (mod 2 l) lista))))


(defun pushlast (item lista)
  (append lista (list item)))


(defun lastitem (l)
  (first (last l)))


(defun remove-pos (pos lista)
  (append (subseq lista 0 pos) (subseq lista (1+ pos))))


(defun print-scheme (F)
  (let ((d 1))
    (dolist (G F)
      (format t "~%~D-Facce : ~D~%" d (length G))
      (dolist (H G)
        (dolist (j H)
          (format t "~D  " j))
        (format t "~%"))
      (setq d (1+ d)))))
      

(defun vcoordlist-to-vertmat (vcoordlist)
  (let* ((m (length vcoordlist))                             ;number of vertices
         (n (length (first vcoordlist))))                    ;E-space dimension
    (mat-trans (make-array (list m n)                        ;matrix of vertices
                           :element-type 'real
                           :initial-contents vcoordlist))))


;-------------------------------------------------------------------------------
; Intersection of Convexes
;-------------------------------------------------------------------------------

; Nota Bene: in the following functions, all vertices must have the
; same number of coordinates.

(defun convex-intersection-test (VCLL)        ;VCLL (Vertex Coord. List List)
  (let ((num-convex (length VCLL)))           ; is a list of convexes
    (if (= num-convex 1)                      ; each convex is a list of vert.
      t                                       ; each vertex is a list of coord.
      (let* ((A (build-initial-matrix VCLL))        
             (nrows (array-dimension A 0))
             (ncols (array-dimension A 1)))
        (if (build-canonical-form A nrows (1- ncols)) t nil)))))


(defun convex-intersection (VCLL)              ;VCLL (Vertex Coord. List List)
  (let* ((A (build-initial-matrix VCLL))       ; is a list of convexes
         (nrows (array-dimension A 0))         ; each convex is a list of vert.
         (ncols (array-dimension A 1))         ; each vertex is a list of coord.
         (vert1 (car VCLL))
         (nvert1 (length vert1)))
    (multiple-value-bind 
      (A m n v0) (build-canonical-form A nrows (1- ncols))
        (calculate-rectangular-coordinates 
         vert1 
         (mapcar 
          #'(lambda (l) (subseq l 0 nvert1)) 
          (generate-vertices A m n v0))))))


(defun build-initial-matrix (VCLL)             ;VCLL (Vertex Coord. List List)
  (let* ((num-convex (length VCLL))            ; is a list of convexes
         (dim (length (caar VCLL)))            ; each convex is a list of vert.
         (num-vert-list (mapcar #'length VCLL)); each vertex is a list of coord.
         (nrows (- (* (1+ dim) num-convex) dim))
         (ncols (1+ (apply '+ num-vert-list)))
         (offs (* dim (1- num-convex)))
         (A (make-array (list nrows ncols)
                        :element-type 'real
                        :initial-element 0)))
    (do* ((s 0 (1+ s))
          (temp VCLL (cdr temp))
          (V1 (first temp) (first temp))
          (V2 (second temp) (second temp))
          (posr 0 (+ dim posr))
          (posc 1 (+ nv posc))
          (nv-list num-vert-list (cdr nv-list))
          (nv (first nv-list) (first nv-list)))
         ((= s (1- num-convex)) (fill-with-one A (+ s offs) posc nv))
      (copy-mat+ V1 A posr posc)
      (copy-mat- V2 A posr (+ nv posc))
      (fill-with-one A (+ s offs) posc nv))
    A))


(defun copy-mat+ (vlist mat posr posc)
  (let ((i posr)
        (j posc))
    (dolist (v vlist)
      (dolist (c v)
        (setf (aref mat i j) c)
        (setq i (1+ i)))
      (setq i posr)
      (setq j (1+ j)))))


(defun copy-mat- (vlist mat posr posc)
  (let ((i posr)
        (j posc))
    (dolist (v vlist)
      (dolist (c v)
        (setf (aref mat i j) (- c))
        (setq i (1+ i)))
      (setq i posr)
      (setq j (1+ j)))))

(defun fill-with-one (mat posr posc nv)
  (dotimes (i nv)
    (setf (aref mat posr (+ posc i)) 1))
  (setf (aref mat posr 0) 1))
