; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Funzioni per calcolare i limiti di un convesso
;;;
;;;  scritte da G. Proietti   31/3/1993
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'mypl)

(inpackage kernel)

(require 'pl)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                            (use-package 'pl)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Attenzione: le funzioni per calcolare il "box" di un complesso
;;;             sono state introdotte per ora nel package KERNEL. Risulta
;;;             impossibile importare la funzione PIVOT definita nel package
;;;             PL perche' nasce un conflitto con un simbolo (non meglio
;;;             identificato) presente nel package KERNEL.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  CONVEX-BOX
;;;
;;;   funzione che ritorna, dato un poliedro espresso tramite intersezione
;;;   di sottospazi, il box di contenimento dello stesso, senza side effects 
;;;   sugli argomenti
;;;
;;;  INPUT
;;;
;;;   equations := lista di liste contenente i coefficienti delle disequazioni
;;;                nel formato ( (b1 a11 a12 ... a1n) (b2 a21 a22 ... a2n).. )
;;;
;;;  OUTPUT
;;;
;;;   lista contenente gli estremi superiore ed inferiore del box di conteni-
;;;   -mento del poliedro nel formato ( (cmax1 ... cmaxn) (cmin1 ... cminn) )

(defun convex-box (equations)
  (let* ((neq    (length equations) )
         (dim    (- (length (car equations)) 1) )
         (plus   (make-list dim ))
         (minus  (make-list dim ))
         (arg    (construct-canonical-tableau neq dim equations))
         )
    (dotimes (i dim (list plus minus) )
      (setf (nth i minus) (1-PL 'min i arg dim neq))
      (setf (nth i plus)  (1-PL 'max i arg dim neq))
      )
    )
  )


(defun calc-limits (equations indices type)
  (let* ((neq     (length equations))
         (dim     (1- (length (car equations))) )
         (arg     (construct-canonical-tableau neq dim equations))
         )
    (mapcar '(lambda (x) (1-PL type x arg dim neq)) indices)
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;Risolve un problema di PL
;
(defun 1-PL (type-symbol index canonical-form dim neq)
  (let ( (objfun (make-list (1+ dim) :initial-element 0))
         res
         )
    (cond ( (equalp type-symbol 'min) (setf (nth (1+ index) objfun) 1) )
          ( (equalp type-symbol 'max) (setf (nth (1+ index) objfun) -1))
          )
    (setq res (nth index (resolvesim neq dim objfun (duplicate-arg neq dim canonical-form))))
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; ritorna una lista contenente un tableau in forma canonica per il
;;; con la riga dei coefficienti della funzione obiettivo vuota e
;;; pronta per la memorizzazione dei suddetti coefficienti,il numero di
;;; righe, di colonne ed un vettore specificante le variabili di base.
;;;
;;;
;;; INPUT :
;;;
;;;         neq           :=  numero delle disuguaglianze ( <= ) costituenti i vincoli
;;;         dim           :=  dimensione dello spazio ( numero delle variabili )
;;;         objfun        :=  lista dei coefficienti della funzione obiettivo
;;;                           ( t.noto coeff1 .... coeffn )
;;; OUTPUT
;;;
;;;  ( #2A( .. tableau ..) nrighe ncol #( .. base .. ) )

(defun construct-canonical-tableau (neq dim equations)
  (let ( (arg (store neq dim equations) ) )
    (progn
      (setq arg (multiple-value-list ( pl::build-canonical-form arg neq (+ dim neq 1))))
      (setf (first arg) (copy-arry neq dim (first arg)) )
      (setf (second arg) (+ 1 (second arg)) )
      (setf (fourth arg) (let ( (vec (make-array (+ neq 1) ) ) )
                           (dotimes (i neq vec)
                             (setf (aref vec (+ i 1)) (nth i (fourth arg))))))
      (return-from construct-canonical-tableau arg)
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; ritorna un duplicato della lista contenente il tableau con tableau e
;;; variabili di base nuovamente memorizzate per salvare gli originali
;;; dai side-effects del calcolo del simplesso

(defun duplicate-arg (neq dim arg)
  (let ( (newarg1 (make-array (list (+ neq 1) (+ dim neq 2)) :initial-element 0))
         (newarg4 (make-array (+ neq 1)))
         (arg1 (first arg))
         (arg4 (fourth arg)))
    (list
      (dotimes (i (+ neq 1) newarg1)
        (dotimes (j (+ dim neq 2))
          (setf (aref newarg1 i j) (aref arg1 i j))
        )
      )
      (second arg)
      (third  arg)
      (dotimes (i neq newarg4)
        (setf (aref newarg4 (+ i 1)) (aref arg4 (+ i 1)))
      )
    )
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  ritorna il punto soluzione del problema specificato, sotto forma di
;;;  lista di razionali

(defun resolvesim (neq dim objfun arg )
    (progn
      (setf (first arg) (store-objective-fun dim (first arg) objfun))
      (report-canonical  neq dim (first arg) (fourth arg))
      (if ( pl::simplex-algorithm (first arg) (second arg) (third arg) (fourth arg))
        (let* ( (result (make-list dim :initial-element 0) )
                (pos (search (list (+ dim 1)) (fourth arg) ) )
                (w   (if pos (aref (first arg) pos 0) 0) ) )
          (progn
            (do* ( (i 1 (+ i 1) ) )
                 ( (> i neq) )
              (let ( (n (aref (fourth arg) i) ) )
                (progn
                  (if (<= n dim )
                    (setf (nth (- n 1) result )  (aref (first arg) i 0) )
                  )
                )
              )
            )
            (dotimes (j dim)
              (setf (nth j result) (- (nth j result) w)))
            (return-from resolvesim result)
          )
        )
        (return-from resolvesim nil)
      )
    )
)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; memorizza i coefficienti della funzione obiettivo, trasformati in rational,
;;; nella prima riga del tableau fornito

(defun store-objective-fun (dim mat objfun)
  (progn
    (dotimes (i (+ 1 dim) mat)
      (setf (aref mat 0 i) (rationalize (nth i objfun)))
    )
    (add-w 0 dim mat)
    (return-from store-objective-fun mat)
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Crea un array dove memorizzare i coefficienti dei vincoli del problema
;;; per il successivo utilizzo della funzione build-canonical-form
;;; Da notare che inverte il segno del termine noto

(defun store  (neq dim equaz)
  (let ((mat (make-array (list neq (+ dim neq 2)) :initial-element 0)))
    (dotimes (i neq mat)
      (let ((a (nth i equaz)))
        (progn
          (setq a (cons (- (car a)) (cdr a)) )
          (if (>= (car a) 0)
            (progn
              (dotimes (j (+ dim 1) )
                (setf (aref mat i j) (rationalize (nth j a)))
              )
              (add-w i dim mat)
              (setf (aref mat i (+ dim i 2)) 1)
            )
            (progn
              (dotimes (j (+ dim 1) )
                (setf (aref mat i j) (rationalize (- (nth j a))))
              )
              (add-w i dim mat)
              (setf (aref mat i (+ dim i 2)) -1)
            )
          )
        )
      )
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Aggiunge il coefficiente della variabile necessaria a gestire
;;; le variabili non vincolate in segno nella posizione dim+1
;;;
;;; INPUT : row  riga sulla quale calcolare ed aggiungere il coefficiente
;;;         dim  dimensione dello spazio
;;;         mat  matrice

(defun add-w (row dim mat)
  (let ((w 0))
    (progn
      (dotimes (j dim)
        (setq w (+ w (aref mat row (+ j 1)))))
      (setf (aref mat row (+ dim 1)) (- w) )
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Copia la matrice in una nuova con una riga in piu ( corrispondente
;;; alla riga 0 ) dove poter memorizzare i coefficienti della f. obiettivo

(defun copy-arry (neq dim mat)
  (let ((newmat (make-array (list (+ neq 1) (+ dim neq 2)) :initial-element 0)))
    (dotimes (i neq newmat)
      (dotimes (j (+ dim neq 2) newmat)
        (setf (aref newmat (+ i 1) j) (aref mat i j))
      )
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Riporta il tableau in forma canonica a partire dal tableau a cui
;;; siano aggiunti i coefficienti della f. obiettivo, eseguendo una serie
;;; di operazioni di pivot in corrispondenza delle variabili di base
;;; che corrispondano a coefficienti di costo non nulli

(defun report-canonical (neq dim mat base)
  (dotimes (i neq mat)
    (if (/= 0 (aref mat 0 (aref base (+ i 1)) ) )
      ( pl::pivot mat (+ neq 1) (+ dim neq 2) (+ i 1) (aref base (+ i 1)))
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Converte una lista generalizzata di rational nella corrispondente
;;; lista di float

(defun ratio2float (lista)
  (let ( (a (car lista)) )
    (cond ( (null a) nil)
          ( (atom a) (cons (coerce a 'float) (ratio2float (cdr lista))))
          (    t     (cons (ratio2float a) (ratio2float (cdr lista))))
    )
  )
)
