; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; G. Proietti (2/4/1993)
;;;
;;;  richiedono le funzioni definite nei seguenti files :
;      MyPL.lisp
;      Pol-Complex-vertices-lim.lisp
;    e la presenza dei packages
;      matlib
;      pl

(provide 'pol_c_f)

(inpackage kernel)

(require 'matlib)
(use-package 'matlib)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  Calcola i valori estremi delle coordinate indicate contenenti il complesso poliedrale
;;;  specificato e le ritorna come lista.
;;;  Se specificato 'min calcola i valori minimi, se specificato 'max i massimi
;;;
;;;  INPUT  :  - un complesso poliedrale
;              - lista di indici ( compresi tra 0 ad n-1 ) individuanti le coordinate interessate
;                dal calcolo.
;              - simbolo indicante se calcolare il minimo o il massimo sulle coordinate
;                puo' essere  'min  oppure  'max
;;;
;;;  OUTPUT :  lista dei valori risultanti, nello stesso ordine specificato nella lista di indici
;;;
;    ESEMPIO : (pol-complex-limits ( (build-pol-complex-cube 3) '(0 2) 'min) )
;              ( 0.0 0.0 )
(defun pol-complex-limits (pol-compl indices type)
  (let ( (temp (recursive-limits (coerce-pol-complex pol-compl 'on-vertices)
                              (mat-identity (1+ (get-npol-complex pol-compl)))
                              indices
                              type
                              'on-vertices)) )
    (ratio2float (coerce temp 'list) )
    )
  )



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; restituisce le disequazioni rappresentanti una cella in funzione della
;;; matrice di trasformazione passata
;;;

; versione vecchia, adatta solo se lo spazio di immersione e' uguale a quello di oggetto

;(defun recalc-diseq (disequaz mat)
;  (dotimes (i (length disequaz) disequaz)
;    (setf (nth i disequaz) (coerce (vet-mat-product (coerce (nth i disequaz) 'vector) mat ) 'list) )
;  )
;)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Ricalcola le disequazioni rappresentanti una cella in funzione della
;;; matrice di trasformazione passata.
;;; Qualora la dimensione dello spazio di immersione sia maggiore della dimensione 
;;; interinseca dell'oggetto, le equazioni vengono estese ed al set delle stesse
;;; viene aggiunta una coppia di disequazioni per ogni coordinata aggiuntiva Xk, specificante
;;; la giacenza dell'oggetto sul piano Xk=0
;
;;;  INPUT  :  - Lista di liste, ogni sottolista rappresenta una faccia da ricalcolare,
;                nel formato ( t.noto  coeff1 ..... coeffn ) tutti intesi al primo membro.
;              - Matrice di trasformazione geometrica.
;;;
;;;  OUTPUT :  lista di liste nello stesso formato, con aggiunte eventualmente in coda
;              le equazioni per portare l'oggetto a dimensione pari a quella di immersione

(defun recalc-diseq (disequaz mat)
  (let ( (dimspace (car (array-dimensions mat)))  ;in effetti e' la dim dello spazio + 1
         (dimobj   (length (car disequaz)))   ;in effetti e' la dim dell'oggetto + 1
         (temp     nil ) )
    (progn
      (if (> dimspace dimobj)
        (progn
          (setq temp (make-list (- dimspace dimobj) :initial-element 0 ))
          (setq disequaz (dotimes (i (length disequaz) disequaz)
                           (setf (nth i disequaz) (append (nth i disequaz) temp))) )
          (dotimes (i (- dimspace dimobj) disequaz)
            (progn
              (setq temp (make-list dimspace :initial-element 0 ))
              (setf (nth (+ i dimobj) temp ) 1)
              (setq disequaz (append disequaz (list temp)) )
              (setq temp (make-list dimspace :initial-element 0 ))
              (setf (nth (+ i dimobj) temp ) -1)
              (setq disequaz (append disequaz (list temp)) )
            )
          )
        )
      )
      (dotimes (i (length disequaz) disequaz)
        (setf (nth i disequaz) (coerce (vet-mat-product (coerce (nth i disequaz) 'vector) mat ) 'list) ) )
    )
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  calcola i limiti di contenimento di un poliedro sottoposto alla trasformazione
;;;  specificata dalla matrice, sia rappresentato con facce che con vertici.
;    Il calcolo e' effettuato solo sulle coordinate specificate, di minimo o massimo
;    a seconda del type-symbol specificato (vedi commenti precedenti)
;    Ritorna la lista dei risultati, posti nello stesso ordine espresso dalla lista
;    di coordinate passata.

(defun calc-pol-limit (pol mat indices type-symbol)
  (if (equalp (get-pol-rep pol) 'on-vertices)
    (calc-limits-pol-on-verts pol mat indices type-symbol)
;else
    (let* ((cells-limits (get-pol-limits pol mat indices type-symbol))
           res
           i-coords
           )
      (dotimes (i (length indices) res)
        (setq i-coords (mapcar #'(lambda  (x) (nth i x)) cells-limits))
        (setq res (append res (list 
                               (cond ( (equalp type-symbol 'max) (gmax i-coords))
                                     ( (equalp type-symbol 'min) (gmin i-coords))
                                     )
                               )
                          )        
              )
        )
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  calcola i limiti di contenimento di tutte le celle di un poliedro e li ritorna in una lista

(defun get-pol-limits (pol mat indices type)
  (let  ((listacelle (get-clist pol))
          results
         (cell-box   nil)
         (disequaz   nil) )
    (dotimes (i (length listacelle) results )
      (setq disequaz (get-disequations (nth i listacelle) pol) )
      (setq cell-box (calc-limits (recalc-diseq disequaz mat) indices type) )
      (setq results (append results (list cell-box) ) )
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; ritorna le disequazioni che definiscono una cella
;;; sotto forma di lista di liste

(defun get-disequations (cell pol)
  (let ((listaeq    nil)
        (facce      nil)
        (segno      nil))
    (setq facce (get-cell-faces cell))
    (dotimes (i (length facce))
      (setq segno (cdr (nth i facce)) )
      (setq listaeq (cons (get-fcoord (car (nth i facce)) pol segno ) listaeq ) )
      )
    (return-from get-disequations listaeq)
    )
  )


;;; Funzioni per ricavare l'ennesima coordinata massima
;;; da un insieme di punti posti come lista di liste

;;; Preleva dalla lista di liste tutti gli elementi pos-esimi di ogni
;;; sottolista di lista e li pone nella lista risultato

(defun getnth (lista pos)
  (let ( (result   nil) )
    (dotimes (i (length lista) result)
      (setq result (append result (list (nth pos (nth i lista)))))
    )
  )
)

;;; Preleva il valore piu' grande presente nella lista argomento

(defun gmax (lista)
  (let* ( (mas    (car lista)) )
    (dotimes (i (1- (length lista)) mas )
      (setq mas (max mas (nth (+ i 1) lista )))
    )
  )
)

;;; Preleva il valore piu' piccolo presente nella lista argomento
(defun gmin (lista)
  (let* ( (mini   (car lista)) )
    (dotimes (i (1- (length lista) ) mini )
      (setq mini (min mini (nth (+ i 1) lista )))
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  routine che esegue effettivamente il calcolo dei limiti di contenimento dell' oggetto passato
;;;

(defun recursive-limits (obj mat1 indices type rep)
  (cond ( (polyhedron?      obj) (calc-pol-limit obj mat1 indices type) )
        ( (polyhedral-node? obj)
          (let* ( (mat2  (if (lisp-extension::symbol-equal rep 'on-faces) 
                           ;(mat-product (get-node-matrix obj) mat1)
                           (mat-product-extended (get-node-matrix obj) mat1)
                           ;else
                           ;(mat-product mat1 (get-node-matrix obj) )
                           (mat-product-extended mat1 (get-node-matrix obj) ) ) )
                  )
            (recursive-limits (get-node-pol-complex obj) mat2 indices type rep)
            )
          )
        ( (polyhedra-complex? obj)
          (let* ( (res (recursive-limits (car obj) mat1 indices type rep))
                  temp
                  )
            (progn
              (dotimes (i (- (length obj) 1) res )
                (setq temp (recursive-limits (nth (+ 1 i) obj) mat1 indices type rep) )
                (cond  
                 ((equalp type 'max) (setq res (mapcar #'(lambda  (x y) (if (> x y) x y)) res temp)))
                 ((equalp type 'min) (setq res (mapcar #'(lambda  (x y) (if (> x y) y x)) res temp)))
                 )
                )
              )
            )
          )
        )
  )
