;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Il seguente modulo mette a disposizione delle funzioni per la  ;;
;; manipolazione delle strutture dei simplessi cubici nelle celle ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'conv_ext)

(provide 'cub_ext)

(inpackage kernel)

;-------------------------------------------------------------------
; La seg. funzione memorizza la tavola hash dei simplessi cubici
; nel campo adeguato della cella 
;-------------------------------------------------------------------

(defun put-cubic-simplexes (pol cell-id simplexes)
  (let ((cell (get-polyhedron-cell (pol cell-id))))
    (setf (convex-sell-CUBIC-SIMPLEXES cell) simplexes)
  pol))

;-------------------------------------------------------------------
; La seg. funzione ottiene la tavola hash dei simplessi cubici della 
; cella data
;-------------------------------------------------------------------

(defun get-cubic-simplexes (pol cell-id)
  (let ((cell (get-polyhedron-cell pol cell-id)))
    (convex-cell-CUBIC-SIMPLEXES cell)))
;-------------------------------------------------------------------
; Definizione della struttura di un simplesso cubico
;-------------------------------------------------------------------

(defstruct (cubic-simplex (:constructor create-cubic-simplex))
    (identifier nil)
    (edges '())
    (vertices '())
    (Bezier-net nil)
    (adiacent-cubic-simplexes '()))
  
;-------------------------------------------------------------------
; La seg. funzione costuisce un simplesso cubico
;-------------------------------------------------------------------

(defun make-cubic-simplex
                        (identifier
                         edges
                         vertices
                         Bezier-net
                         adiacent-cubic-simplexes)
  (create-cubic-simplex
    :identifier identifier
    :edges edges
    :vertices vertices
    :Bezier-net bezier-net
    :adiacent-cubic-simplexes adiacent-cubic-simplexes))
  
;-------------------------------------------------------------------
; La seg. funzione crea l'identificatore per un simplesso cubico
;-------------------------------------------------------------------

(defun make-cubic-simplex-symbol (cell-id j)
  (intern
    (concatenate 'string (string cell-id) "_" (integer->string j))))

;-------------------------------------------------------------------
; La seg. funzione crea dalla lista degli indici delle coordinate
; baricentriche di un punto della bezier-net, l'identificatore
; costituito da "P_" e gli indici di cui sopra.
; Questo serve per la HASH TABLE delle dei punti della Bezier-net.
;-------------------------------------------------------------------

(defun make-id-coord-baric (coord-baric)
  (let ((y ""))
    (dolist (x coord-baric)
      (setq y (concatenate 'string y (integer->string x))))
    (intern
      (concatenate 'string "P_" y))))

;--------------------------------------------------------------------
; La seg. funzione restituisce la lista degli indici delle coordinate
; baricentriche a partire dall'identificatore dei punti di controllo.
; La funzione deve saltare i due caratteri iniziali "P_" 
; dell'identificatore.
;--------------------------------------------------------------------

(defun make-list-from-id-coord-baric (id-coord-baric)
  (let* ((nn (- (length id-coord-baric) 2))
         (y (make-list nn :initial-element 0)))
    (dotimes (i nn)
      (setf (nth i y) 
            (string->integer
              (subseq id-coord-baric (+ i 2) (+ i 3)))))
   y))
    
;--------------------------------------------------------------------
; La seg. funzione memorizza un punto nella hash-table della 
; Bezier-net del simplesso cubico dato in input
;--------------------------------------------------------------------

(defun put-point-in-Bezier-net (id-coord-baric coord-baric c-simplex)
  (let ((table-point (cubic-simplex-BEZIER-NET c-simplex)))
    (setf (gethash id-coord-baric table-point) coord-baric)))


;--------------------------------------------------------------------
; La seg. funzione estrae un punto dalla tabella hash della
; Bezier-net del simplesso cubico dato in input
;--------------------------------------------------------------------

(defun get-point-of-Bezier-net (id-coord-baric c-simplex)
  (let ((table-point (cubic-simplex-BEZIER-NET c-simplex)))
    (gethash id-coord-baric table-point)))

;--------------------------------------------------------------------
; La seg. funzione ritorna la hash-table dei simplessi cubici di una
; cella.
;--------------------------------------------------------------------

(defun get-cell-cubic-simplexes (cell)
  (convex-cell-CUBIC-SIMPLEXES cell))

;--------------------------------------------------------------------
; La seg. funzione ritorna il centroide di una cella
;--------------------------------------------------------------------

(defun get-cell-centroid (cell)
  (convex-cell-CENTROID cell))


;------------------------------------------------------------------
; La seg. funzione calcola il centroide di una cella in base alla
; sua lista di vertici del tipo :
; ((V0 (. . .)) (V1 (. . .) ...................)
; Restituisce la lista delle coordinate del centroide.
;------------------------------------------------------------------

(defun calculate-centroid (points)
  (let* ((n (length points))
         (m (length (second (first points))))
         (sum (make-list m :initial-element 0))
         (centroid (make-list m :initial-element 0))
         point
         temp)
    (dotimes (i n)
      (setq point (second (nth i points)))
      (dotimes (j m)
        (setf (nth j sum) (+ (nth j sum) (nth j point)))))
    (dotimes (j m)
      (setf (nth j centroid) (/ (nth j sum) n)))
    centroid)) 
      
;-------------------------------------------------------------------
; La seg. funzione restituisce l'identificatore del centroide
;-------------------------------------------------------------------

(defun make-centroid-symbol (cell-id)
   (intern
      (concatenate 'string (string cell-id) "_C")))
   
        
;------------------------------------------------------------------
; La seg. funzione calcola la lista dei vertici di una cella con le
; rispettive coordinate a partire dalla lista dei simboli dei 
; vertici di una cella e da un poliedro. 
; Restituisce una lista del tipo:
; ((V0 (. . .)) (V1 (. . .) ...................)
;------------------------------------------------------------------

(defun get-cell-vertices-coord (vcell pol)
  (let ((vpol (kernel::polyhedron-VLIST pol))
         vcell-coord)
    (dolist (vertex vcell)
      (dolist (coord vpol)
        (if (equal (car coord) vertex)
            (setq vcell-coord 
               (append vcell-coord (list coord))))))
    vcell-coord))

;------------------------------------------------------------------
; La seg. funzione a partire da una lista di vertici e da una lista
; di vertici e coordinate del tipo:
; ((V0 (. . .)) (V1 (. . .) ...................)
; restituisce una lista dello stesso tipo dei soli vertici 
; contenuti in vlist.
;------------------------------------------------------------------

(defun get-coord-vlist (vlist clist)
  (let ((vlist-coord nil))
    (dolist (vertex vlist)
      (dolist (coord clist)
        (if (equal (car coord) vertex)
            (setq vlist-coord 
               (append vlist-coord (list coord))))))
    vlist-coord))


;------------------------------------------------------------
;La seg. funzione restituisce gli indici delle coord.
;baricentriche di un punto in base allo spigolo dato ed al
;simplesso cubico in esame.
;------------------------------------------------------------

(defun get-coord-baric-point (vertex edge c-simplex)
  (let* ((list-vertex (cubic-simplex-VERTICES c-simplex))
         (list-edges (cubic-simplex-EDGES c-simplex))
         (coord-baric (make-list (length list-edges)
                                 :initial-element 0))
          index1
          index2)
    (dotimes (i (length list-vertex))
      (if (equal (first edge) (first (nth i list-vertex)))
         (setq index1 i))
      (if (equal (second edge) (first (nth i list-vertex)))
         (setq index2 i)))

    (if (equal vertex (first edge))
      (progn 
         (setf (nth index1 coord-baric) 2)
         (setf (nth index2 coord-baric) 1)))
    (if (equal vertex (second edge))
       (progn
         (setf (nth index2 coord-baric) 2)
         (setf (nth index1 coord-baric) 1)))
    coord-baric))
         

;--------------------------------------------------------------------------
; La seg. funzione controlla se il simbolo dato in input e' il simbolo
; del centroide della cella. 
;--------------------------------------------------------------------------

(defun centroid-symbol? (symbol 
                         &aux YorN)
  (let ((end-symbol (subseq (reverse (string symbol)) 0 1)))
    (if (equal end-symbol "C")
      (setq YorN T)
      (setq YorN nil))
   YorN))    
