;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Il seg. modulo contiene le funzioni per triangolare le celle  ;;
;; di un complesso poliedrale in base al loro centroide          ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'cub_ext)

(provide 'step1)

(inpackage kernel)


;------------------------------------------------------------------
; La seg. funzione triangolarizza una cella di un certo poliedro  
; rispetto al suo centroide e crea la struttura di simplesso cubico
; per ogni triangolo. Restituisce quindi la cella con la hash-table 
; dei simplessi cubici.
; Vengono poi memorizzate le coordinate dei vertici nella
; hash-table della Bezier-net in quanto esse non vengono modificate
; dall'applicazione dell'algoritmo. 
; In ultimo vengono settate le adiacenze dei simplessi cubici
; interne alla cella.
;------------------------------------------------------------------
                          
(defun split-cell (cell pol)
  (let* ((coord (get-cell-vertices-coord (get-cell-vertices cell) pol))
         (simplexes-table (make-hash-table :test 'equal))
         (faces (get-cell-faces cell))
         (cell-id (get-cell-identifier cell))
         (centroid-id (make-centroid-symbol cell-id))
         (centroid (list centroid-id (calculate-centroid coord))) 
         (n (length coord)) )
     

    (dotimes (j n)  ;il num. di vertici di una cella coincide
                    ;               con il numero delle facce
      (let* ((cell-face (nth j faces))
             (identifier (make-cubic-simplex-symbol cell-id j))
             (edges (list cell-face 
                         (list (second cell-face) centroid-id)
                         (list centroid-id (first cell-face))))
             (vertices (append 
                          (get-coord-vlist cell-face coord) 
                          (list centroid)))
             (Bezier-net (make-hash-table :test 'equal))
             (adiacent-cubic-simplexes (make-list (length edges))) )


                                      ;;memorizza le coordinate dei vertici
                                      ;;dei simplessi, eccetto che il
                                      ;;centroide della cella, nella 
                                      ;;hash-table della Bezier-net.
        (setq Bezier-net 
           (dotimes (i (1- (length vertices)) Bezier-net)
             (let ((coord-baric (make-list (length vertices)
                                            :initial-element 0.0d0) ))

             (setf (nth i coord-baric) 3)  ;;3 e' il grado del simplesso 
             (setf 
                (gethash 
                   (make-id-coord-baric coord-baric) 
                    Bezier-net)
              (second (nth i vertices)) )) ))

        (setf 
          (gethash identifier simplexes-table) 
          (make-cubic-simplex identifier
                              edges
                              vertices
                              Bezier-net 
                              adiacent-cubic-simplexes)) ))

                         ;setta le adiacenze interne alla cella per i
                         ;simplessi cubici
    (setq simplexes-table (internal-adiacences simplexes-table))

                         ;memorizza la hash table dei simplessi
                         ; cubici nella cella
    (setf (convex-cell-CUBIC-SIMPLEXES cell) simplexes-table) 
                        
     cell))


;----------------------------------------------------------------------------
; Calcolo delle adiacenze interne dei simplessi cubici
;----------------------------------------------------------------------------

(defun internal-adiacences (simplexes-table)
  (let (edges1
        edges2)
    (maphash #'(lambda (id1 simplex1)
                 (setq edges1 (cubic-simplex-EDGES simplex1))
                 (maphash #'(lambda (id2 simplex2)
                     (if (not (equal id1 id2))
                         (progn
                           (setq edges2 (cubic-simplex-EDGES simplex2))
                           (dotimes (i (length edges1))
                             (dolist (x edges2)
                               (if (or (equal (nth i edges1) x)
                                       (equal (nth i edges1) (reverse x)))
                                   (setf (nth i 
                                           (cubic-simplex-ADIACENT-CUBIC-SIMPLEXES
                                              (gethash id1 simplexes-table)))
                                            id2))
                             )))))
                       simplexes-table))
      simplexes-table)
   simplexes-table))

;-------------------------------------------------------------------------
; La seg. funzione visita la struttura di complesso poliedrale
; ------------------------------------------------------------------------

(defun Split-Pol-Complex-with-Centroid (obj)
  (cond ((polyhedron?      obj) (Split-Pol-with-Centroid obj))
        ((polyhedra-complex? obj)
          (mapcar  #'(lambda (nodo) (Split-Pol-Complex-with-Centroid nodo)) obj)
        )
        ((polyhedral-node? obj)
         (progn
           (make-pol-node 
             (Split-Pol-Complex-with-Centroid (get-node-pol-complex obj))
             (get-node-matrix obj)
             (get-node-id obj) ))) ))


;-------------------------------------------------------------------------
; La seg. funzione visita le celle del poliedro dato in input
;-------------------------------------------------------------------------

(defun Split-Pol-with-Centroid (pol)
  (let ((cells (get-cells pol)))
        
    (maphash #'(lambda (id cell)
      (setf (gethash id cells) (split-cell cell pol)))
     cells)
                 ;; restituisce il poliedro con la nuova hash table
  pol ))


;-------------------------------------------------------------------------
; La seg. funzione esegue una copia del poliedro dato in input.
; E' utile quando si vogliono evitare i side-effects.
;-------------------------------------------------------------------------

(defun copy-pol (pol)
  (let ((newpol (make-empty-polyhedron)))
    (setf (polyhedron-M newpol) (polyhedron-M pol))
    (setf (polyhedron-N newpol) (polyhedron-N pol))
    (setf (polyhedron-K0 newpol) (polyhedron-K0 pol))
    (setf (polyhedron-KM newpol) (polyhedron-KM pol))
    (setf (polyhedron-SIGN newpol) (polyhedron-SIGN pol))
    (setf (polyhedron-VLIST newpol) (polyhedron-VLIST pol))
    (setf (polyhedron-SLIST newpol) (polyhedron-SLIST pol))
    (setf (polyhedron-INFO newpol) (polyhedron-INFO pol))
    (setf (polyhedron-REPRESENTATION newpol) 
          (polyhedron-REPRESENTATION pol))
    (setf (polyhedron-FACES newpol) (polyhedron-FACES pol))
    (setf (polyhedron-F newpol) (polyhedron-F pol))
    (setf (polyhedron-CONVEX-CELLS newpol) 
          (polyhedron-CONVEX-CELLS pol))
    (setf (polyhedron-C newpol) (polyhedron-C pol))
    (setf (polyhedron-KIND-OF-BRICK newpol) 
          (polyhedron-KIND-OF-BRICK pol))
    (setf (polyhedron-GENERAL-CELLS newpol) 
          (polyhedron-GENERAL-CELLS pol)) 
     newpol))
