; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;           Universita' di Roma "La Sapienza"                              
;
;       Dipartimento di Informatica e Sistemistica                         
;
;            Via Buonarroti 12 - 00185 Roma                                
;
;                                                                          
;
;                      CAD Group                                           
;                    
;                                                                          
;
; Project: SimpleXn                                                        
;
; Identification: Costruzione di complessi da vertici                      
;               
;                                                                          
;
;  Version: 1.0                                                            
;
;                                                                          
;
;  Author: G. Proietti                                                     
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; (1/9/93)
;
; (30/9/93) Eliminato side effect sulla lista di punti
;           Aggiunta possibilita' di non eseguire calcolo della matrice
;            di trasformazione
;           Aggiunta funzione per il calcolo dello schema di un convesso
;            coerente con le necessita' dell'algoritmo di triangolazione
;            di Von H.

; (21/1/94) Eliminato warning per Risc 6000 e variabile DIM nella funzione 
;           UPDATE. Aggiunta funzione
;GetListOfFacesWithVerticesAndNormalizedCoeff
;

;Contiene le funzioni che implementano, ad alto livello, l'algoritmo di
;Beneath-Beyond


(inpackage kernel)

(provide 'Beneath)

(require 'affines)
(use-package 'hyperplanes)

(defstruct ConvexHull
  Graph              ;Grafo di incidenza del poliedro convesso
  Root               ;Identificatore faccia di dimensione massima del pol.
  Empty              ;Identificatore faccia di dimensione -1 del pol.
  DimSpace           ;Dimensione dello spazio di immersione del poliedro
  Dim                ;Dimensione intrinseca del poliedro
  InternalPoint      ;Un punto interno al poliedro
  AffineSpace        ;Lista di iperpiani specificanti lo spazio affine 
                     ;contenente il poliedro.
  NVerts             ;Numero di vertici finora aggiunti
  Matrix             ;Matrice di trasformazione da aggiungere al nodo
  )

  
;Applicazione dell'algoritmo
(defun Beneath-Beyond (PointList &optional (CalculateMatrix t))
  (Reset-GraphNode-Counter)
  (setq PointList (Lex-Sort-with-Labels (copy-tree PointList)))
  (setq PointList (Lex-Purge-Equal-Verts PointList))
  (let* ((Convex (NewConvexHull))
         (FirstPoint (first PointList))
         )
    (if PointList
      (progn
        (setf (ConvexHull-Dimspace Convex) (length (second FirstPoint)))
;Dim spazio di immersione
        (Pyr-Update Convex nil FirstPoint) ;Primo punto a partire da convesso vuoto
        (CalcNewInternalPoint Convex (second FirstPoint)) ;Punto interno per verso degli iperpiani
        (mapcar #'(lambda (x) (Update Convex x)) (cdr PointList))
        (if (and CalculateMatrix (ConvexHull-AffineSpace Convex))
          ; Siamo in un sottospazio affine  
          (let ( (Translation (my-make-translation-matrix 
                               (ConvexHull-Dimspace Convex)  (second FirstPoint))))
            (setf (ConvexHull-Matrix Convex) 
                  (matlib::mat-product Translation
                                       (Transf-Matrix (ConvexHull-AffineSpace Convex))
                                       ))
            )
          )
        )
      )
    Convex
    )
  )


;(defun my-make-translation-matrix (n l)
;   (do ((tmat (make-idmat (+ n 1)))    
;        (j 1 (+ j 1)))
;       ((= j (1+ n)) tmat)
;     (setf (aref tmat j 0) (nth (1- j) l))))

(defun my-make-translation-matrix (n l)
  (let ((l-double (mapcar #'(lambda (n) (coerce n 'double-float)) l)))
    (do ((tmat (make-idmat (+ n 1)))    
         (j 1 (+ j 1)))
        ((= j (1+ n)) tmat)
      (setf (aref tmat j 0) (nth (1- j) l-double)))))

;Aggiunta di un nuovo punto al convesso.
;A seconda che il punto appartenga allo spazio affine gia' calcolato,
; nel qual caso soddisfa tutte le equazioni che definiscono lo
; spazio stesso, oppure no, nel qual caso non ne soddisfa almeno una,
; equazione che sara' utilizzata come base della piramide e nei calcoli di
; rotazione dei piani della facet e del nuovo spazio affine,
; in modo da eliminarla dal nuovo spazio affine.
(defun Update (Convex Point)
  (let* ( (Affine (ConvexHull-AffineSpace Convex)) 
          (Index  (Is-Point-Not-In-Affine? Affine (second Point)))
          )
    (CalcNewInternalPoint Convex (second  Point))
    (if Index 
      (Pyr-Update Convex Index Point)
      ;else
      (Not-Pyr-Update Convex Point)
      )
    )
  )

(defun CalcNewInternalPoint (Convex PointCoords)
  (let* ( (n (ConvexHull-NVerts Convex))
          (nn (1+ n))
          (p (ConvexHull-InternalPoint Convex))
          )
    (if (null p) (setf p (make-list (length PointCoords) :initial-element 0)))
    (setf (ConvexHull-InternalPoint Convex)
          (mapcar #'(lambda (x y) (/ (+ (* n x) y) nn)) p PointCoords)
          )
    (setf (ConvexHull-NVerts Convex) nn)
    )
  )

;Stampa struttura dati del guscio convesso
(defun PrintConvexHull (C)
  (print C)
  (PrintGraph (ConvexHull-Graph C))
  )


(defun NewConvexHull ()
  (let* ( (NewConv (make-ConvexHull :Graph (MakeEmptyGraph) :Dim -1 :NVerts 0))
          (Face (NewFace (ConvexHull-Graph NewConv)))
          )
    (setf (ConvexHull-Root NewConv) Face)
    (setf (ConvexHull-Empty NewConv) Face)
    NewConv
    )
  )


(defun GetVerticesOfFacets (Convex)
  (let ( (Graph       (ConvexHull-Graph Convex))
         (Root        (ConvexHull-Root  Convex))
         (DimPol      (ConvexHull-Dim   Convex))
         )
    (if (= DimPol 0)
      (list (list (first (GetVertex Graph Root))))
      ;else
      (mapcar #'(lambda (x) (GetVerticesOfFace Graph x (1- DimPol))) 
              (GetSubFaces Graph Root)
              )
      )
    )
  )

(defun GetAllVertices (Convex)
  (let ( (Graph (ConvexHull-Graph Convex)) )

    (mapcar #'(lambda (x) (FaceInfo-VID (GraphNode-Data (second (GetNode x Graph)))))
            (GetSuperFaces Graph (ConvexHull-Empty Convex))
            )
    )
  )

;Data una k-faccia, ritorna tutti i vertici che la individuano
(defun GetVerticesOfFace (Graph FaceID Dim)
  (let ( Res )
    (if (/= Dim 0)
      (dolist (sub (GetSubFaces Graph FaceID) Res)
        (setq Res (union Res (GetVerticesOfFace Graph sub (1- Dim))))
        )
      ;else
      (list (first (GetVertex Graph FaceID)))
      )
    )
  )


;
(defun Get-Scheme-Of-Convex (Convex &optional Graph FaceID ListRes Dim)
  ;(print FaceID)
  (let (Res
        )
    (if (null Graph)
      (progn
        (setq Graph   (ConvexHull-Graph Convex))
        (setq FaceID  (ConvexHull-Root Convex))
        (setq Dim     (ConvexHull-dim Convex))
        (setq ListRes (make-list (1+ Dim) :initial-element nil))
        (resetinduced Graph)
        ))
     
    (if (/= Dim 1)
      (progn
        (dolist (sub (GetSubFaces Graph FaceID))
            (setq Res (union Res (Get-Scheme-Of-Convex Convex Graph sub ListRes (1- Dim)))) 
          )
        (if (not (getinduced Graph FaceID))
          (progn
            (setf (nth Dim ListRes) (cons Res (nth Dim ListRes)))
            (setinduced Graph FaceID T)
            )
          )
        )
      ;else
      (progn
        (setq Res (mapcar #'(lambda (NodeID) (first (GetVertex Graph NodeID)))
                          (GetSubFaces Graph FaceID)))
        (if (not (getinduced Graph FaceID))
          (progn
            (setf (nth Dim ListRes) (cons Res (nth Dim ListRes)))
            (setinduced Graph FaceID T)
            )
          )
        )
      )
    (if (= Dim (ConvexHull-dim Convex))
      (cdr ListRes)
      (progn
        ;(print "------")
        Res
        )
      )
    )
  )


;-----------------------------------------------------------------------
;Ritorna la lista delle facce, ciascuna faccia espressa come una lista di
;due componenti: l'insieme dei vertici che vi giacciono ed i suoi coefficienti.
;La faccia e' nel formato (tNoto coeff1 ... coeffn)

(defun GetListOfFacesWithVerticesAndCoeff (Convex)
 (let ( (Graph (ConvexHull-Graph Convex)) )
    (mapcar #'(lambda (x)
                (list
                 (GetVerticesOfFace Graph x (1- (ConvexHull-Dim Convex)))
                 (let ( (eqn (GetHyperplane Graph x)))
                   (setq eqn (append (last eqn) (butlast eqn)))
                   )
                 )
                )
            (GetSubFaces Graph (ConvexHull-Root Convex))
            )
    )
 )

;-----------------------------------------------------------------------
;Ritorna la lista delle facce, ciascuna faccia espressa come una lista di
;due componenti: l'insieme dei vertici che vi giacciono ed i suoi coefficienti.
;La faccia e' nel formato (tNoto coeff1 ... coeffn) e gli stessi sono normalizzati.
;Il primo coefficiente non nullo e' positivo.

(defun GetListOfFacesWithVerticesAndNormalizedCoeff (Convex)
 (let ( (Graph (ConvexHull-Graph Convex)) )
    (mapcar #'(lambda (x)
                (list
                 (GetVerticesOfFace Graph x (1- (ConvexHull-Dim Convex)))
                 (let ( (eqn (GetHyperplane Graph x)))
                   (first (normalize-parameters 
                              (setq eqn (append (last eqn) (butlast eqn)))))
                   )
                 )
                )
            (GetSubFaces Graph (ConvexHull-Root Convex))
            )
    )
 )


;Dato un vertice, ritorna la lista delle k-facce che vi incidono
(defun GetFacesOnVertice (Graph ID k)
  (if (= k 1) (GetSuperFaces Graph ID)
      ;else
      (let ((temp (mapcar '(lambda (x) (GetFacesOnVertice Graph x (1- k)))
                          (GetSuperFaces Graph ID)))
            Res
            )
        (setq Res (car temp))
        (dolist (el (cdr temp) Res)
          (setq Res (union Res el))
          )
        )
      )
  )
