; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Costruzione di complessi da vertici                       ;               
;                                                                           ;
;  Version: 1.02                                                            ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; (28/9/93)

; In questa nuova versione e' stato eliminato il bug causato da facce
; coincidenti comuni a piu' celle di un unico poliedro.

;(4/10/93)  eliminato bug sul segno delle facce con adiacenza
;(24/1/94)  risolto problema segno delle facce

(inpackage kernel)

(require 'affines)
(use-package 'hyperplanes)

;-----------------------------------------------------
;A partire da un database di vertici, contenuto nella vlist, in cui ciascun
; vertice e' rappresentato dalla lista delle sue coordinate non omogenee
; , un database di celle convesse clist, ciascuna identificata dalla lista degli
; indici di vertici che la identificano ed una lista di poliedri plist, 
; ciascuno identificato dalla lista degli indici delle celle convesse che lo
; compongono, costruisce un complesso poliedrale con la rappresentazione 
; specificata dal simbolo rep. Default 'on-vertices.

(defun Generate-PolComplex (vlist clist plist &optional rep)
  (let ( (dim (length (car vlist))) 
         polyhedra-list
         )
    ; Aggiungo i nomi ai vertici, trasformando vlist in una lista associativa
    (setq vlist (name-vertices vlist))
    ; Costruisco una lista contenente tutti i poliedri
    (reset-polyhedral-node-counter)
    (dolist (pol plist)
      (push (Make-Polyhedral-Node-List-From-Vertices-And-Cells-Lists 
             vlist
             (Make-List-Of-Cells-Of-Polyhedron clist pol)
             rep
             )
            polyhedra-list)
      )
    ; Costruisco ora il complesso come lista di nodi, ognuno costituito
    ; da una matrice identita' , il poliedro ed il simbolo
    (setq dim (1+ dim))   ;dimensione matrice identita'
    (mapcar #'(lambda (x) 
               (if (polyhedral-node? x)
                 x
                 ;else
                 (if (symbol-equal rep 'on-faces) 
                   (list x (matlib::mat-identity dim) (advance-polyhedral-node-counter))                                   
                   ;else 
                   (list (matlib::mat-identity dim) x (advance-polyhedral-node-counter))                                   
                   )
                 )
               )
            polyhedra-list)
    )
  )
  

;-----------------------------------------------------------------------
; Data una lista di vertici, ritorna la lista stessa, in cui, ad ogni vertice,
; e' associato un nome.
(defun Name-Vertices (vlist)
  (let ( temp )
    (dotimes (i (length vlist) temp)
      (setq temp (append temp (list 
                               (list (make-vertex-symbol i)
                                     (nth i vlist))
                               )
                         )
            )
      )
    )
  )

;-------------------------
;A partire dal database delle celle e dalla lista di indici di celle
;del poliedro, costruisce una lista di liste, dove ciascuna cella e'
;identificata da un elemento, costituito dalla lista di indioci di vertici.
(defun Make-List-Of-Cells-Of-Polyhedron (c-list pol-cells-index)
  (mapcar #'(lambda (x) (nth x c-list)) pol-cells-index)
  )

;---------------------------
;A partire dal database di vertici etichettati VertList e dalla lista di 
;liste di indici dei vertici costituenti le celle CellsList, genera una 
;lista di nodi poliedrali, distinguendo due casi.
;Se tutte le celle sono a dimensione piena, la lista conterra' un solo 
;nodo, costituito, oltre che dalla matrice identita' e dall-etichetta, 
;dal poliedro formato dalle celle.
;Se vi sono invece celle a dimensione non piena, la lista conterra' come 
;primo elemento, se possibile, un nodo contenente il poliedro costituito 
;dall'insieme delle celle a dimensione piena e poi tanti nodi quante sono 
;le celle a dimensione non piena, essendo questi ultimi nodi costituiti da 
;una matrice di trasformazione affine necessaria all'immersione nello 
;spazio a dimensione piena, dal simbolo identificante il nodo poliedrale e 
;da un poliedro costituito dall'unica cella.
;Se rep e' pari ad 'on-faces avro' rappresentazione tramite facce.

(defun Make-Polyhedral-Node-List-From-Vertices-And-Cells-Lists (VertList CellsList rep)
  (let* ( (VerticiCelle (Transform-CellsList-in-Vertices-List-of-List VertList CellsList))
          (ConvHulls    (Make-List-Of-Convex-Hulls VerticiCelle))
          (ListeConv    (Separate-Entire-Dimensions-Convex-Hulls ConvHulls))
          Node-List
          Pol
          )
    (if (first ListeConv) 
      (setq Pol  (Make-Pol-From-ConvexHull-List VertList (first ListeConv) rep))
      )
    (if Pol ;Costruzione primo nodo poliedrale della lista
        (setq Node-List (make-pol-node pol 
                                       (matlib::mat-identity (1+ (get-npol Pol))) 
                                       (advance-polyhedral-node-counter)) 
              )
      )
    (if (second ListeConv)
      ;(if (not (null Node-List)) (setq Node-List (list Node-List)))
      (setq Node-List 
            (cons Node-List 
                    (Make-Nodes-From-SubDimension-ConvexHulls VertList (second ListeConv) rep)
                    )
            )
      )
    (setq Node-List (remove nil Node-List))  ;per disperazione
    (if (= (length Node-List) 1) (setq Node-List (car Node-List)))
    Node-List
    )
  )

;------------------------
;A partire dal database dei vertici etichettati VertList e da una lista di 
;strutture ConvexHull, ConvexHullList, definenti gusci convessi a dimensione 
;piena, costruisce un poliedro rappresentato tramite vertici. Ogni guscio convesso nella lista 
;corrispondera' ad una cella convessa del poliedro.
;rep indica il tipo di rappresentazione richiesta.

(defun Make-Pol-From-ConvexHull-List (VertList ConvexHullList rep)
  (let ( c-list     ;Lista celle convesse costruite
         v-list     ;lista vertici del poliedro
         f-list     ;lista facce del poliedro (solo se richiesta rappresentazione con facce)
         faces
         (dim    (ConvexHull-dim (first ConvexHullList)))
         (pol (make-empty-polyhedron))   ;il poliedro risultato
         )
    (reset-face-counter)
    (setq f-list (make-empty-face-database))
    ;Costruzione lista celle
    (dotimes (i (length ConvexHullList) )
      (setq c-list (append c-list (list (Build-Cell-From-ConvexHull (nth i ConvexHullList) i rep f-list))))
      )
    (Rebuild-Adiacens c-list rep) ;Costruzione relaz. adiacenza tra' celle
    (if (not (symbol-equal rep 'on-faces))
      (progn
        ;Costruzione lista vertici
        (dolist (cell c-list)
          (setq v-list (union v-list (kernel::convex-cell-vertices cell)))
          )
        ;Costruisce lista associativa contenente solo i vertici del poliedro
        (setq v-list (mapcar #'(lambda (x) (assoc x VertList)) v-list))
        ;Costruzione effettiva poliedro
        (mapcar #'(lambda (x) (put-vertex x pol)) v-list)
        (dolist (cella c-list)
          (setq pol (put-cell cella pol) )
          )
        (setf (kernel::polyhedron-REPRESENTATION pol) 'ON-VERTICES)
        )
      ;else con facce
      (progn
        ;Costruzione effettiva poliedro
        (setq Faces (Faces-List-for-Put-Faces-from-DataBase f-list))
        ;(pprint Faces)
        (setq Faces (reduce-Redundant-Faces-from-Cells-List c-list Faces))
        (put-faces Faces pol)
        (dolist (cella c-list)
          (setq pol (put-cell cella pol) )
          )
        (setf (kernel::polyhedron-REPRESENTATION pol) 'ON-FACES)
        )       
      )
    (set-npol pol dim)  
    (set-dpol pol dim)
    (set-#kd  pol)
    (set-#k0  pol)
    (set-#F  pol)
    (set-#C  pol)
    (setf (kernel::polyhedron-KIND-OF-BRICK pol)  'CONVEX-CELLS)
    pol
    )
  )


(defun Faces-List-for-Put-Faces-from-DataBase (DataBase)
  (mapcar #'(lambda (x) (list 
                        (first x)            ; simbolo
                        (second (second x))  ; equazione
                        )
            )
          DataBase)
  )

;------------------------
;A partire da una lista di strutture ConvexHull, ConvexHullList, definenti 
;gusci convessi a dimensione non piena, e dal 
;database di vertici etichettati VertList, crea una lista di nodi 
;poliedrali, tanti quanti sono i gusci convessi, in cui ciascun nodo 
;contiene un poliedro, rappresentato come specificato da rep.
(defun Make-Nodes-From-SubDimension-ConvexHulls (VertList ConvexHullList rep)
  (mapcar #'(lambda (x) (Make-Pol-Node-From-Single-ConvexHull-with-matrix VertList x rep)
             )
          ConvexHullList)
  )

;-------------------------
;Costruisce un singolo nodo poliedrale a partire da un guscio convesso a 
;dimensione non piena, 
;che viene trasformato nell'unica cella del poliedro, rappresentato come 
;specificato da rep.
;Calcola anche la trasformazione affine da associare al poliedro.
(defun Make-Pol-Node-From-Single-ConvexHull-with-matrix (VertList ConvexHull Rep)
  (let (Cell
        v-list
        f-list     ;lista facce del poliedro (solo se richiesta rappresentazione con facce)
        (dim (ConvexHull-dim ConvexHull))
        (pol (make-empty-polyhedron))
        Matrix
        pol-node
        )
    (reset-face-counter)
    (setq f-list (make-empty-face-database))
    ;Costruisce cella corrispondente al guscio convesso
    (setq Cell (Build-Cell-From-ConvexHull ConvexHull 0 Rep f-list))
    ;Lista associativa con vertici del poliedro
    (setq v-list (mapcar #'(lambda (x) (assoc x VertList)) (convex-cell-vertices Cell)))
    (if (not (symbol-equal rep 'on-faces))
      (progn
        ;Costruzione effettiva poliedro
        (put-vlist (Transform-Vertices-with-Matrix-Reducing-Coords 
                    v-list 
                    (matlib::mat-inv (ConvexHull-Matrix ConvexHull))
                    (- (ConvexHull-dimspace ConvexHull)
                       (ConvexHull-dim ConvexHull)))
                   pol)
          (setf (kernel::polyhedron-REPRESENTATION pol) 'ON-VERTICES)
          (setq Matrix (ConvexHull-Matrix ConvexHull))
          )
      ;else rappresentazione con facce
      (progn
        (if (> dim 0)
          (put-faces (Transform-Hyperplanes-with-Matrix-Reducing-Coeff 
                      (Faces-List-for-Put-Faces-from-DataBase f-list)
                      (ConvexHull-Matrix ConvexHull)
                      (- (ConvexHull-dimspace ConvexHull)
                         (ConvexHull-dim ConvexHull)))
                     pol)
          )
        (setf (kernel::polyhedron-REPRESENTATION pol) 'ON-FACES)
        (setq Matrix (matlib::mat-inv (ConvexHull-Matrix ConvexHull)))
        )      
      )
    (setq pol (put-cell Cell pol))
    (set-npol pol dim)  
    (set-dpol pol dim)
    (set-#kd  pol)
    (set-#k0  pol)
    (set-#F  pol)
    (set-#C  pol)
    (setf (kernel::polyhedron-KIND-OF-BRICK pol)  'CONVEX-CELLS)
    (setq pol-node (make-pol-node pol Matrix (advance-polyhedral-node-counter)))
    pol-node
    )
  )

;-----------------------------
;Trasforma le coordinate dei vertici in VertList, sottoponendole alla 
;trasformazione specificata da Matrix, con una moltiplicazione del vettore 
;a destra, previa estensione delle coordinate con la coordinata omogenea, 
;aggiunta in cima alle stesse.
;Ritorna poi solo le prime NCoords coordinate

(defun Transform-Vertices-with-Matrix-Reducing-Coords (VertList Matrix NCoords)
  (let ( (OmogCoords (mapcar #'(lambda (x) (list (first x) 
                                                (coerce (cons 1 (second x) ) 'vector)))
                             VertList))
         )
    (mapcar #'(lambda (x) (list (first x)
                               (cdr
                                (butlast 
                                 (coerce 
                                  (matlib::mat-vet-product Matrix (second x)) 
                                  'list)
                                 NCoords)
                                )))
            OmogCoords)     
    )
  )

;-----------------------------
;Trasforma i coefficienti degli iperpiani in HyperpList, sottoponendole alla 
;trasformazione specificata da Matrix, con una moltiplicazione del vettore 
;a sinistra.
;Ritorna poi solo i primi NCoords coefficienti.

(defun Transform-Hyperplanes-with-Matrix-Reducing-Coeff (HyperpList Matrix NCoords)
  (let ( (HyperpCoeffs (mapcar #'(lambda (x) (list (first x) 
                                                  (coerce (second x) 'vector)))
                               HyperpList))
         )
    (mapcar #'(lambda (x) (list (first x)
                               (butlast
                                (coerce 
                                 (matlib::vet-mat-product (second x) Matrix) 
                                 'list)
                                NCoords)
                               ))
            HyperpCoeffs)
    )
  )

;---------------------------
;Trasforma una lista di indici di celle in una lista di liste,
;in cui ciascun elemento costituisce la lista dei vertici della cella 
;stessa. 
(defun Transform-CellsList-in-Vertices-List-of-List (VertList CellsList)
  (let (ListOfListOfVertices)
    (dolist (Cell CellsList ListOfListOfVertices)
      ; prende i vertici della cella
      (setq ListOfListOfVertices (append ListOfListOfVertices 
                                         (list (mapcar #'(lambda (x) (nth x VertList)) Cell))
                                         )
            )
      )
    )
  )

;--------------------------
;Data una struttura ConvexHull, costruisce una cella convessa che ne abbia 
;la stessa topologia, estraendo le facce dal grafo di incidenza del 
;convesso.
;Il simbolo identificante la cella e' ricavato dal valore CellIndex.
;Se rep specifica una rappresentazione 'on-faces, utilizza ed aggiorna il 
;database delle facce del poliedro f-list.
(defun Build-Cell-From-ConvexHull (ConvexHull CellIndex rep f-list)
  (if (not (symbol-equal rep 'on-faces)) ; Costruzione cella per rappresentazione con vertici
    (let* ( (Faces  (kernel::GetVerticesOfFacets ConvexHull))
            )
      (make-convex-cell 
       (make-convex-cell-symbol CellIndex)
       '+
       (if (= (length Faces) 1) nil Faces)    
       (let ( Res ) (dolist (F Faces Res) (setq Res (union Res F))))
       (if (= (length Faces) 1) nil (make-list (length Faces) :initial-element 0))
       nil
       )
      )
    ;else Costruzione cella per rappresentazione con facce
    (let* ((Faces (GetListOfFacesWithVerticesAndNormalizedCoeff ConvexHull))
           (FacesSymbolsAndSigns (mapcar #'(lambda (x) (Make-a-Face-and-Update-DataBase 
                                               (first x) 
                                               (second x)
                                               (CalcFaceSign (second x) (ConvexHull-internalpoint ConvexHull) )  ;Per come escono da Beneath-Beyond
                                               f-list
                                               )
                                   )
                                 Faces
                                 )
                         )
           (FacesSymbols (mapcar #'car FacesSymbolsAndSigns))
           (Verts (GetAllVertices ConvexHull))
           (FacesForVerts (mapcar #'(lambda (x) ;debbo scartare le facce del poliedro ma non
                                               ;della cella.
                                     (intersection x FacesSymbols))
                                  (Transform-VIDs-in-List-Of-List-Of-FIDs Verts f-list))
                          )      
           (nfacce  (length FacesSymbols)) ;Numero di facce = numero di identificatori di facce
           (adj       (if (> nfacce 1) 
                        (make-list nfacce :initial-element 0)
                        ;else 0-dimensional case
                        nil)
                      )
           )
      (make-convex-cell  
       (make-convex-cell-symbol CellIndex)
       '+
       (if (> 1 (length FacesSymbolsAndSigns))
         (mapcar #'(lambda (x) (assoc x FacesSymbolsAndSigns)) FacesSymbols )
         ;else this is 0-dimensional case : faces nil
         nil
         )
       FacesForVerts
       adj
       nil) ) 
    )
  )

(defun CalcFaceSign (Plane Point)
  (if (> (Calcola-Equazione-last-Omog-coord Plane (cons 1 Point )) 0)
    '-
    ;else
    '+
    )
  )


(defun Calcola-Equazione-last-Omog-coord (Eqn Point)
  (let ( (acc 0) )
    (dotimes (i (length Eqn) acc)
      (setq acc (+ acc (* (nth i Eqn) (nth i Point))))
      )
    )
  )

;------------------------
;A partire da una lista di gusci convessi ListaConvexHull, ritorna una 
;lista di due elementi, in cui il primo elemento e' una lista 
;contenente tutti i gusci convessi di dimensione pari alla dimensione 
;dello spazio di immersione, mentre il secondo elemento e' una lista di 
;tutti i gusci convessi aventi dimensione minore della dimensione dello 
;spazio di immersione.
(defun Separate-Entire-Dimensions-Convex-Hulls ( ListaConvexHull )
  (let ( EntireDim
         SubDim
         )
    (dolist (ConvexHull ListaConvexHull (list EntireDim SubDim))
      (if (= (kernel::ConvexHull-Dim ConvexHull) (kernel::ConvexHull-DimSpace ConvexHull))
        (setq EntireDim (append EntireDim (list ConvexHull)))
        (setq SubDim (append SubDim (list ConvexHull)))
        )
      )
    )
  )

;---------------------------
;A partire da una lista di liste di vertici, calcola, applicando 
;l'algoritmo di Beneath-Beyond, la corrispondente lista di gusci convessi.
(defun Make-List-Of-Convex-Hulls ( Lista )
  (mapcar #'kernel::Beneath-Beyond Lista)
  )


;------------------------------
;Data una lista di strutture Convex-Cell, calcola tutte le adiacenze 
;tra' di esse, settando gli opportuni campi delle strutture stesse.
;In Rep deve essere riportato il genere di rappresentazione delle celle.
(defun Rebuild-Adiacens ( c-list Rep)
  (let ( cell1 
         cell2 
         remaining-cells
         )
    (dotimes (icell1 (length c-list))
      (setq cell1 (nth icell1 c-list))
      (setq remaining-cells (nthcdr (1+ icell1) c-list))
      (dotimes (icell2 (length remaining-cells) )
        (setq cell2 (nth icell2 remaining-cells))
        (if (symbol-equal rep 'on-faces)
          (Adjust-two-cells-adiacens-for-Faces-Rep cell1 cell2)
          (Adjust-two-cells-adiacens-for-Verts-Rep cell1 cell2)
          )
        )
      )
    )
  )

;------------------
;Data una lista di celle, elimina tutte le adiacenze
(defun Reset-Adiacens ( clist )
  (dolist (cell clist)
    (let ( (adj (CONVEX-CELL-ADJACENT-CONVEX-CELLS cell)) 
          )
      (dotimes (i (length adj) )
        (setf (nth i adj) 0))
      )
    )
  )

;------------------
;Data una coppia di celle, espresse come strutture convex-cell 
;rappresentate tramite vertici, ne calcola l'eventuale adiacenza
; e la memorizza nelgli appositi campi.
(defun Adjust-two-cells-adiacens-for-Verts-Rep (cell1 cell2)
  (let ( (faces1 (CONVEX-CELL-FACES cell1))
         (faces2 (CONVEX-CELL-FACES cell2))
         (adj1   (CONVEX-CELL-ADJACENT-CONVEX-CELLS cell1))
         (adj2   (CONVEX-CELL-ADJACENT-CONVEX-CELLS cell2))
         face1 face2 len
         )
    (dotimes (iface1 (length faces1))
      (setq face1 (nth iface1 faces1))
      (dotimes (iface2 (length faces2))
        (setq face2 (nth iface2 faces2))
        (setq len (length face2))
        (if (= (length face1) len) ;stesso numero di vertici
          (if (= (length (intersection face1 face2)) len) ; stessi vertici
            (progn
              (setf (nth iface2 adj2) (CONVEX-CELL-IDENTIFIER cell1))
              (setf (nth iface1 adj1) (CONVEX-CELL-IDENTIFIER cell2))
              )
            )
          )
        )
      )
    )
  )

;------------------
;Data una coppia di celle, espresse come strutture convex-cell 
;rappresentate tramite facce, ne calcola l'eventuale adiacenza
; e la memorizza nelgli appositi campi.
(defun Adjust-two-cells-adiacens-for-Faces-Rep (cell1 cell2)
  (let ( (faces1 (mapcar #'car (CONVEX-CELL-FACES cell1)))
         (faces2 (mapcar #'car (CONVEX-CELL-FACES cell2)))
         (adj1   (CONVEX-CELL-ADJACENT-CONVEX-CELLS cell1))
         (adj2   (CONVEX-CELL-ADJACENT-CONVEX-CELLS cell2))
         face1 face2
         )
    (dotimes (iface1 (length faces1))
      (setq face1 (nth iface1 faces1))
      (dotimes (iface2 (length faces2))
        (setq face2 (nth iface2 faces2))
        (if (equalp face1 face2) ;stesso identificatore di facce
          (progn
            (setf (nth iface2 adj2) (CONVEX-CELL-IDENTIFIER cell1))
            (setf (nth iface1 adj1) (CONVEX-CELL-IDENTIFIER cell2))
            )
          )
        )
      )
    )
  )
  
(defun Reduce-Redundant-Faces-from-Cells-List (c-list Faces)
  (let ( NewFaces )
    (setq NewFaces 
          (multiple-value-list 
           (kernel::Lex-Purge-Equal-Verts
            (kernel::Lex-Sort-with-Labels Faces))
           )
          )
    (Reassign-Faces-To-Cells c-list (second NewFaces))
    (first NewFaces)
    )
  )

(defun Reassign-Faces-To-Cells (Cells Faces-AList)
  (mapcar #'(lambda (Cell)
              (Reassign-Faces-To-Single-Cell Cell
                                             Faces-AList)
              ) Cells
          )
  )


(defun Reassign-Faces-To-Single-Cell (Cell Faces-AList)
  (setf (kernel::Convex-Cell-Vertices Cell)
         (mapcar #'(lambda (x)
                    (mapcar #'(lambda (y)
                               (let ( (Nuovo (assoc y Faces-AList) ))
                                 (if Nuovo
                                   (second Nuovo)
                                   y
                                   )
                                 )
                               ) x )) (kernel::Convex-Cell-Vertices Cell) )
         )
  (setf (kernel::Convex-Cell-Faces Cell)
        (mapcar #'(lambda (y)
                   (let ( (Nuovo (assoc (car y) Faces-AList)) )
                     (if Nuovo
                       (cons
                        (second Nuovo)
                        (cdr y)
                        )
                       ;else
                       y
                       )
                     )
                   ) (kernel::Convex-Cell-Faces Cell) )
        )
  Cell
  )





