; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Costruzione di complessi da vertici                       ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; (30/8/93)

(provide 'Gram_S)

(inpackage kernel)

(require 'matlib)

;--------------------------------------------
;A partire dai coefficienti degli iperpiani individuanti un
;sottospazio affine, ricava la matrice di trasformazione geometrica che 
;trasforma il sottospazio vettoriale associato al sottospazio affine in
;un sottospazio vettoriale di dimensione pari alla dimensione del
; sottospazio affine, ma contenente un numero di assi coordinati pari
; alla sua dimensione, con gli assi a partire dall'asse x.
;P. Es. un sottospazio affine di dimensione 1 origina una trasformazione
;che riporta all'asse x, un sottospazio affine di dimensione 2 origina 
;una trasformazione che riporta al piano xy .
(defun Transf-Matrix (AffineSpace)
  (Make-Generic-Transf-Matrix (mapcar 'butlast AffineSpace))
  )

;Dato un insieme di vettori, espresso come lista di liste, ritorna la
; matrice che consente il passaggio dai punti del sottospazio 
;vettoriale contenuto nei piani che hanno come vettori ortogonali i vettori 
;passati, espressi nelle coordinate della base canonica dello spazio,
;al sottospazio vettoriale coincidente con un asse, piano ,spazio, ecc.
;coordinato.

;che consente l'immersione dell'oggetto in uno spazio di dimensione superiore.

(defun Make-Generic-Transf-Matrix ( Vet-List )
  (let* ( (OrtoBase (Ortogonal-Base-Of-Space Vet-List))
          (Mat (listoflist-to-mat (Extend-Matrix-First-Row-and-Col (new-pl:transpose-matrix 
                (Swap-List OrtoBase (length Vet-List))))))
          )
     Mat
    )
  )


;Calcola una base ortogonale dello spazio Rn a partire da un insieme
; di vettori dati.
;Il calcolo viene eseguito nel seguente modo:
;All'insieme dei vettori, passato attraverso una lista, viene accodata 
;una base canonica dello spazio Rn.
;Il tutto viene poi ortogonalizzato con il metodo di Gram-Schmidt

(defun Ortogonal-Base-Of-Space ( Vet-List ) 
  (let* ( (Dim (length (car Vet-List)))
          )
    (setq Vet-List (append Vet-List
                         (Mat-To-ListOfList (MATLIB:Mat-Identity Dim))))
    (mapcar #'(lambda (x) (coerce x 'list)) 
            (Ortogon-Gram-Schmidt Vet-List))
    )
  )
  
;Test se un vettore e' nullo, o comunque tutte le sue
;componenti sono piu' piccole di un certo epsilon.
(defun IsVetNull? ( Vet )
  (dolist (Comp (coerce Vet 'list) T)
    (if (>= (abs Comp) 1E-12) (return-from IsVetNull? nil))
    )
  )


; Calcolo di un insieme di vettori indipendenti a partire da un
; insieme di vettori qualsiasi, secondo il metodo di 
; ortogonalizzazione di Gram-Schmidt.
; I vettori sono espressi come liste di componenti.
; I vettori che dovessero eventualmente risultare nulli vengono scartati
; e non inseriti nella lista risultante.

(defun Ortogon-Gram-Schmidt (Vet-List)
  (let ( u-list
         Temp
         )
    (dolist (vet Vet-List u-list)
      (setq Temp (1-Vet (coerce vet 'vector) u-list))
      (if (not (IsVetNull? Temp)) 
        (setq u-list (append u-list
                             (list Temp)
                             )
              )
        )
      )
    )
  )

;--------------------------------
; Calcolo del vettore ortogonale all'insieme dei vettori elencati in
; U-List,  partire dal vettore vec, mediante la formula di Gram-Schmidt.
;
(defun 1-Vet (vec U-List)
  (dolist (u U-List vec)
    (setq vec (matlib:vet-diff 
               vec 
               (matlib:vet-scalar-product 
                u
                (/ (matlib:vet-product vec u)
                   (matlib:vet-product u u)
                   )
                )
               )
          )
    )
  )


;Le seguenti due routines si occupano della conversione di
;rappresentazione di matrici da array a lista di liste e
;viceversa.
(defun Mat-To-ListOfList ( mat )
  (let* ( (dims (array-dimensions mat))
          (nRows (first dims))
          (nCols (second dims))
          Res
          )
    (dotimes (i nRows Res)
      (setq Res 
            (append Res 
                    (list 
                     (let ( (Temp (make-list nCols)) )
                       (dotimes (j nCols Temp) 
                         (setf (nth j Temp) (aref mat i j))
                         )
                       )
                     )
                    )
            )
      )
    )
  )

(defun ListOfList-To-Mat ( LOL )
  (let* ( (nRows (length LOL))
          (nCols (length (car LOL)))
          (Res   (make-array (list nRows nCols)))
          Temp
          )
    (dotimes (i nRows Res)
      (setq Temp (nth i LOL) )
      (dotimes (j nCols) 
        (setf (aref Res i j) (nth j Temp))
        )
      )
    )
  )

;-------------------------------------
;data una matrice rappresentata come lista di liste, la estende
; con una riga ed una colonna della matrice unitaria posta in
; ultima posizione.
;P. Es.
; '((2 3)          '((2 3 0)
;   (5 9))   --->    (5 9 0)
;                    (0 0 1))
;-------------------------------------
(defun Extend-Matrix (Mat)
  (let* ( (nCols     (length (car Mat)))
          (LastRow   (append (make-list  nCols :initial-element 0) (list 1)))
          )
    (append (mapcar #'(lambda (x) (append x (list 0))) Mat)
            (list LastRow))
    )
  )

(defun Extend-Matrix-First-Row-and-Col (Mat)
  (let* ( (nCols     (length (car Mat)))
          (FirstRow   (cons 1 (make-list  nCols :initial-element 0)))
          )
    (cons FirstRow 
          (mapcar #'(lambda (x) (cons 0 x )) Mat)
          )
    )
  )

;-------------------------------------
;sposta i primi n elementi della lista nell'ultima posizione
;-------------------------------------
(defun Swap-List (lista n)
  (append (nthcdr n lista) (GetFirst lista n))
  )

;-------------------------------------
;Preleva i primi n elementi della lista
;-------------------------------------
(defun GetFirst (lista n)
  (let ( (len (length lista))
         )
    (butlast lista (- len n))
    )
  )
