; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Proietti Giuseppe  (30/7/93)
;

(provide 'IncGraph)

(inpackage kernel)

(export '(Reset-GraphNode-Counter 
          MakeEmptyGraph 
          NewNode 
          PutNode
          GetNode
          DeleteNode
          CreatAdj
          DeleteAdj
          PrintGraph
          )
        )

;----------------------------------------------
; G. Proietti 
;
; Implementazione del tipo di dato GRAFO non orientato

; Questa implementazione e' rispondente alle particolari necessita'
; dell'algoritmo di Beneath-Beyond per la costruzione di gusci convessi
; a partire da insiemi di punti.
; La particolarita' di questi grafi e' quella di necessitare di due liste di
; nodi adiacenti, una per i nodi indicanti le superfacce ed una le subfacce
; della faccia relativa al nodo in esame.
; Si e' comunque scelto di fornire una struttura dati e delle operazioni
; in grado di gestire anche grafi di tipo ordinario.
; In questo caso viene utilizzata solamente la lista delle superfacce
; per gestire le adiacenze.
; 

; Operazioni implementate :
;
;  Creazione di grafo vuoto
;  Creazione di nodo
;  Inserzione in grafo
;  Cancellazione di nodo da grafo
;  Presa di un nodo da grafo
;  Creazione di adiacenza
;  Cancellazione di adiacenza
;  Stampa di grafo

; Operazioni specifiche per l'algoritmo di BB
;
; Creazione di una faccia
; Cancellazione di una faccia
; Creazione di una relazione Super-Sub faccia
; Rimozione di una relazione Super-Sub faccia
; 

; Genera simbolo di nodo del grafo di incidenza
;


;------ Strutture dati del grafo -----------

(defstruct GraphNode
  Super                  ;Lista superfacce  (adiacenti)
  Sub                    ;Lista subfacce    (aux)
  Data                   ;Dati del nodo     (data)
  )

(defstruct FaceInfo
  Color                  ;Colore della faccia
  Induced                ;Puntatore ausiliario a nodo
  Coeff                  ;Coefficienti piano facet
  TNote                  ;Termine noto piano facet
  Coords                 ;Coordinate del vertice
  VID                    ;Identificatore del vertice
  )

;------- Contatore per generazione identificatori nodi ----------

(defvar GraphNode-Counter  0)  ; Contatore per identificatori nodi

(defun Reset-GraphNode-Counter () (setq GraphNode-Counter 0))

; Generazione di nuovo identificatore di nodo
(defun MakeNodeId ()
  (setq GraphNode-Counter (1+ GraphNode-Counter))
  (intern (concatenate 
                'string "N" (integer->string GraphNode-Counter))
               )
  )

;------- Primitive per la manipolazione del grafo -----------

; Grafo vuoto
(defun MakeEmptyGraph () (make-hash-table))

; Nuovo Nodo
(defun NewNode ()
  (let ( (New (make-GraphNode))
         (Id  (MakeNodeId))
         )
    (list Id New)
    )
  )

; Immissione di nodo in grafo
(defun PutNode (Node Graph)
  (setf (gethash (first Node) Graph) (second Node))
  )

; Ricavo nodo da grafo
(defun GetNode (NodeId Graph)
  (let ( (Data (gethash NodeId Graph))
         )
    (if Data
      (list NodeId Data)
      )
    )
  )

; Cancellazione di nodo da grafo
(defun DeleteNode (Graph NodeId)
  (let ( (Data (gethash NodeId Graph))
         )
    (if Data
      (progn
        (mapcar #'(lambda (x) (DeleteAdj Graph NodeId x)) (GraphNode-Super Data))
        (remhash NodeId Graph)
        )
      )
    )
  )
  
; Creazione adiacenza tra' due nodi
(defun CreatAdj (Graph NodeId1 NodeId2)
  (let ( (Data1 (gethash NodeId1 Graph))
         (Data2 (gethash NodeId2 Graph))
         )
    (if (and Data1 Data2)
      (progn
        (setf (GraphNode-Super Data1) (cons NodeId2 (GraphNode-Super Data1)))
        (setf (GraphNode-Super Data2) (cons NodeId1 (GraphNode-Super Data2)))
        T
        )
      )
    )
  )

; Cancellazione adiacenza tra' due nodi
(defun DeleteAdj (Graph NodeId1 NodeId2)
  (let ( (Data1 (gethash NodeId1 Graph))
         (Data2 (gethash NodeId2 Graph))
         )
    (if (and Data1 Data2)
      (progn
        (setf (GraphNode-Super Data1) (remove NodeId2 (GraphNode-Super Data1) ) )
        (setf (GraphNode-Super Data2) (remove NodeId1 (GraphNode-Super Data2) ) )
        T
        )
      )
    )
  )

; Stampa grafo
(defun PrintGraph (Graph)
  (maphash #'(lambda (x y) (progn (print x) (print y))) Graph)
  )

;---- Funzioni specifiche per algoritmo di Beneath-Beyond -----

; --- Colori delle facce ---

(defun UnColor () (make-list 3 :initial-element nil) )

;Mappatura dei 'colori' che possono assumere le facce

;                     R   Y   B
(defun Red     () '(  T  nil nil ) )
(defun Yellow  () '( nil  T  nil ) )
(defun Blue    () '( nil nil  T  ) )
(defun Purple  () '(  T  nil  T  ) )
(defun Orange  () '(  T   T  nil ) )
(defun Green   () '( nil  T   T  ) )
(defun Brown   () '(  T   T   T  ) )

; Fusione di due colori 
(defun MergeColors (Color1 Color2)
  (let ((res (make-list 3 :initial-element nil)))
    (if (or (first  Color1) (first  Color2)) (setf (first res)   T))
    (if (or (second Color1) (second Color2)) (setf (second res)  T))
    (if (or (third  Color1) (third  Color2)) (setf (third res)   T))
    res
    )
  )

; Test di uguaglianza tra' due colori
(defun EqualColor? (Color1 Color2)
  (equal Color1 Color2)
  )

(defun ResetColor (Graph)
  (maphash #'(lambda (x y) (setf (FaceInfo-Color (GraphNode-Data y)) (UnColor))) Graph)
  )

(defun GetColor (Graph FaceID)
  (FaceInfo-Color (GraphNode-Data (gethash FaceID Graph)))
  )

(defun SetColor (Graph FaceID Color)
  (setf (FaceInfo-Color (GraphNode-Data (gethash FaceID Graph))) Color)
  )

;------ Funzioni manipolazione grafo di incidenza ---------

; Creazione di una faccia : ritorna identificatore faccia
(defun NewFace (Graph) 
  (let ((Node (NewNode)))
    (setf (GraphNode-Data (second Node)) (make-FaceInfo))
    (setf (FaceInfo-Color (GraphNode-Data (second Node))) (UnColor))
    (PutNode Node Graph)
    (first Node)
    )
  )

; Creazione di una relazione Super-Sub faccia : ritorna booleano
(defun Creat-Super-Sub-Face-Rel (Graph SuperFaceID SubFaceID)
  (let ( (Data1 (gethash SuperFaceID Graph))
         (Data2 (gethash SubFaceID   Graph))
         )
    (if (and Data1 Data2)
      (progn
        (setf (GraphNode-Sub   Data1) (cons SubFaceID   (GraphNode-Sub   Data1)))
        (setf (GraphNode-Super Data2) (cons SuperFaceID (GraphNode-Super Data2)))
        T
        )
      )
    )
  )

; Rimozione di una relazione Super-Sub faccia : boolean
; Cancellazione adiacenza tra' due nodi
(defun Delete-Super-Sub-Face-Rel (Graph SuperFaceID SubFaceID)
  (let ( (Data1 (gethash SuperFaceID Graph))
         (Data2 (gethash SubFaceID   Graph))
         )
    (if (and Data1 Data2)
      (progn
        (setf (GraphNode-Sub    Data1) (remove SubFaceID   (GraphNode-Sub Data1) ) )
        (setf (GraphNode-Super  Data2) (remove SuperFaceID (GraphNode-Super Data2) ) )
        T
        )
      )
    )
  )

;Superfacce di una faccia
(defun GetSuperfaces (Graph FaceID)
  (GraphNode-Super (gethash FaceID Graph))
  )

;Subfacce di una faccia
(defun GetSubfaces (Graph FaceID)
  (GraphNode-Sub (gethash FaceID Graph))
  )


; Cancellazione di faccia da grafo  : boolean
(defun DeleteFace (Graph FaceId)
  (let ( (Data (gethash FaceId Graph))
         )
    (if Data
      (progn
        (mapcar #'(lambda (x) (Delete-Super-Sub-Face-Rel Graph FaceId x)) (GraphNode-Sub   Data))
        (mapcar #'(lambda (x) (Delete-Super-Sub-Face-Rel Graph x FaceId)) (GraphNode-Super Data))
        (remhash FaceId Graph)
        )
      )
    )
  )

(defun ResetInduced (Graph)
  (maphash #'(lambda (x y) (setf (FaceInfo-Induced (GraphNode-Data y)) nil)) Graph)
  )

(defun SetInduced (Graph FaceID InducedID)
  (setf (FaceInfo-Induced (GraphNode-Data (gethash FaceID Graph))) InducedID)
  )

(defun GetInduced (Graph FaceID)
  (FaceInfo-Induced (GraphNode-Data (gethash FaceID Graph)))
  )

; Preleva i coefficienti ed il termine noto dell'iperpiano
; contenente la facet specificata.
; Se non li trova, allora non e' una facet e torna nil.
; Il risultato e' una lista il cui primo elemento e' la lista dei
; coefficienti ed il secondo il termine noto riportato al primo membro
; della disequazione.
(defun GetHyperPlane (Graph FaceID)
  (let* ( (data (GraphNode-Data (gethash FaceID Graph)))
          (Res (append (FaceInfo-Coeff Data) (list (FaceInfo-TNote Data))))
          )
    (if (or (null (first Res)) (null (second Res))) 
      nil
      Res
      )
    )
  )

; Setta i coefficienti ed il termine noto dell'iperpiano
; contenente la facet specificata.
; Il termine noto deve essere l'ultimo della lista ed essere riportato
; al primo membro.
(defun SetHyperPlane (Graph FaceID HyperPlane)
  (let ( (Data (GraphNode-Data (gethash FaceID Graph))) )
    (setf (FaceInfo-Coeff Data ) (butlast HyperPlane))
    (setf (FaceInfo-TNote Data ) (car (last HyperPlane)))
    )
  )

; Come sopra per i vertici.
(defun GetVertex (Graph FaceID)
  (let* ( (data (GraphNode-Data (gethash FaceID Graph)))
          (Res  (list (FaceInfo-VID Data) (FaceInfo-Coords Data)))
          )
    (if (or (null (first Res)) (null (second Res)) ) 
      nil
      Res
      )
    )
  )

(defun SetVertex (Graph FaceID Vertex)
  (let ( (Data (GraphNode-Data (gethash FaceID Graph))) )
    (setf (FaceInfo-Coords Data ) (second Vertex))
    (setf (FaceInfo-VID    Data ) (first Vertex))
    )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Test
;
;(Reset-GraphNode-Counter)
;(setq a (MakeEmptyGraph))
;(NewFace a)
;(ColorFacet a 'N1)