; -*- package: hyperplanes -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package hyperplanes                                       ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; note:
;   in this package are collected the functions useful to calculate
;   hyperplane parameters from its points, hyperplane intersection
;   and so on (many function are taken from the package utilities
;   written by F. Bernardini & V. Ferrucci).
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'hyperp)
(defpackage hyperplanes)
(inpackage hyperplanes)

;(require 'f_ext)
(use-package 'kernel)


(require 'lisp_ext)
(use-package 'lisp-extension)

(require 'matlib)
(use-package 'matlib)

(require 'new_pl)
(use-package 'new-pl)

(export '(calc-plane-parameters
          calc-plane-parameters-with-test
          calc-intersection-point
          same-vertex
          same-plane
          normalize-parameters))

;------------------------------------------------------------------------------
; This function returns the list of the parameters definig the hyperplane
; passing through the points listed in vertcood.
; INPUT: positive-sign_  the sign specifies the directin of the UP vector of
;                        the hyperplane ;
;        vertcoord_ is the list of the points of the plane. Each point is
;                   defined by the list of its coordinates;
;
; OUTPUT: cons of thwo elements:
;             - the first is a list of parameters (a0 a1 ... an) where the
;                hyperplane is defined by the implicit 
;                equation a0 + a1 x1 + ... + an xn = 0 .
;             - the second is the sign of the face
;------------------------------------------------------------------------------
(defun calc-plane-parameters-with-test
       (positive-sign vertcoord test-vertex)
  (let* ((n (length (first vertcoord)))
         (m (1- (length vertcoord)))
         (the-matrix (mapcar #'(lambda (coord)
                                (append (list 1) (copy-list coord)))
                             vertcoord))
         (positions (reduce-matrix 0 m 0 n n the-matrix))
         (result-column-index (catch 'end
                                (dotimes (i n n)
                                  (if (not (member i positions))
                                    (throw 'end i)))))
         (known-terms-vector nil))
    (dotimes (i  (1+ n))
      (when (/= i result-column-index)
        (push (nth result-column-index 
                   (nth (position i positions ) the-matrix))
              known-terms-vector)))
    (setf known-terms-vector (nreverse known-terms-vector))
    (push-nth  result-column-index -1 known-terms-vector)
    (setf known-terms-vector
          (first (normalize-parameters known-terms-vector)))
     (cons known-terms-vector (if  (minusp
                                   (+ (first known-terms-vector)
                                      (apply '+ (mapcar '* (cdr known-terms-vector)
                                                        test-vertex))))
                               positive-sign (opposit-sign positive-sign)))))

;------------------------------------------------------------------------------
; Returns a list of numbers proportionals to parameters-list but with the sum
; of squares of the numbers equal to 1.
;------------------------------------------------------------------------------
#+:APPLE
(defun normalize-parameters (parameters-list &optional (sign '+))
  ;(WITHOUT-FLOATING-POINT-UNDERFLOW
  (let ((normalized-parameters
         (mapcar 
          #'(lambda (x)
             (round-number
              (/ x 
                 (sqrt 
                   (apply 
                    '+ (mapcar '* parameters-list parameters-list))))))
          parameters-list)))
    (do ((i 0 (1+ i))
         (coord (first normalized-parameters)
                (nth i normalized-parameters)))
        ((not (zerop coord))
         (if (plusp coord)
           (cons normalized-parameters sign)
           (cons (mapcar '- normalized-parameters) (opposit-sign sign))))))) ;)

#-:APPLE
(defun normalize-parameters (parameters-list &optional (sign '+))
  (WITHOUT-FLOATING-POINT-UNDERFLOW(let ((normalized-parameters
         (mapcar 
          #'(lambda (x)
             (round-number
              (/ x 
                 (sqrt 
                   (apply 
                    '+ (mapcar '* parameters-list parameters-list))))))
          parameters-list)))
    (do ((i 0 (1+ i))
         (coord (first normalized-parameters)
                (nth i normalized-parameters)))
        ((not (zerop coord))
         (if (plusp coord)
           (cons normalized-parameters sign)
           (cons (mapcar '- normalized-parameters) (opposit-sign sign))))))))

;------------------------------------------------------------------------------
; This function returns the list of coordinates of the point intersection 
; of the n hyperplanes given in input.
; INPUT: hyperplanes-parameters_ list of the n hyperlanes. Each hyperlane
;                                is represented by the list of the n+1
;                                parameters of its implicit equation (the 
;                                first one is known term.
;
; OUTPUT: list of coordivates (x1 x2 ... xn) of the resulting point.
;------------------------------------------------------------------------------
(defun calc-intersection-point (hyperplanes-parameters)
  (let* ((n (1- (length (first hyperplanes-parameters))))
         (m (1- (length hyperplanes-parameters)))
         (matrix (copy-tree hyperplanes-parameters))
         (positions (reduce-matrix 0 m 1 n n matrix))
         (known-terms-vector nil))
    (dotimes (i n)
      (push (- (first (nth (position (- n i) positions ) matrix)))
            known-terms-vector))
    known-terms-vector))

;------------------------------------------------------------------------------
; Tests if the vertex has the same coordinates of coordlist
;------------------------------------------------------------------------------
(defun same-vertex (coord-list vertex)
  (if (every #'(lambda (x y)
                (almost-equal x y)) coord-list (get-coord vertex))
    (get-id vertex)))

;------------------------------------------------------------------------------
; Tests if the plane defined by parameters-list is coincident with plane.
; If the test is successful returns the plane-identifier.
;------------------------------------------------------------------------------
(defun same-plane (parameters-list plane)
  (if (every 'lisp-extension::almost-equal_ parameters-list (get-coord plane))
    (get-id plane)))
