; -*- package: lisp-extension -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package lisp-extension                                    ;
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'lisp_ext)

(defpackage lisp-extension)

(inpackage lisp-extension)

(export '(symbol-equal
          but-nth
          push-nth
          pop-function
          almost-equal
          swap
          revert-sign
          opposit-sign
          round-number
          corresponding-list
          copy-hash-table
          never-ending-list
          nested-dolist
          make-pathname**
          from-to
          list-of-lists-of-numbers?))

(require 'kernel)
(use-package 'kernel)

;------------------------------------------------------------------
; Returns a list of integers starting from "FROM" to "TO"
;------------------------------------------------------------------
(defun from-to (from to &optional elements-after-to)
  (if (> from to)
    elements-after-to
    (from-to from (1- to) (cons to elements-after-to))))

;------------------------------------------------------------------

;------------------------------------------------------------------
;
;------------------------------------------------------------------
(defvar formatting-string-for-round-number 
  (let ((number-of-digits (truncate (- (log **epsilon 10)))))
  (concatenate 'string 
               "~"
               (write-to-string (+ number-of-digits 10))
               "," 
               (write-to-string number-of-digits)
               "@E")))

;------------------------------------------------------------------
; Returns the number x rounded based on the value of the 
; value if the global constant **epsilon
;------------------------------------------------------------------
(defun round-number (x)
  (declare (special formatting-string-for-round-number))
  (if (< (abs x) **epsilon) 0
      (read-from-string
       (format nil formatting-string-for-round-number x))))

;------------------------------------------------------------------
; tests the two parameters sym-1 sym-2 are both symbols and
; if their names are equal
;    es:  (symbol-equal 'a 'a) --> t
;         (symbol-equal 'io::a 'kernel::a) --> t
;         (symbol-equal 'a 'b) --> nil
;         (symbol-equal '(a) '(a)) --> nil
;------------------------------------------------------------------
(defun symbol-equal (sym-1 sym-2)
  (and (symbolp sym-1)
       (symbolp sym-2)
       (string-equal sym-1 sym-2)))

;------------------------------------------------------------------
; Returns a list equal to list but without the nth element
;------------------------------------------------------------------
(defun but-nth (n list) 
  (append (subseq list 0 n) (nthcdr (1+ n) list)))

;------------------------------------------------------------------
; Pushes the element item to the list in position n.
;------------------------------------------------------------------
(defmacro push-nth (n item list)
  `(setf ,list 
         (append (subseq ,list 0 ,n)
                 (list ,item)
                 (nthcdr ,n ,list))))

;------------------------------------------------------------------
; Tests if the two numbers are almost equal 
;------------------------------------------------------------------
(defmacro almost-equal (number1 number2)
  `(< (abs (- ,number1 ,number2)) **epsilon))

;------------------------------------------------------------------------------
; Swaps the values of the two variables.
;------------------------------------------------------------------------------
(defmacro swap (var1 var2)
  `(let ((temp ,var1))
     (setf ,var1 ,var2
           ,var2 temp)))

;------------------------------------------------------------------------------
; Sets the value of var to the sign opposit to the current value of var
;------------------------------------------------------------------------------
(defmacro revert-sign (var)
  `(setf ,var (opposit-sign ,var)))

;------------------------------------------------------------------------------
; Returns the sign opposit to the given sign.
;------------------------------------------------------------------------------
(defun opposit-sign (sign)
  (if (eq sign '+) '- '+))

;------------------------------------------------------------------------------
; Creates a new list wiht each old name of the old-list replaced wth the new 
; name of the list new-names of the same posistion of the old name in old-names
;------------------------------------------------------------------------------
(defun corresponding-list (old-list old-names new-names)
  (mapcar #'(lambda (old-name)
             (nth (position old-name old-names) new-names))
         old-list))

;------------------------------------------------------------------------------
; Returns a hash table containing the same elements of the old-hash-table.
;------------------------------------------------------------------------------
(defun copy-hash-table (old-hash-table)
  (let ((new-hash-table (make-hash-table :test 'equal)))
    (maphash #'(lambda (key value)
                 (setf (gethash key new-hash-table) value))
             old-hash-table)
    new-hash-table))

;------------------------------------------------------------------------------
; Returns a never ending list composed of the elements given as parameters.
;------------------------------------------------------------------------------
(defun never-ending-list (&rest elements-list)
  (setf (cdr (last elements-list)) elements-list))

;------------------------------------------------------------------------------
; Function used for unification of make-pathname to allegro common lisp
; standard.
; In akcl the key-parameter requires a list of strings instead of a string
;------------------------------------------------------------------------------
(defun make-pathname** (&key host device directory name
                             type version defaults)
  #+:apple
 (make-pathname :host host :device device
                  :directory directory :name name
                  :type type :version version :defaults defaults)
  #+:akcl
 (make-pathname :host host :device device
                  :directory (list directory) :name name
                  :type type :version version :defaults defaults))
 
  
;-------------------------------------------------------------------------------------------
; This macro extends the standard dolist function to the case of a parametric
; number of nested dolist cicles.
; In particular, instead of a single (variable-name list [resultform]) before the body
; of the cycle ( {declaration}* {tag |form}*) you will give the list of var definitions:
; ((variable1-name list1 [resultform1])...(variablen-name listn [resultformn])).
;-------------------------------------------------------------------------------------------
(defmacro nested-dolist (var-definitions &rest body)
  (if (= 1 (length var-definitions))
      (append (cons 'dolist var-definitions)  body)
      (list 'dolist (car var-definitions)
            (append (list 'nested-dolist (cdr var-definitions))  body))))

;-------------------------------------------------------------------------------------------
; This function tests if the argument is a list of lists of numbers
;-------------------------------------------------------------------------------------------
(defun list-of-lists-of-numbers? (arg)
  (and (listp arg)
       (every #'(lambda (x) (and (listp x)
                                 (every (function numberp) x))) 
              arg)))
       

      
