; -*- package: matlib -*-
;-------------------------------------------------------------------------------
;           Universita' di Roma "La Sapienza"
;       Dipartimento di Informatica e Sistemistica
;            Via Buonarroti 12 - 00185 Roma
;
;                      CAD Group                 
; 
; Project: SimpleXn
; Identification: Package Matlib                         
; Date: 31-7-92
; Note: Extension of Matrix Functions Library
;             
; Version: 1.0
; Last modify: 
;              
; Author: V. Pascucci
;--------------------------------------------------------------------------------

(require 'matlib)
(provide 'matlib_e)        ;module name


(inpackage matlib)     ;package name

(export '(mat-identity
          mat-identity?
          mat-composition-for-power
          mat-product-extended
          mat-vet-product-extended
          rotate-rows
          move-rows-block))


(defvar $$number-type 'real)

;-----------------------------------------------------------------------------
; Crea e restituisce una matrice identita` di dimensione dim
;-----------------------------------------------------------------------------
(defun mat-identity (dim)
  (let ((mat (make-array (list dim dim)
                         :element-type $$number-type
                         :initial-element 0.0d0)))
    (dotimes (i dim)
      (setf (aref mat i i) 1))
    mat))

;-----------------------------------------------------------------------------
; Test per verificare se mat  una matrice identit
;-----------------------------------------------------------------------------
(defun mat-identity? (mat)
  (catch 'test
    (dotimes (i (array-dimension mat 0))
      (dotimes (j (array-dimension mat 0))
        (if (= i j)
          (when (/= (aref mat i j) 1) (throw 'test nil))
          (when (/= (aref mat i j) 0) (throw 'test nil)))))
    t))

;-----------------------------------------------------------------------------
; Composizione di due matrici secondo la regola specificata per l'esecuzione
; della potenza di due complessi poliedrali.
;-----------------------------------------------------------------------------
(defun mat-composition-for-power (M1 M2)
  (let* ((n1 (array-dimension m1 0))
         (n2 (array-dimension m2 0))
         (M (mat-identity (+ n1 n2 -1))))
    (dotimes (i (1- n1))
      (setf (aref M (1+ i) 0)(aref M1 (1+ i) 0))
      (dotimes (j (1- n1))
        (setf (aref M (1+ i) (1+ j))(aref M1 (1+ i) (1+ j)))))
    (dotimes (i (1- n2))
      (setf (aref M (+ n1 i) 0)(aref M2 (1+ i) 0))
      (dotimes (j (1- n2))
        (setf (aref M (+ n1 i) (+ n1 j))(aref M2 (1+ i) (1+ j)))))
    M))

;-----------------------------------------------------------------------------
; Opera una rotazione sulle ultime righe di una matrice partendo
; dalla i-esima.
; In particolare l'ultima riga prendera` il posto dell'i-esima, mentre
; le righe dll'i-esima in poi vengono spostate avanti di una posizione.
;-----------------------------------------------------------------------------
(defun rotate-rows (matrix d)
  (let* ((n1 (array-dimension matrix 0))
         (n2 (array-dimension matrix 1))
         tmp)
    (dotimes (j n2)
      (setf tmp (aref matrix (1- n1) j))
      (dotimes (i (- n1 d 1))
        (setf (aref matrix (- n1 1 i) j)(aref matrix (- n1 2 i) j)))
      (setf (aref matrix d j) tmp))
    matrix))

;-----------------------------------------------------------------------------
; Muove il blocco di righe dalla d+1 alla n in fondo alla matrice matrix 
;-----------------------------------------------------------------------------
(defun move-rows-block (matrix d n)
  (let* ((n1 (array-dimension matrix 0))
         (n2 (array-dimension matrix 1))
         (tmp (make-array (list (- n d) n2))))
    (dotimes (i (- n d))
      (dotimes (j n2)
        (setf (aref tmp i j)(aref matrix (+ d i) j))))
    (dotimes (i (- n1 n))
      (dotimes (j n2)
        (setf (aref matrix (+ d i) j)(aref matrix (+ n i) j))))
    (dotimes (i (- n d))
      (dotimes (j n2)
        (setf (aref matrix (+ (- n1 n) d i) j)(aref tmp i j))))
    matrix))

;-----------------------------------------------------------------------------
; Estensione del prodotto di matrici al caso di matrici con dimensioni
; communi non uguali.
; Se il parametro opzionale to-max vale true, alla matrice piu` piccola si
; aggiunge un blocco di matrice identita`, altrimenti la matrice piu` grande
; viene troncata.
;-----------------------------------------------------------------------------
(defun mat-product-extended (A B &optional (to-max t))
  (let* ((n (array-dimension A 0))
         (p1 (array-dimension A 1))
         (p2 (array-dimension B 0))
         (m (array-dimension B 1))
         (C (make-array (list n m)
                        :element-type 'real
                        :initial-element 0.0d0)))
    (dotimes (i n C)
      (dotimes (j m)
        (dotimes (k (if to-max (max p1 p2)(min p1 p2)))
          (setf (aref C i j)
                (+ (aref C i j) (* (if (< k p1)
                                     (aref A i k)
                                     (if (= i k) 1 0))
                                   (if (< k p2)
                                     (aref B k j)
                                     (if (= j k) 1 0))
                                   ))))))))
;-----------------------------------------------------------------------------
; Estensione del prodotto matrice vettore al caso di coppie con dimensioni 
; communi non uguali.
; Se il parametro opzionale to-max vale true, alla matrice piu` piccola si
; aggiunge un blocco di matrice identita`, altrimenti la matrice piu` grande
; viene troncata.
;-----------------------------------------------------------------------------
(defun mat-vet-product-extended (mat vet &optional (to-max t))
  (let* ((d  (array-dimension mat 0))
         (p1 (array-dimension mat 1))
         (p2 (array-dimension vet 0))
         (res (make-array (list d)
                            :element-type 'real
                            :initial-element 0.0d0)))
    (dotimes (j (if to-max (max p1 p2)(min p1 p2)) res)
      (dotimes (i d) 
        (setf (aref res i) 
              (+ (aref res i) (* (if (< j p2)
                                    (aref vet j) 0)
                                  (if (< j p1)
                                     (aref mat i j)
                                     (if (= i j) 1 0)))))))))
