; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Kernel                                            ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Estensione del nucleo di SimpleXn alla rappresentazione di complessi
; poliedrali .
; Un complesso poliedrale  una struttura a semireticolo con un solo nodo
; sorgente memorizzata tramite una lista i cui elementi sono delle strutture
; dette nodi poliedrali (sempre memorizzate in delle liste) composte da:
;
;          a) poliedro di dimensione intrinseca uguale alla dimensione
;              dello spazio di immersione
;             oppure complesso poliedrale;
;
;          b) matrice quadrata che definisce una trasformazione lineare
;             invertibile che trasforma il sottospazio Rk (prime k
;             componenti di Rn) in un opportuno sottospazio generico di Rn.
;             Le caratteristiche di questa matrice dipendono dal tipo
;             di rappresentazione del poliedro:
;                   sulle facce o sui vertici.
;             Il passaggio da una rappresentazione all'altra implica una
;             operazione di inversione della matrice (elemento che 
;             giustifica la nacessit di utilizzare una matrice quadrata)
;             Essendo di dimensione k lo spazio di immersione del poliedro
;             (o complesso poliedrale) a cui va applicata la trasformazione
;             si considera implicita una trasformazione di immersione 
;             dell'oggetto da Rk ad Rn.
;             Per rendere pi evidente il fatto che nel caso di 
;             rappresentazione a facce la matrice va postmoltiplicata per
;             i parametri che descrivono le facce, mentre per i vertici
;             la matrice va premoltiplicata per il vettore delle 
;             relative coordinate, i due elementi a) e b) verranno anche 
;             scambiati di ordine nel passaggio da una rappresentazione 
;             all'altra.
;
;          c) Nome opzionale (oppure assegnato automaticamente) del nodo. 
;             Si definiscon due classi di nomi:
;                  - i nomi rappresentati tramite simboli
;                    sono nomi assegnati arbitrariamente
;                    tramite meccanismi interni, e quindi sono nomi che
;                    possono essere alterati in modo altrettanto
;                    arbitrario;
;                  - i nomi rappresentati trimite stringhe sono nomi
;                    ai quali si considera abbinato un significato
;                    particolare (assegnato dall'utente o calcolato
;                    automaticamente) la cui alterazione dovr avvenire
;                    con particolare cautela.
;
;          d) Flag di visita del nodo del complesso poliedrale (non
;             presente nella definizione formale di complesso poliedrale
;             ma necessario nell'implementazione di visite di strutture 
;             non ad albero ma a semireticolo).          
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'conv_ext)

(provide 'pol_c_ex)

(inpackage kernel)

(export '(current-polyhedral-node-counter
	  advance-polyhedral-node-counter
	  reset-polyhedral-node-counter
	  previous-polyhedral-node-counter
	  make-polyhedral-node-symbol
	  
	  set-node-id set-node-pol-complex set-node-matrix
	  get-node-id get-node-pol-complex get-node-matrix
	  get-node-rep
	  get-pol-complex-rep
	  make-pol-node
	  
	  make-empty-polyhedra-complex
	  put-node-in-pol-complex
	  put-node-in-pol-complex-as-first
	  get-pol-complex-node-ids
	  get-npol-complex
	  get-dpol-complex
	  get-node
	  get-all-node
	  
	  polyhedral-node?
	  polyhedra-complex?
	  
	  build-pol-complex-cube
	  build-pol-complex-simplex
	  reduce-pol-complex
	  general-transform
	  rotate-pol-complex
	  translate-pol-complex
	  scale-pol-complex
	  shear-pol-complex
	  
	  power
	  skeleton
	  coerce-pol-complex
	  ))

(require 'matlib_e)
(use-package 'matlib)

(require 'new_pl)
(use-package 'new-pl)

;-----------------------------------------------------------------------------
; Controlla se pol-complex  un complesso poliedrale 
;-----------------------------------------------------------------------------
(defun polyhedra-complex? (pol-complex)
  (or (polyhedron? pol-complex)
      (and (listp pol-complex)
	   (every 'polyhedral-node? pol-complex))))

;-----------------------------------------------------------------------------
; Controlla se pol-node  un nodo poliedrale 
;-----------------------------------------------------------------------------
(defun polyhedral-node? (node)
  (and (listp node)
       (> (length node) 2)
       (or (and (arrayp (first node))
		(polyhedra-complex? (second node)))
	   (and (arrayp (second node))
		(polyhedra-complex? (first node))))
       (or (symbolp (third node))
	   (stringp (third node)))))

;-----------------------------------------------------------------------------
; creazione di un complesso poliedrale vuoto
;-----------------------------------------------------------------------------
(defun make-empty-polyhedra-complex ())

;-----------------------------------------------------------------------------
; Carattere usato come prefisso standard per l'assegnazione
;-----------------------------------------------------------------------------
(defvar $$polyhedral-node-prefix "PC")

;-----------------------------------------------------------------------------
;create the symbol PCi with the given i
; es: i=5 -> PC5 
;-----------------------------------------------------------------------------
(defun make-polyhedral-node-symbol (i)
  (if (< i 0)
    (intern 
     (concatenate 
      'string $$polyhedral-node-prefix (integer->string i)))
    (intern 
     (concatenate 
      'string $$polyhedral-node-prefix (integer->string i)))))

;-----------------------------------------------------------------------------
;  counter functions for convex cells
;-----------------------------------------------------------------------------
(defvar $$polyhedral-node-counter 0)

(defun current-polyhedral-node-counter ()
  (make-polyhedral-node-symbol $$polyhedral-node-counter))

(defun advance-polyhedral-node-counter ()
  (prog1
    (make-polyhedral-node-symbol $$polyhedral-node-counter)
    (setq $$polyhedral-node-counter 
	  (+ $$polyhedral-node-counter 1))))

(defun reset-polyhedral-node-counter (&optional (n 0))
  (setq $$polyhedral-node-counter  n))

(defun previous-polyhedral-node-counter ()
  (make-polyhedral-node-symbol 
   (- $$polyhedral-node-counter 1)))

;-----------------------------------------------------------------------------
; Restitusce la dimensione dello spazio di immersione del complesso poliedrale
;-----------------------------------------------------------------------------
(defun get-npol-complex (pol-complex)
  (if pol-complex
    (if (polyhedron? pol-complex)
      (get-npol pol-complex)
      (1- (array-dimension
	   (get-node-matrix (first pol-complex)) 0)))
    0))

;-----------------------------------------------------------------------------
; Restitusce la dimensione intrinseca del complesso poliedrale
;-----------------------------------------------------------------------------
(defun  get-dpol-complex (pol-complex)
  (if pol-complex
    (if (polyhedron? pol-complex)
      (get-dpol pol-complex)
      (get-dpol-complex
       (get-node-pol-complex (first pol-complex))))))

;-----------------------------------------------------------------------------
; Restitusce la dimensione dello spazio di immersione del complesso poliedrale
;-----------------------------------------------------------------------------
(defun get-pol-complex-rep (pol-complex)
  (if (polyhedron? pol-complex)
    (get-pol-rep pol-complex)
    (get-node-rep  (first pol-complex))))

;-----------------------------------------------------------------------------
; Restitusce l'identificatore del nodo poliedrale
;-----------------------------------------------------------------------------
(defun get-node-id (node) (third node))

;-----------------------------------------------------------------------------
; Restitusce il tipo di rappresentazione su cui si basa il nodo polyhedrale
;-----------------------------------------------------------------------------
(defun get-node-rep (node) 
  (if (arrayp (first node))
    'on-vertices
    'on-faces))

;-----------------------------------------------------------------------------
; Restitusce il poliedro del nodo poliedrale
;-----------------------------------------------------------------------------
(defun get-node-pol-complex (node) 
  (if (arrayp (first node))
    (second node)
    (first node)))

;-----------------------------------------------------------------------------
; Restitusce la matrice di trasformazione del nodo poliedrale
;-----------------------------------------------------------------------------
(defun get-node-matrix (node) 
  (if (arrayp (first node))
    (first node)
    (second node)))

;-----------------------------------------------------------------------------
; Restitusce il nodo del complesso poliedrale di identificatore node-id
;-----------------------------------------------------------------------------
(defun get-node (pol-complex &optional node-id)
  (find node-id pol-complex 
	:test #'(lambda (id node) (if node-id (equalp id (get-node-id node)) t))))

;-----------------------------------------------------------------------------
; Restitusce il nodo del complesso poliedrale di identificatore node-id
;-----------------------------------------------------------------------------
;(defun get-all-nodes (pol-complex node-id recursions)
;  (let ((nodes nil))
;    (dolist (node pol-complex)
;      
;(find node-id pol-complex 
;        :test #'(lambda (id node) (equalp id (get-node-id node)))))
;
;-----------------------------------------------------------------------------
; Assegna l'identificatore new-id al nodo poliedrale
;-----------------------------------------------------------------------------
(defun set-node-id (node new-id) 
  (setf (third node) new-id))

;-----------------------------------------------------------------------------
; Assegna il complesso poliedrale new-pol-complex al nodo poliedrale node
;-----------------------------------------------------------------------------
(defun set-node-pol-complex (node new-pol-complex) 
  (if (arrayp (first node))
    (setf (second node) new-pol-complex)
    (setf (first node) new-pol-complex)))

;-----------------------------------------------------------------------------
; Assegna la matrice di trasformazione new-matrix al nodo poliedrale node
;-----------------------------------------------------------------------------
(defun set-node-matrix (node new-matrix) 
  (if (arrayp (first node))
    (setf (first node) new-matrix)
    (setf (second node) new-matrix)))

;-----------------------------------------------------------------------------
; Creazione di un nodo di un complesso poliedrale come composizione di
; un poliedro (complesso poliedrale) una matrice di trasformazione
; un nome.
;   - Il poliedro (complesso poliedrale)  l'unico parametro che
;     deve essere necessariamente fornito in input, come primo
;     parametro;
;     il poliedro deve avere dimensione intrinseca uguale allo spazio
;     di immersione.
;   - Il secondo parametro debe essere la matrice di trasformazione (di
;     default matrice identit) o la di mensione dello spazio di immersione
;     (di default pari alla dimensione dello spazio di immersione del
;     poliedro).
;     Il passaggio del parametro nil equivale alla richiesta del
;     calcolo dei parametri di default.
;   - Il terzo parametro  l'identificatore del nodo (di default si assume 
;     la chiamata alla funzione make-polyhedral-node-symbol
;     con parametro 0.
;-----------------------------------------------------------------------------
(defun make-pol-node (pol &optional matrix id)
  (let ((representation (get-pol-complex-rep pol))
	(matrix (or matrix
		    (mat-identity (1+ (get-npol-complex pol)))))
	(id (or id (make-polyhedral-node-symbol 0))))
    (cond 
     ((symbol-equal representation 'on-faces)
      (list pol matrix id))
     ((symbol-equal representation 'on-vertices)
      (list matrix pol id)))))

;-----------------------------------------------------------------------------
; Inserisce in nodo poliedrale in un complesso 
; Il primo parametro  il nuovo nodo da inserire.
; Il secondo  il complesso poliedrale in cui deve essere inserito.
; Se il terzo parametro vale t il nuovo nodo andr a sostituire l'eventuale
; nodo con stesso identificatote.
; Se il terzo parametro vale nil (valore di default)
;-----------------------------------------------------------------------------
(defmacro put-node-in-pol-complex (node  pol-complex 
					 &optional (replace nil))
  `(let* ((the-node ,node)
	  (current-id (get-node-id the-node))
	  (same-id-node (get-node ,pol-complex current-id)))
     (if same-id-node
       (if ,replace
	 (progn 
	   (set-node-matrix       same-id-node (get-node-matrix      the-node))
	   (set-node-pol-complex  same-id-node (get-node-pol-complex the-node)))
	 (if (symbolp current-id)
	   (do* ((illegal-ids  (get-pol-complex-node-ids ,pol-complex 0))
		 (dummy  (reset-polyhedral-node-counter))
		 (new-id (advance-polyhedral-node-counter)
			 (advance-polyhedral-node-counter)))
		((not (member new-id illegal-ids))
		 (set-node-id the-node new-id)
		 (setf ,pol-complex (cons the-node ,pol-complex)))
	     (declare (special dummy)))
	   (error-handler " Polyhedral node identifier already used in the polyhedral complex "
			  " put-node-in-pol-complex ")))
       (setf ,pol-complex (cons the-node ,pol-complex)))
     (values)))


;-----------------------------------------------------------------------------
; Inserisce un nodo poliedrale in un complesso come prima istanza
; Il primo parametro  il nuovo nodo da inserire.
; Il secondo  il complesso poliedrale in cui deve essere inserito.
; Se il terzo parametro vale t il nuovo nodo andr a sostituire l'eventuale
; nodo con stesso identificatote.
; Se il terzo parametro vale nil (valore di default)
;-----------------------------------------------------------------------------
(defmacro put-node-in-pol-complex-as-first (node  pol-complex 
					     &optional (replace nil))
  `(let* ((the-node ,node)
	  (current-id (get-node-id the-node))
	  (same-id-node (get-node ,pol-complex current-id)))
     (if same-id-node
       (if ,replace
	 (progn 
	   (set-node-matrix       same-id-node (get-node-matrix      the-node))
	   (set-node-pol-complex  same-id-node (get-node-pol-complex the-node)))
	 (if (symbolp current-id)
	   (do* ((illegal-ids  (get-pol-complex-node-ids ,pol-complex 0))
		 (dummy  (reset-polyhedral-node-counter))
		 (new-id (advance-polyhedral-node-counter)
			 (advance-polyhedral-node-counter)))
		((not (member new-id illegal-ids))
		 (set-node-id the-node new-id)
		 (setf ,pol-complex (cons the-node ,pol-complex)))
	     (declare (special dummy)))
	   (error-handler " Polyhedral node identifier already used in the polyhedral complex "
			  " put-node-in-pol-complex ")))
       (setf ,pol-complex (cons the-node ,pol-complex)))
     (values)))

;-----------------------------------------------------------------------------
; Restitusce la lista dei nomi dei nodi di un complesso poliedrale.
; Si innesca un meccanismo di ricerca ricorsiva che scende in profondit
; di un numero di livelli pari al valore del parametro recursions.
; Se recursions vale nil (valore di default) l'albero che rappresenta il
; complesso poliedrale viene visitato tutto.
;-----------------------------------------------------------------------------
(defun get-pol-complex-node-ids (pol-complex &optional recursions)
  (when (not (polyhedron?  pol-complex))
    (let ((identifiers nil))
      (mapcar #'(lambda (identifier-to-add)
		  (setf identifiers 
			(adjoin identifier-to-add identifiers :test 'equal)))
	      (mapcar 'get-node-id pol-complex))
      (if recursions
	(when (plusp recursions)
	  (let ((identifiers (mapcar 'get-node-id pol-complex)))
	    (mapcar #'(lambda (identifiers-to-add)
			(setf identifiers 
			      (union identifiers identifiers-to-add :test 'equal)))
		    (mapcar 'get-pol-complex-node-ids 
			    (mapcar 'get-node-pol-complex pol-complex)
			    (never-ending-list (1- recursions))))
	    identifiers))
	(mapcar #'(lambda (identifiers-to-add)
		    (setf identifiers 
			  (union identifiers identifiers-to-add :test 'equal)))
		(mapcar 'get-pol-complex-node-ids 
			(mapcar 'get-node-pol-complex pol-complex))))
      identifiers)))

;-----------------------------------------------------------------------------
; Creazione di un complesso poliedrale formato da un solo simplesso standard
;-----------------------------------------------------------------------------
(defun build-pol-complex-cube (n)
  (let ((the-cube (build-cube n))
	(the-complex (make-empty-polyhedra-complex)))
    (put-node-in-pol-complex (make-pol-node the-cube)
			     the-complex)
    the-complex))

;-----------------------------------------------------------------------------
; Creazione di un complesso poliedrale formato da un solo cubo standard
;-----------------------------------------------------------------------------
(defun build-pol-complex-simplex (n)
  (let ((the-simplex (build-simplex n))
	(the-complex (make-empty-polyhedra-complex)))
    (put-node-in-pol-complex (make-pol-node the-simplex)
			     the-complex)
    the-complex))


;-----------------------------------------------------------------------------
; Ricava il parametro da passare alla funzione di estrusione stright
; per operare come la potenza il cui esponenete sia il poliedro di
; dimensione intrinseca e dimesione dello spazio di immersione pari a 1.
;-----------------------------------------------------------------------------
(defun reduce-P1_1-to-extrusion-parameter (pol)
  (let ((pol-faces (sort (mapcar #'(lambda (face)
				     (let ((coord (get-coord face)))
				       (list (get-id face) (- (/ (first coord)
								 (second coord))))))
				 (get-flist pol))
			 #'(lambda (f1 f2) (< (second f1) (second f2)))))
	(cells (get-clist pol))
	(extrusion-parameter nil))
    (dolist (cell cells)
      (let ((faces (get-cell-faces cell))
	    (adjacences (get-cell-adjacences cell)))
	(when  (not (symbolp (first adjacences)))
	  (setf (first (nth (position
			     (caar faces)
			     pol-faces
			     :test #'(lambda (f1 f2) 
				       (eql f1
					    (first f2)))) pol-faces)) 0))
	(when  (not (symbolp (second adjacences)))
	  (setf (first (nth (position
			     (caadr faces)
			     pol-faces
			     :test #'(lambda (f1 f2) 
				       (eql f1
					    (first f2)))) pol-faces)) 0))))
    (setf pol-faces (nreverse pol-faces))
    (do ((val (pop pol-faces)(pop pol-faces)))
	((null val))
      (push (list (second val)) extrusion-parameter)
      (do ((val (pop pol-faces)(pop pol-faces)))
	  ((not (symbolp (first val)))
	   (push (second val) (first extrusion-parameter)))
	(push (second val) (first extrusion-parameter))))
    extrusion-parameter))

;-----------------------------------------------------------------------------
; Potenza di un poliedro per un poliedro
;-----------------------------------------------------------------------------
(defun power-pol-for-pol (base exponent)
  (declare (special known-powers))
  (if (= 0 (get-npol exponent))
    base
    (cdr (or (assoc (cons base exponent) known-powers :test #'equalp)
	     (car (push (cons (cons base exponent)
			 (general-extrude base
					  (REDUCE-P1_1-TO-EXTRUSION-PARAMETER exponent)))
		   known-powers))))))

;-----------------------------------------------------------------------------
; Potenza di un nodo poliedrale per un poliedro
;-----------------------------------------------------------------------------
(defun power-node-for-pol (base exponent)
  (let* ((n (get-npol exponent))
	 (i (mat-identity (1+ n)))
	 (pol (get-node-pol-complex base))
	 (node-matrix (get-node-matrix base))
	 (result-matrix 
	  (mat-composition-for-power (get-node-matrix base)i))
	 )
    (make-pol-node (if (polyhedron? pol)
		     (power-pol-for-pol pol exponent)
		     (power-complex-for-pol pol exponent))
		   (if (> (array-dimension node-matrix 0)
			  (1+ (get-npol-complex pol)))
		     (rotate-rows result-matrix (1+ n))
                     

		     result-matrix)
		   (get-node-id base))))

;(defun power-node-for-pol (base exponent) (print '-------)
;  (let* ((n (get-npol exponent))
;         (i (mat-identity (1+ n)))
;         (pol (get-node-pol-complex base))
;         (node-matrix (get-node-matrix base))
;         (n-dim (print (array-dimension node-matrix 0)))
;         (d-dim (print (get-npol-complex pol)))
;         (result-matrix 
;          (mat-composition-for-power (get-node-matrix base) i))
;         (n1-dim (array-dimension result-matrix 0)))
;    (make-pol-node (if (polyhedron? pol)
;                     (power-pol-for-pol pol exponent)
;                     (power-complex-for-pol pol exponent))
;                   (if (< d-dim n-dim n1-dim)
;                     (progn 
;                       (print (move-rows-block  result-matrix
;                                                d-dim
;                                               n-dim)))
;                     result-matrix)
;                   (get-node-id base))))

;-----------------------------------------------------------------------------
; Potenza di un complesso poliedrale per un poliedro
;-----------------------------------------------------------------------------
(defun power-complex-for-pol (base exponent)
  (mapcar 'power-node-for-pol
	  base
	  (never-ending-list exponent)))

(defun power-complex-for-pol (base exponent)
  (declare (special known-powers))
  (cdr (or (assoc (cons base exponent) known-powers :test #'equalp)
	   (car (push (cons (cons base exponent)
			    (mapcar 'power-node-for-pol
				    base
				    (never-ending-list exponent)))
		      known-powers)))))

;-----------------------------------------------------------------------------
; Potenza di un complesso poliedrale per un nodo poliedrale
;-----------------------------------------------------------------------------
(defun power-complex-for-node (base exponent)
  (let* ((n (get-npol-complex base))
	 (i (mat-identity (1+ n)))
	 (pol (get-node-pol-complex exponent)))
    (if (listp pol)
      (make-pol-node (power-complex-for-complex base pol)
		     (mat-composition-for-power i 
						(get-node-matrix exponent))
		     (get-node-id exponent))
      (make-pol-node (power-complex-for-pol base pol)
		     (mat-composition-for-power i 
						(get-node-matrix exponent))
		     (get-node-id exponent)))))

;-----------------------------------------------------------------------------
; Potenza di un complesso poliedrale per un complesso poliedrale
;-----------------------------------------------------------------------------
(defun power-complex-for-complex (base exponent)
  (mapcar 'power-complex-for-node 
	  (never-ending-list base)
	  exponent))

(defun power-complex-for-complex (base exponent)
  (declare (special known-powers))
  
  (cdr (or (assoc (cons base exponent) known-powers :test #'equalp)
	   (car (push (cons (cons base exponent)
			    (mapcar 'power-complex-for-node 
				    (never-ending-list base)
				    exponent))
		      known-powers)))))

;-----------------------------------------------------------------------------
; Innesca il calcolo della potenza di un complesso poliedrale
; per un complesso poliedrale
;-----------------------------------------------------------------------------
(defun power (base exponent)
  (setf base (coerce-pol-complex base 'on-faces))
  (setf exponent (coerce-pol-complex exponent 'on-faces))
  (let ((known-powers nil))
    (declare (special known-powers))
    (power-complex-for-complex base exponent)))

;-----------------------------------------------------------------------------
; Eliminazione dei nodi superflui di un complesso poliedrale
;  - i nodi di primo livello sono superflui se hanno matrice di trasformazione
;    identi di dimensione pari allo spazio di immersione del suo complesso
;    poliedrale
;  - i nodi intermedi sono superflui se hanno matrice di trasformazione
;    identi
;-----------------------------------------------------------------------------
(defun reduce-pol-complex (pol-complex)
  (apply 'nconc (mapcar 'reduce-top-node pol-complex)))

(defun reduce-top-node (node)
  (let ((the-matrix (get-node-matrix node))
	(pol-complex (get-node-pol-complex node))
	(identifier (get-node-id node)))
    (if (polyhedron? pol-complex) (list node)
	(if (and (symbolp identifier)
		 (= (array-dimension the-matrix 0)
		    (get-npol-complex pol-complex))
		 (mat-identity? the-matrix))
	  (reduce-pol-complex pol-complex)
	  (list (make-pol-node
		 (reduce-intermediate-pol-complex pol-complex)
		 the-matrix
		 identifier))))))

(defun reduce-intermediate-pol-complex (pol-complex)
  (if (= 1 (length pol-complex))
    (let* ((node (first pol-complex))
	   (the-matrix (get-node-matrix node))
	   (pol (get-node-pol-complex node))
	   (identifier (get-node-id node)))
      (and (symbolp identifier)
	   (mat-identity? the-matrix))
      (if (polyhedron? pol)
	pol
	(reduce-intermediate-pol-complex pol)))
    (apply 'nconc (mapcar 'reduce-node pol-complex))))

(defun reduce-node (node)
  (let ((the-matrix (get-node-matrix node))
	(pol-complex (get-node-pol-complex node))
	(identifier (get-node-id node)))
    (if (and (symbolp identifier)
	     (mat-identity? the-matrix))
      (if (polyhedron? pol-complex)
	pol-complex
	(reduce-intermediate-pol-complex pol-complex))
      (list (make-pol-node
	     (if (polyhedron? pol-complex)
	       pol-complex
	       (reduce-intermediate-pol-complex pol-complex))
	     the-matrix
	     identifier)))))

(defun reduce-pol-complex (pol-complex)
  (if (polyhedron? pol-complex)
    pol-complex
    (apply #'append (mapcar 'reduce-pol-node pol-complex))))

(defun reduce-pol-node (pol-node)
  (list
   (let ((matrix (get-node-matrix pol-node))
	(name (get-node-id pol-node))
	(pol-complex (reduce-pol-complex (get-node-pol-complex pol-node))))
    (if (or (polyhedron? pol-complex)
	    (/= 1 (length pol-complex))
	    (not (symbolp name)))
      (make-pol-node pol-complex matrix name)
      (make-pol-node (get-node-pol-complex (first pol-complex))
		     (if (symbol-equal 'on-faces (get-pol-complex-rep pol-complex))
		       (mat-product (get-node-matrix (first pol-complex))
				    matrix)
		       (mat-product matrix
				    (get-node-matrix (first pol-complex))))
		     name)))))

;-----------------------------------------------------------------------------
; Calcolo del k-scheletro di un complesso poliedrale
;-----------------------------------------------------------------------------
(defun skeleton (pol-complex &optional k)
  (setf pol-complex (coerce-pol-complex pol-complex 'on-faces))
  (let ((konwn-skeletons nil))
    (declare (special konwn-skeletons))
    (if (null k)
      (error-handler " Boundary of al polyhedra complex not supported "
		     " skeleton ")
      (mapcar 'node-skeleton pol-complex (never-ending-list k)))))

;-----------------------------------------------------------------------------
; Calcolo del k-scheletro di un complesso poliedrale
;-----------------------------------------------------------------------------
(defun complex-skeleton (pol-complex &optional k)
    (mapcar 'node-skeleton pol-complex (never-ending-list k)))

(defun complex-skeleton (pol-complex &optional k)
  (declare (special konwn-skeletons))
  
  (cdr (or (assoc pol-complex konwn-skeletons :test #'equalp )
	   (car (push (cons pol-complex
			    (mapcar 'node-skeleton pol-complex (never-ending-list k)))
		      konwn-skeletons))))
  
  )

;-----------------------------------------------------------------------------
; Calcolo del k-scheletro di un nodo poliedrale
;-----------------------------------------------------------------------------
(defun node-skeleton (node k)
  (let ((pol (get-node-pol-complex node))
	(matrix (get-node-matrix node))
	(id (get-node-id node)))
    (make-pol-node (if (polyhedron? pol)
		     (polyhedron-skeleton pol k)
		     (complex-skeleton pol k))
		   matrix
		   id)))

;-----------------------------------------------------------------------------
; Calcolo del k-scheletro di un poliedro
;-----------------------------------------------------------------------------
(defun polyhedron-skeleton (pol k)
  (if (= (get-dpol pol) k) pol
      (let* ((n (get-npol pol))
	     (n-k (- n k))
	     (faces (get-flist pol))
	     (cells (get-clist pol))
	     cells-groups-list
	     faces-not-adjacent)
	(dolist (cell cells)
	  (let* ((cell-faces (get-cell-faces cell))
		 (cell-vertices (get-cell-vertices cell))
		 (cell-adjacences (get-cell-adjacences cell))
		 (max  (length cell-faces))
		 (indices nil))
	    (dotimes (i n-k)
	      (push i indices))
	    (do ((indices indices)
		 (all-bundary-faces t t)
		 (test t (incf-indices indices max)))
		((not test))
	      (dotimes (i n-k)
		(if (symbolp (nth (nth i indices) cell-adjacences))
		  (setf all-bundary-faces nil)))
	      (when all-bundary-faces
		(let ((faces-names   ; facce di base del sottospazio 
		       (mapcar 'first 
			       (mapcar 'nth indices (never-ending-list cell-faces))))
		      all-vertices   ; insieme dei vertici appartenenti al sottospazio
		      (all-faces     ; insieme di tutte le facce che generano il sottospazio 
		       (mapcar 'first cell-faces))
		      all-faces-par  ; parametri delle facce che generano il sottospazio
		      other-faces    ; altre facce che intersecano il sottospazio
		      boundary-faces ; facce che delimitano il sottospazio
		      )
		  (dotimes (i (length cell-adjacences))
		    (if (symbolp (nth i cell-adjacences))
		      (setf faces-not-adjacent
			    (adjoin (first (nth i cell-faces)) faces-not-adjacent))))
		  (dolist (vertex cell-vertices)
		    (if (subsetp faces-names vertex)
		      (push vertex all-vertices)))
		  (when all-vertices
		    (dolist (vertex all-vertices)
		      (setf all-faces (intersection all-faces vertex)))
		    (setf all-vertices 
			  (mapcar 'set-difference
				  all-vertices
				  (never-ending-list all-faces)))
		    (dolist (vertex all-vertices)
		      (setf other-faces (union other-faces vertex)))
		    (setf all-faces-par
			  (mapcar 'get-coord
				  (mapcar 'assoc 
					  all-faces
					  (never-ending-list faces)
					  (never-ending-list  :test)
					  (never-ending-list  #'equalp))))
		    (when (= n-k
			     (calc-matrix-rank 0 (1- (length all-faces))
					       0 n
					       (copy-tree all-faces-par)))
		      (dolist (boundary-face other-faces)
			(let (boundary-face-vertices
			      (common-faces other-faces))
			  (dolist (vertex all-vertices)
			    (if (member boundary-face vertex)
			      (push vertex boundary-face-vertices)))
			  (dolist (vertex boundary-face-vertices)
			    (setf common-faces
				  (intersection common-faces vertex)))
			  (if (and (= 1 (length common-faces))
				   (= (1+ n-k)
				      (calc-matrix-rank 
				       0 (length all-faces)
				       0 n
				       (copy-tree 
					(cons 
					 (get-coord
					  (assoc (first common-faces)
						 faces :test #'equalp))
					 all-faces-par)))))
			    (setf boundary-faces
				  (adjoin (first common-faces) boundary-faces)))))
		      (setf all-vertices 
			    (mapcar 'intersection
				    all-vertices
				    (never-ending-list boundary-faces)))
		      (setf cells-groups-list
			    (put-new-cell-in-list
			     all-faces-par
			     boundary-faces
			     all-vertices
			     cells-groups-list
			     n-k
			     n
			     cell-faces
			     (get-cell-sign cell))))))))))
	(reset-polyhedral-node-counter)
	(mapcar 'calc-node-from-cells-group 
		cells-groups-list
		(never-ending-list faces)
		(never-ending-list n-k)
		(never-ending-list faces-not-adjacent)))))

(defun polyhedron-skeleton (pol k)
  (declare (special konwn-skeletons))
  
  (cdr (or (assoc pol konwn-skeletons :test #'equalp)
	   (car (push (cons pol
			    (if (= (get-dpol pol) k) pol
				(let* ((n (get-npol pol))
				       (n-k (- n k))
				       (faces (get-flist pol))
				       (cells (get-clist pol))
				       cells-groups-list
				       faces-not-adjacent)
				  (dolist (cell cells)
				    (let* ((cell-faces (get-cell-faces cell))
					   (cell-vertices (get-cell-vertices cell))
					   (cell-adjacences (get-cell-adjacences cell))
					   (max  (length cell-faces))
					   (indices nil))
				      (dotimes (i n-k)
					(push i indices))
				      (do ((indices indices)
					   (all-bundary-faces t t)
					   (test t (incf-indices indices max)))
					  ((not test))
					(dotimes (i n-k)
					  (if (symbolp (nth (nth i indices) cell-adjacences))
					    (setf all-bundary-faces nil)))
					(when all-bundary-faces
					  (let ((faces-names   ; facce di base del sottospazio 
						 (mapcar 'first 
							 (mapcar 'nth indices (never-ending-list cell-faces))))
						all-vertices   ; insieme dei vertici appartenenti al sottospazio
						(all-faces     ; insieme di tutte le facce che generano il sottospazio 
						 (mapcar 'first cell-faces))
						all-faces-par  ; parametri delle facce che generano il sottospazio
						other-faces    ; altre facce che intersecano il sottospazio
						boundary-faces ; facce che delimitano il sottospazio
						)
					    (dotimes (i (length cell-adjacences))
					      (if (symbolp (nth i cell-adjacences))
						(setf faces-not-adjacent
						      (adjoin (first (nth i cell-faces)) faces-not-adjacent))))
					    (dolist (vertex cell-vertices)
					      (if (subsetp faces-names vertex)
						(push vertex all-vertices)))
					    (when all-vertices
					      (dolist (vertex all-vertices)
						(setf all-faces (intersection all-faces vertex)))
					      (setf all-vertices 
						    (mapcar 'set-difference
							    all-vertices
							    (never-ending-list all-faces)))
					      (dolist (vertex all-vertices)
						(setf other-faces (union other-faces vertex)))
					      (setf all-faces-par
						    (mapcar 'get-coord
							    (mapcar 'assoc 
								    all-faces
								    (never-ending-list faces)
								    (never-ending-list  :test)
								    (never-ending-list  #'equalp))))
					      (when (= n-k
						       (calc-matrix-rank 0 (1- (length all-faces))
									 0 n
									 (copy-tree all-faces-par)))
						(dolist (boundary-face other-faces)
						  (let (boundary-face-vertices
							(common-faces other-faces))
						    (dolist (vertex all-vertices)
						      (if (member boundary-face vertex)
							(push vertex boundary-face-vertices)))
						    (dolist (vertex boundary-face-vertices)
						      (setf common-faces
							    (intersection common-faces vertex)))
						    (if (and (= 1 (length common-faces))
							     (= (1+ n-k)
								(calc-matrix-rank 
								 0 (length all-faces)
								 0 n
								 (copy-tree 
								  (cons 
								   (get-coord
								    (assoc (first common-faces)
									   faces :test #'equalp))
								   all-faces-par)))))
						      (setf boundary-faces
							    (adjoin (first common-faces) boundary-faces)))))
						(setf all-vertices 
						      (mapcar 'intersection
							      all-vertices
							      (never-ending-list boundary-faces)))
						(setf cells-groups-list
						      (put-new-cell-in-list
						       all-faces-par
						       boundary-faces
						       all-vertices
						       cells-groups-list
						       n-k
						       n
						       cell-faces
						       (get-cell-sign cell))))))))))
				  (reset-polyhedral-node-counter)
				  (mapcar 'calc-node-from-cells-group 
					  cells-groups-list
					  (never-ending-list faces)
					  (never-ending-list n-k)
					  (never-ending-list faces-not-adjacent)))))
		      konwn-skeletons))))
  
  )

;-----------------------------------------------------------------------------
; Aggiunge una cella di dimensione intrinseca k allalista di celle gi
; calcolate.
; Se il sottospazio che la contiene  gi stato creato, la cella verr 
; semplicemente aggiunta alla lista di celle appartenenti a quel sottospazio.
;-----------------------------------------------------------------------------
(defun put-new-cell-in-list (all-faces-par
			     boundary-faces
			     all-vertices
			     cells-groups-list
			     n-k
			     n
			     cell-faces
			     cell-sign)
  (let* ((signed-faces (mapcar 'assoc 
			       boundary-faces
			       (never-ending-list cell-faces)
			       (never-ending-list  :test)
			       (never-ending-list  #'equalp)))
	 (group-index (position
		       all-faces-par
		       cells-groups-list
		       :test #'(lambda (faces-par cells-group)
				 (let ((all-pars
					(copy-tree (append faces-par
							   (first cells-group)))))
				   (= n-k (calc-matrix-rank 
					   0 (1- (length all-pars))
					   0 n
					   all-pars))))))
	 (group (if group-index
		  (nth group-index cells-groups-list)
		  (first (push (list all-faces-par nil nil)
			       cells-groups-list)))))
    (when (eql '+ cell-sign)
      (setf signed-faces (mapcar 'opposit-face signed-faces)))
    (setf (second group)
	  (union (second group) boundary-faces))
    (setf (third group)
	  (cons (list boundary-faces all-vertices signed-faces)
		(third group))))
  cells-groups-list)

;-----------------------------------------------------------------------------
; Calcola la trasposta della matrice che porta il sottospazio 
;-----------------------------------------------------------------------------
(defun calc-node-transposed-matrix (faces-pars)
  (let* ((faces-num (length faces-pars))
	 (n (1- (length (first faces-pars))))
	 (matrix (mapcar 'append 
			 (transpose-matrix faces-pars)
			 (identity-matrix (1+ n))))
	 (indices (append (reduce-matrix 1 n 0 (1- faces-num) faces-num matrix)
			  (make-list (- n faces-num) :initial-element -1)))
	 row-index)
    (dotimes (i faces-num)
      (setf row-index (position (- faces-num i 1) indices))
      (swap-rows (- n i) (1+ row-index) matrix) 
      (swap-rows (- n i 1) row-index indices))
    (setf matrix (mapcar 'nthcdr 
			 (never-ending-list faces-num)
			 matrix))))

;-----------------------------------------------------------------------------
; Calcola il nodo poliedrale dato dalla trasformazione affine che posiziona
; il sottospazio Rk in Rn e dal poliedro Pk,k composto dalle celle date.
; Il gruppo di celle  una lista di 3 elementi:
;   1) matrice (lista di liste) dei parametri delle facce che generano
;      il sottospazio;
;   2) lista dei nomi di tutte le facce usate in tutte le celle;
;   3) lista dei parametri che definiscono ciascuna cella; 
;    I parametri di definizione di una cella sono a loro volta forniti
;    in una lista di tre elementi:
;       1) lista dei nomi delle facce di bordo della cella;
;       2) lista dei vertici (indicati come lista delle facce che li generano)
;          di bordo della cella;
;       3) lista delle coppie (nome_faccia . segno) che delimitano la cella
;          considerata con segno positivo.
;-----------------------------------------------------------------------------
(defun calc-node-from-cells-group (cells-group pol-faces n-k faces-not-adjacent)
  (reset-face-counter) 
  (reset-convex-cell-counter) 
  (let* ((transposed-matrix (calc-node-transposed-matrix (first cells-group)))
	 (n (length transposed-matrix))
	 (faces (copy-tree (mapcar 'assoc 
				   (second cells-group)
				   (never-ending-list pol-faces)
				   (never-ending-list  :test)
				   (never-ending-list  #'equalp))))
	 faces-to-invert
	 new-face-pars
	 (matrix (mat-inv (make-array (list n n)
				      :initial-contents 
				      (transpose-matrix transposed-matrix))))
	 (cells-list (mapcar #'(lambda (cell)
				(push (advance-convex-cell-counter)
				      cell))(third cells-group)))
	 (new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells)))
    (dolist (face faces)
      (setf new-face-pars (normalize-parameters 
			   (butlast  
			    (mapcar `(lambda (row) 
				       (apply '+ 
					      (mapcar '* row 
						      (quote ,(get-coord face)))))
				    transposed-matrix)
			    n-k))
	    (second face) (first new-face-pars))
      (when (eql '+ (cdr new-face-pars))
	(push (first face) faces-to-invert)))
    (dolist (cell cells-list)
      (setf (fourth cell)
	    (mapcar #'(lambda (face)
			(if (member (first face) faces-to-invert)
			  (opposit-face face)
			  face))
		    (fourth cell)))
      (push 
       (mapcar `(lambda (face)
		  (if (member face (quote ,faces-not-adjacent))
		    (dolist (cell (quote ,cells-list) 0)
		      (if (and (not (eq (first cell)
					(quote ,(first cell))))
			       (member face (second cell)))
			(return (first cell))))
		    0))
	       (second cell))
       (cddr cell)))
    (dolist (cell cells-list)
      (put-cell (make-convex-cell (first cell)
				  '+
				  (fifth cell)
				  (fourth cell)
				  (third cell)
				  nil)
		new-polyhedron))
    (put-faces faces new-polyhedron)
    (set-npol new-polyhedron (- n n-k 1))
    (set-dpol new-polyhedron (- n n-k 1))
    (set-#kd  new-polyhedron)
    (set-#k0  new-polyhedron) 
    (set-#F  new-polyhedron)
    (set-#C  new-polyhedron) 
    (make-pol-node new-polyhedron
		   matrix
		   (advance-polyhedral-node-counter))))

;-----------------------------------------------------------------------------
; Incrementa una lista di indici
;   - gli indici devono essere una sequenza decrescente
;   - ogni indice deve appartenere all'intervallo [0 max]
;-----------------------------------------------------------------------------
(defun incf-indices (indices max)
  (if (and indices
	   (= max (incf (first indices))))
    (if (incf-indices (rest indices) (1- max))
      (setf (first indices)
	    (1+ (second indices))))
    indices))

;-----------------------------------------------------------------------------
; Applicazione di una generica trasformazione affine ad un complesso 
; poliedrale
;-----------------------------------------------------------------------------
(defun general-transform (pol-complex matrix)
  (list  (make-pol-node pol-complex matrix)))

;-----------------------------------------------------------------------------
; Applica uno scalamento di fattore scale-factor alla variabile i-esima del
; complesso poliedrale pol-complex
;-----------------------------------------------------------------------------
(defun scale-pol-complex (i scale-factor pol-complex)
  (let* ((n (get-npol-complex pol-complex))
	 (matrix (mat-identity  (1+ n)))
	 (rep (get-pol-complex-rep pol-complex))
	 (i+1 (1+ i)))
    (if (symbol-equal rep 'on-faces)
      (setf (aref matrix i+1 i+1) (/ scale-factor))
      (setf (aref matrix i+1 i+1) scale-factor))
    (general-transform pol-complex matrix)))

;-----------------------------------------------------------------------------
; Applica una traslazione della quantit translate-factor alla variabile
; i-esima del complesso poliedrale pol-complex
;-----------------------------------------------------------------------------
(defun translate-pol-complex (i translate-factor pol-complex)
  (let* ((n (get-npol-complex pol-complex))
	(matrix (mat-identity  (1+ n)))
	(rep (get-pol-complex-rep pol-complex)))
    (if (symbol-equal rep 'on-faces)
      (setf (aref matrix (1+ i) 0) (- translate-factor))
      (setf (aref matrix (1+ i) 0) translate-factor))
    (general-transform pol-complex matrix)))


;-----------------------------------------------------------------------------
; Applica uno shearing con vettore shearing-vector-list sulla variabile
; i-esima del complesso poliedrale pol-complex
;-----------------------------------------------------------------------------
(defun shear-pol-complex (i shearing-vector-list pol-complex)
  (let* ((n (get-npol-complex pol-complex))
	 (matrix (identity-matrix  (1+ n)))
	 (rep (get-pol-complex-rep pol-complex)))
    (setf (nth i matrix)
	  (cons 0 shearing-vector-list))
    (setf matrix (make-array  (list (1+ n)(1+ n))
			      :initial-contents matrix ))
    (if (symbol-equal rep 'on-faces)
      (setf matrix (mat-inv  matrix)))
    (general-transform pol-complex matrix)))

;-----------------------------------------------------------------------------
; Applica una rotazione della quantit alpha nel piano delle variabili i j
;  del complesso poliedrale pol-complex
;-----------------------------------------------------------------------------
(defun rotate-pol-complex (i j alpha pol-complex)
  (let* ((n (get-npol-complex pol-complex))
	(matrix (mat-identity  (1+ n)))
	(rep (get-pol-complex-rep pol-complex)))
    (if (symbol-equal rep 'on-faces)
      (progn
      (setf (aref matrix (1+ i) (1+ i)) (cos alpha))
      (setf (aref matrix (1+ i) (1+ j)) (sin alpha))
      (setf (aref matrix (1+ j) (1+ i)) (- (sin alpha)))
      (setf (aref matrix (1+ j) (1+ j)) (cos alpha))
      )
      (progn
      (setf (aref matrix (1+ i) (1+ i)) (cos alpha))
      (setf (aref matrix (1+ i) (1+ j)) (- (sin alpha)))
      (setf (aref matrix (1+ j) (1+ i)) (sin alpha))
      (setf (aref matrix (1+ j) (1+ j)) (cos alpha))
      ))
    (general-transform pol-complex matrix)))

;-----------------------------------------------------------------------------
; Restituisce il complesso polyhedrale uguale a quello dato ma basato sul
; nuovo tipo di rappresentazione.
;-----------------------------------------------------------------------------
(defun coerce-pol-complex (pol-complex new-rep)
  (let ((old-rep (get-pol-complex-rep pol-complex)))
    (if (symbol-equal new-rep old-rep)
      pol-complex
      (if (polyhedron? pol-complex)
	(coerce-polyhedron pol-complex new-rep)
	(mapcar #'(lambda (old-node)
		   (make-pol-node (coerce-pol-complex (get-node-pol-complex old-node)
						      new-rep)
				  (mat-inv (get-node-matrix old-node))
				  (get-node-id  old-node)))
		pol-complex)))))

(defun coerce-pol-complex (pol-complex new-rep)
  (let ((coerced-pol-complex nil))
    (declare (special coerced-pol-complex))
    (coerce-polyhedra-complex pol-complex new-rep)))

;-----------------------------------------------------------------------------
; Restituisce il complesso polyhedrale uguale a quello dato ma basato sul
; nuovo tipo di rappresentazione.
;-----------------------------------------------------------------------------
(defun coerce-polyhedra-complex (pol-complex new-rep)
  (declare (special coerced-pol-complex))

  (cdr (or (assoc pol-complex coerced-pol-complex :test #'equalp)
	   (car (push (cons pol-complex
			    (let ((old-rep (get-pol-complex-rep pol-complex)))
			      (if (symbol-equal new-rep old-rep)
				pol-complex
				(if (polyhedron? pol-complex)
				  (coerce-polyhedron pol-complex new-rep)
				  (mapcar #'(lambda (old-node)
					      (make-pol-node (coerce-polyhedra-complex (get-node-pol-complex old-node)
										new-rep)
							     (mat-inv (get-node-matrix old-node))
							    (get-node-id  old-node)))
					  pol-complex)))))
		      coerced-pol-complex)))))



