; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;           Universita' di Roma "La Sapienza"                              
;
;       Dipartimento di Informatica e Sistemistica                         
;
;            Via Buonarroti 12 - 00185 Roma                                
;
;                                                                          
;
;                      CAD Group                                           
;
;                                                                          
;
; Project: SimpleXn                                                        
;
; Identification: Module Functlib                                          
;
;                                                                          
;
;  Version: 1.0                                                            
;
;                                                                          
;
;  Author: V. Ferrucci                                                     
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;

(provide 'functlib) 

(inpackage kernel)  

(export '(make-translation-matrix
	  make-shearing-matrix
	  make-rotation-matrix
	  make-idmat
	  post-multiply
	  add-1dim
	  add-1
	  merge-vlist
	  merge-pol
	  setpol
	  chop
	  get-face
	  check
	  translate
	  shear
	  scale
	  rotate
	  extract-admat
	  update-ad
	  array->list
	  integer->string
	  string->integer
	  opposite
	  kill
	  remove-elements
	  resetreadtable))


#+:APPLE (require 'io)
#+:akcl
  (load (concatenate 'string user::**simplexn-directory 
		     "/core/matlib.lisp"))

(require 'matlib)

;#+:APPLE
;(require 'maclib)

#+:APPLE (use-package 'io)
(use-package 'matlib)
;#+:APPLE
;(use-package 'maclib)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Matrix functions

;(defun make-translation-matrix (n l)
;   (do ((tmat (make-idmat (+ n 1)))    
;	(j 0 (+ j 1)))
;       ((= j n) tmat)
;     (setf (aref tmat n j) (nth j l))))

(defun make-translation-matrix (n l)
  (let ((l-double (mapcar #'(lambda (n) (coerce n 'double-float)) l)))
   (do ((tmat (make-idmat (+ n 1)))    
	(j 0 (+ j 1)))
       ((= j n) tmat)
     (setf (aref tmat n j) (nth j l-double)))))

;(defun make-shearing-matrix (n i l) ;;shearing w.r.t x_i
;  (let ((shmat (make-idmat n)))
;    (mat-set-a-column 
;     shmat
;     i 
;     (coerce (append (subseq l 0 i) '(1) (subseq l i (decf n))) 'vector))
;    shmat))

(defun make-shearing-matrix (n i l) ;;shearing w.r.t x_i
  (let ((shmat (make-idmat n))
        (l-double (mapcar #'(lambda (n) (coerce n 'double-float)) l)))
    (mat-set-a-column 
     shmat
     i 
     (coerce (append (subseq l-double 0 i) 
                     '(1.0) 
                     (subseq l-double i (decf n))) 'vector))
    shmat))

(defun make-rotation-matrix (n xi xj alfa)
  (let ((mat (make-idmat n))
	(xi-1 (- xi 1))
	(xj-1 (- xj 1))
	(cosalfa (cos (* alfa **pi/180)))
	(sinalfa (sin (* alfa **pi/180))))
    (setf (aref mat xi-1 xi-1) cosalfa)
    (setf (aref mat xj-1 xj-1) cosalfa)
    (setf (aref mat xi-1 xj-1) (- sinalfa))
    (setf (aref mat xj-1 xi-1) sinalfa)
    mat))

(defun make-idmat (n)
  (do ((idmat (make-array (list n n) 
                         :element-type 'double-float
;			 :element-type 'float
			 :initial-element 0.0d0))
      (i 0 (+ i 1)))
     ((= i n) idmat)
   (setf (aref idmat i i) 1.0d0)))

(defun post-multiply (vect mat)
  (let ((cv (length vect))
       (rm (array-dimension mat 0))
       (cm (array-dimension mat 1)))
   (if (= cv rm)
    (do ((result (make-array cm :element-type 'double-float))
	(j 0 (+ j 1)))
       ((= j cm) result)
     (setf (aref result j)
	  (do ((sum 0)
	     (i 0 (+ i 1)))
	    ((= i cv) sum)
	  (setq  sum (+ sum (* (aref vect i)
			     (aref mat i j)))))))
    (error-handler 5 "post-multiply"))))

;;; Functions for simplices and polyhedra

(defmacro setpol (x pol)
  `(progn (setf ,x ,pol) (values)))

(defun chop (pol)
  (make-polyhedron 
   (mapcar #'(lambda (v)
	       (set-coord
		(mapcar #'(lambda (coord)
			    (if (< (abs coord) **epsilon) 0 coord))
			(get-coord v))
		v))
	   (get-vlist pol))
   (get-slist pol)))

(defun get-face (s i)
  (if (get-type-checking)
   (if (not (<= 0 i (- (get-dsim s) 1))) (error-handler 1 "get-face")))
  (values (if (eql (get-sign s) '+)
	   (if (oddp i) '- '+)
	   (if (oddp i) '+ '-))
	 (kill (get-vert s) i)))

(defun check (pol &optional (print-messages-flag nil))
  (if (get-type-checking) 
    (if (not (polyhedron? pol)) (error-handler 1 "check")))
  (let* ((slist (get-slist pol))
	 (admat (extract-admat slist))
	 (m+1 (+ (get-dpol pol) 1))
	 (S (get-#kd pol))
	 (S-1 (- S 1))
	 (table (make-array (list S)))
	 (l1 '()) (l2 '())
	 (l3 '()) (l4 '())
	 (l5 '()) (l6 '()))
    (flet 
      ((look-for (p h sphvert)
		 (do 
		   ((q 0 (+ q 1)))
		   ((or (= q S)
			(and (subsetp sphvert (get-vert (aref table q)))
			     (not (= q p))))
		    (if (< q s) (push (list p h (make-simplex-symbol q)) l2))))))
      (mapc #'(lambda (s) 
		(setf (aref table (get-index (get-id s))) s)) slist)
      (do ((p 0 (+ p 1)))
	  ((= p S))
	(do ((h 0 (+ h 1)))
	    ((= h m+1))
	  (let ((q (aref admat p h))
		(sp (aref table p)))
	    (cond 
	     ((= q -2) (push (list p h) l1))
	     ((= q -99))
	     ((= q -1) (multiple-value-bind 
			 (sphsign sphvert)
			 (get-face sp h)
			 (declare (ignore sphsign))
			 (look-for p h sphvert)))
	     ((not (<= 0 q S-1)) (push (list p h q) l3))
	     (t (let ((k (do ((k 0 (+ k 1)))
			     ((or (= k m+1) (= (aref admat q k) p)) k)))
		      (sq (aref table q)))
		  (if (= k m+1) (push (list p h q q p) l4)
		      (multiple-value-bind (sphsign sphvert) 
					   (get-face sp h)
			(multiple-value-bind (sqksign sqkvert) 
					     (get-face sq k)
			  (unless (equal sphvert sqkvert) 
			    (push 
			     (list p h q q k p sphsign 
				   sphvert sqksign sqkvert) 
			     l5))
			  (unless (eql sphsign (opposite sqksign))
			    (push 
			     (list p h q q k p sphsign
				   sphvert sqksign sqkvert) 
			     l6))
			  (setf (aref admat q k) -99))))))))))
      (let ((all-is-ok (and (null l1)  (null l2)  (null l3) 
			    (null l4)  (null l5)  (null l6))))
	(if print-messages-flag 
	  (if all-is-ok 
	    t
	    (progn 
	      (write-line "Diagnostics :") 
	      (printmessages 
	       "Adjacency ~D ~D is not resolved" l1)
	      (printmessages 
	       "Adjacency ~D ~D indicates a border face which is instead a face of
~A" l2)
	      (printmessages 
	       "Adjacency ~D ~D is ~D, which is not in 0...S-1" l3)
	      (printmessages
	       "Adjacency ~D ~D is ~D, but A~Dk = ~D doesn't hold for any k" l4)
	      (printmessages 
	       "Adjacency ~D ~D is ~D, and A~D~D = ~D but extracted faces are ~A~A
and ~A~A" l5)
	      (printmessages 
	       "Adjacency ~D ~D is ~D, and A~D~D = ~D but extracted faces have the
same sign: ~A~A and ~A~A" l6)
	      nil))
	  (if all-is-ok t nil))))))

(defun translate (pol tvect) ;tvect is a list
  (make-polyhedron (mapcar #'(lambda (v)
			      (set-coord
			  (mapcar #'+ (get-coord v) tvect)
			  v))
			  (get-vlist pol))
		  (get-slist pol)))

;;; shearing w.r.t x_i, shvect is a list of length n-1 

(defun shear (pol i shlist)  
  (let ((shmat (make-shearing-matrix (get-npol pol) (decf i) shlist)))
    (make-polyhedron (mapcar #'(lambda (v)   
				 (set-coord
				  (coerce 
				   (mat-vet-product shmat (coerce (get-coord v) 'vector))
				   'list)
				  v))
			     (get-vlist pol))
		     (get-slist pol))))

(defun scale (pol splist)
  (make-polyhedron (mapcar #'(lambda (v)
			      (set-coord
			  (mapcar #'* (get-coord v) splist)
			  v))
			  (get-vlist pol))
		  (get-slist pol)))

(defun rotate (pol alfa &optional(rotation-plane (get-rotation-plane)))
  (let ((rotmat (make-rotation-matrix (+ 1 (get-npol pol))
				      (first rotation-plane) 
				      (second rotation-plane)
				      alfa)))
    (make-polyhedron 
     (mapcar #'(lambda (v)
		 (set-coord
		  (butlast (coerce (post-multiply 
				    (coerce (add-1 (get-coord v)) 'vector) 
				    rotmat) 'list))
		  v))
	     (get-vlist pol))
     (get-slist pol))))

(defun add-1dim (vcoord)
  (append vcoord '(0)))

(defun add-1 (vcoord)
  (append vcoord '(1)))

(defun merge-vlist (vlist)
(mapcar #'(lambda (v) (set-coord (add-1dim (get-coord v)) v))
vlist))

(defun merge-pol (pol)
(make-polyhedron (merge-vlist (get-vlist pol))
(get-slist pol)))

(defun extract-admat (slist)
  (let ((adlist
	 (mapcar #'(lambda (s)
		     (mapcar #'(lambda (ad) 
				 (cond
				  ((eql ad 0) -1)
				  ((eql ad '*) -2)
				  ((simplex-symbol? ad) (get-index ad))
				  (t (error-handler 1 "extract-admat"))))
			     (get-ad s)))
		 slist)))
    (make-array (list (length adlist) (length (car adlist)))
		:initial-contents adlist
		:element-type 'integer)))

(defun update-ad (admat slist)
  (let ((adlist (array->list admat)))
    (mapcar #'(lambda (s ad) 
		(set-ad 
		 (mapcar #'(lambda (ad) 
			     (cond ((= ad -2) '*)
				   ((= ad -1) 0)
				   ((>= ad 0) (make-simplex-symbol ad))
				   (t (error-handler 1 "extract-admat"))))
			 ad) s))
	    slist adlist)))

;;; Conversion functions

(defun array->list (mat)
  (do ((l '())
      (i 0 (+ i 1)))
     ((= i (array-dimension mat   0)) l)
   (setq l (append 
	   l 
	   (list (do ((row '())
		    (j 0 (+ j 1)))
		   ((= j (array-dimension mat   1)) row)
		 (setq row 
		    (append row 
			 (list (aref mat i j))))))))))

(defun integer->string (int)
  (let* ((ndigit (do ((greater 10 (* greater 10))
		      (i 1 (+ i 1)))
		     ((> greater int) i)))
	 (listofdigit 
	  (do ((l '() (append l (list (floor (/ theinteger aux)))))
	       (theinteger int (mod int aux))
	       (aux (expt 10 (- ndigit 1)) (/ aux 10))
	       (i 0 (+ i 1)))
	      ((= i ndigit) l))))
    (coerce (mapcar #'digit-char listofdigit) 'string)))

(defun string->integer (str)
  (do ((sum 0)
      (len (length str))
      (i 0 (+ i 1)))
     ((= i len) sum)
   (setq  sum (+ sum (let* ((ch (char str i))
			   (n  (cond ((char= ch #\0) 0)
				 ((char= ch #\1) 1)
				 ((char= ch #\2) 2)
				 ((char= ch #\3) 3)
				 ((char= ch #\4) 4)
				 ((char= ch #\5) 5)
				 ((char= ch #\6) 6)
				 ((char= ch #\7) 7)
				 ((char= ch #\8) 8)
				 ((char= ch #\9) 9)
				 (t (error-handler 1 "string->integer")))))
		      (* n (expt 10 (- len i 1))))))))

;;;;  For general use

(defun opposite (sign)
  (cond ((eql sign '-) '+ )
       ((eql sign '+)'- )
       (t (error-handler 1 "opposite"))))

;Per l'1.3 va sostituita con la seguente
;(defun kill (l i)
;  (remove (nth i l) l :start i :end (+ i 1)))

(defun kill (l pos)
  (let ((i 0) (res ()))
    (dolist (item l res)
      (unless (= i pos) (setf res (append res (list item))))
      (setf i (1+ i)))))

(defun remove-elements (list indlist)
  (if (endp indlist)
   list
   (remove-elements (kill list (car indlist))
		   (mapcar #'(lambda (i)
			    (- i 1))
			(cdr indlist)))))


;;;ripristina la readtable standard

(defun resetreadtable () (setf *readtable* (copy-readtable nil)))





