; -*- package: io -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Input&Output                                      ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Ferrucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'io)                           ;module name

(defpackage io)
(inpackage io) ;package name

(export '(printpol
          printvlist
          printslist
          printv
          prints
          printmat
          printhash
          printmessages
          savepol
          readpol
          exportpol
          dxf
          printpolTex))
                        
(require 'kernel)
(require 'dvrlib)
  
(use-package 'kernel)
(use-package 'dvrlib)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun printpol (pol &optional (printalot t))
  (if (polyhedron? pol)
    (progn (princ "Intrinsic dimension : ") (princ (get-dpol pol)) (terpri)
           (princ "E-space dimension : ") (princ (get-npol pol)) (terpri)
           (princ "Simplices number : ")  (princ (get-#kd pol)) (terpri)
           (princ "Vertices number : ") (princ (get-#K0 pol)) (terpri)
           (princ "Sign : ") (princ (get-pol-sign pol)) (terpri)
           (if printalot (progn
                           (printvlist (get-vlist pol))
                           (printslist (get-slist pol))
                           (terpri)
                           (values))))
    (error-handler 1 "printpol")))

(defun printvlist (l)
  (if (get-type-checking)
    (if (not (vertex-list? l))
      (error-handler 1 "printvlist")))
  (mapc #' printv l)
  (values))

(defun printslist (l)
  (if (get-type-checking)
    (if (not (vertex-list? l))
      (error-handler 1 "printslist")))
  (mapc #'prints l)
  (values))

(defun printv (v)
  (if (get-type-checking)
    (if (not (vertex? v))
      (error-handler 1 "printv")))
  (terpri)
  (princ (get-id v))
  (princ " : ")
  (princ (get-coord v))
  (values))

(defun prints (s)
  (if (get-type-checking)
    (if (not (simplex? s))
      (error-handler 1 "prints")))
  (terpri)
  (princ (get-id s))
  (princ " : ")
  (princ (get-sign s))
  (princ (get-vert s))
  (princ " ")
  (princ (get-ad s))
  (values))

(defun printmat (mat)
  (if (get-type-checking)
   (if (not (arrayp mat))
    (error-handler 1 "printmat")))
  (let ((m (array-dimension mat 0))
        (n (array-dimension mat 1)))
   (dotimes (i m)
    (dotimes (j n)
     (format t "~6,:@A" (aref mat i j)))
    (terpri)))
  (terpri)
  (values))

(defun printhash (hashtable)
  (maphash #'(lambda (key val)
               (format t "~A  ~A~%" key val))
           hashtable)
  (values))

(defun printmessages (format-string l &optional (outstream t))
  (mapc #'(lambda (itemlist)
            (format outstream "~?~%" format-string itemlist))
        l)
  (values))

(defun savepol (pol &optional (filename nil))
  (if (polyhedron? pol)
   (let ((filename (if filename filename
                       (prompt-for-a-new-file))))
    (if (not (string-equal (pathname-type filename) "pol"))
     (setq filename (concatenate 'string (pathname-host  filename)
                                (pathname-device filename)
                                (pathname-directory filename)
                                (pathname-name filename)
                                ".pol")))
    (with-open-file (outfile filename
                            :direction :output
                            :if-exists :overwrite
                            :if-does-not-exist :create)
     (prin1 pol outfile)
     t))
   (error-handler 1 "save-pol")))

(defun readpol (&optional(filename nil))
  (if (not filename) (setq filename (prompt-for-an-existing-file)))
      (with-open-file (infile filename
                              :direction :input)
        (let ((pol (read infile)))
          (if (polyhedron? pol)
            pol
            (error-handler 6 "readpol")))))


(defun exportpol (pol &optional (filename (prompt-for-a-new-file)))  
  (cond ((not (polyhedron? pol)) (error-handler 1 "conv"))
        ((or (/= (get-dpol pol) 2) (/= (get-npol pol) 3)) (error-handler 9 "conv"))
        (t (let* ((S (get-#Kd pol))
                  (nvert (get-#K0 pol))
                  (vlist (get-vlist pol))
                  (slist (get-slist pol))
                  (table (do ((l '())
                              (i 0 (+ i 1)))
                             ((= i nvert) l)
                           (push (list (get-id (nth i vlist)) (+ i 1)) l))))
             (if (not (string-equal (pathname-type filename) "convt"))
               (setq filename (concatenate 'string (pathname-host  filename)
                                           (pathname-device filename)
                                           (pathname-directory filename)
                                           (pathname-name filename)
                                           ".convt")))
             (with-open-file (outfile filename
                                      :direction :output
                                      :if-exists :overwrite
                                      :if-does-not-exist :create)
               (format outfile "; Winged Triangle representation of a (2)-polyhedron merged in R3~%")
               (format outfile "; in ASCII format compatible with MINERVA.~%")
               (format outfile "; generated by SimpleXn~%~D~%~D~%" -200 nvert)
               (mapc #'(lambda (v)
                         (format outfile "~?~%" "~18,12,3@E ~18,12,3@E ~18,12,3@E" (get-coord v)))
                     vlist)
               (format outfile "~D~%" S)
               (mapc #'(lambda (s)
                         (let ((svert (get-vert s))
                               (sad (get-ad s)))
                           (if (eql (get-sign s) '+)
                             (format outfile "~D ~D ~D   ~D ~D ~D~%"
                                     (second (assoc (first svert) table))
                                     (second (assoc (second svert) table))
                                     (second (assoc (third svert) table))
                                     (if (eql (third sad) 0) 0 (+ 1 (get-index (third sad))))
                                     (if (eql (first sad) 0) 0 (+ 1 (get-index (first sad))))
                                     (if (eql (second sad) 0) 0 (+ 1 (get-index (second sad)))))
                             (format outfile "~D ~D ~D   ~D ~D ~D~%"
                                     (second (assoc (second svert) table))
                                     (second (assoc (first svert) table))
                                     (second (assoc (third svert) table)) 
                                     (if (eql (third sad) 0) 0 (+ 1 (get-index (third sad))))
                                     (if (eql (second sad) 0) 0 (+ 1 (get-index (second sad))))
                                     (if (eql (first sad) 0) 0 (+ 1 (get-index (first sad))))))))
                     slist))
             t))))


(defun dxf (pol &optional (filename (prompt-for-a-new-file)))  
  (cond ((not (polyhedron? pol)) (error-handler 1 "conv"))
        ((or (/= (get-dpol pol) 2) (/= (get-npol pol) 3)) (error-handler 9 "conv"))
        (t (if (not (string-equal (pathname-type filename) "dxf"))
             (setq filename (concatenate 'string (pathname-host  filename)
                                         (pathname-device filename)
                                         (pathname-directory filename)
                                         (pathname-name filename)
                                         ".dxf")))
           (with-open-file (outfile filename
                                    :direction :output
                                    :if-exists :overwrite
                                    :if-does-not-exist :create)
             
             
             (format outfile "999~%Flat DXF output from SimpleXn~%")
             (format outfile "  0~%SECTION~%  2~%ENTITIES~%")
             (let ((svcoordtable (make-hash-table :size (get-#Kd pol))))
               (mapc #'(lambda (s)
                         (setf (gethash (get-id s) svcoordtable) (get-svertcoord (get-id s) pol)))
                     (get-slist pol))
               (mapc #'(lambda (s)
                         (let ((svcoord (gethash (get-id s) svcoordtable)))
                           (if (eql (get-sign s) '+)
                             (write-a-face 
                              (append svcoord (last svcoord)) outfile)
                             (write-a-face 
                              (list (second svcoord) (first svcoord) 
                                    (third svcoord) (third svcoord)) outfile))))
                     (get-slist pol)))
             (format outfile "  0~%ENDSEC~%  0~%EOF~%"))))
  t)

(defun write-a-face (vertices outfile)
  (format outfile "  0~%3DFACE~%  8~%SV3D~%")
  (mapc #'(lambda (code coord)
            (format outfile "~3D~%~18,12,3@E~%" code coord))
        '(10 20 30 11 21 31 12 22 32 13 23 33)
        (let ((l '()))
          (mapc #'(lambda (vcoords)
                    (mapc #'(lambda (coord) (setf l (append l (list coord))))
                          vcoords))
                vertices)
          l)))


(defun printpolTex (pol &optional (ssym 'sigma))  
  (if (not (polyhedron? pol)) 
    (error-handler 1 "printTex")
    (let* ((outstream t)
           (mpol (get-dpol pol))
           (nvert (get-#K0 pol))
           (vlist (get-vlist pol))
           (slist (get-slist pol))
           (table (do ((l '())
                       (i 0 (+ i 1)))
                      ((= i nvert) l)
                    (push (list (get-id (nth i vlist)) i) l))))
      (format outstream "~%\\begin{eqnarray}~%{\\cal W}(P)=\\left\\{~%\\begin{array}{ll}~%" mpol)
      (mapc #'(lambda (s)
                (printsimplexTex
                 ssym
                 (get-id s)
                 (get-sign s)
                 (mapcar #'(lambda (vid) (second (assoc vid table))) (get-vert s))
                 (get-ad s)))
            slist) 
      (format outstream "\\end{array}~%\\right.~%\\end{eqnarray}~%")
      (values))))

(defun printsimplexTex (ssym sid sign vlist adlist)
  (let ((simplexname (concatenate 'string "\\~(" (symbol-name ssym) "~)_{" 
                                  (integer->string (get-index sid)) "}")))
  (format t  (concatenate 'string simplexname "= "))
  (format t "~A \\langle " sign)
  (format t "v_{~D}" (car vlist))
  (format t "~{,v_{~D}~}" (cdr vlist))
  (format t (concatenate 'string " \\rangle; & {\\cal A}(" simplexname ")= \\langle ")))
  (let ((adlistTex (mapcar #'(lambda (ad) 
                               (if (eql ad 0) 
                                 '\\bot
                                 (intern (concatenate 'string "\\" (symbol-name ssym) "_{" 
                                                      (integer->string (get-index ad)) "}"))))
                           adlist)))
    (format t "~(~A~)" (car adlistTex))
    (format t "~{~(,~A~)~}" (cdr adlistTex))
    (format t " \\rangle \\\\~%")))
