; -*- package: math-lib -*-

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;           Universita' di Roma "La Sapienza"                               ;

;       Dipartimento di Informatica e Sistemistica                          ;

;            Via Buonarroti 12 - 00185 Roma                                 ;

;                                                                           ;

;                      CAD Group                                            ;

;                                                                           ;

; Project: SimpleXn                                                         ;

; Identification: Package math-lib                                          ;

;                                                                           ;

;  Version: 1.0                                                             ;

;                                                                           ;

;  Authors: V. Pascucci & G. Proietti                                       ;

;  date:  9-2-1994                                                           ;

;  rev : 22-2-1994

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; 

; This package contains a set of mathematical macro to manipolate

; vectors and matrices. These macros accepts two differents representations

; of matrix, an array and a list of lists structures. Users must specify

; in macro call which representation is used for data representation. 



; To avoid name conflicts with existings functions, all the macros are defined

; with the "_" prefix character.

;::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

; (16/4/94) risolto bug in _find-pivot-element

;



(provide 'math_lib)        ;module name



(defpackAge math-lib)

(inpackage math-lib)     ;package name



(use-package 'lisp-extension)

(use-package 'kernel)



(export '( _array  ; type definition symbols

           _list   

           _vector 

           

           _mat-new  ; matrix management operations

           _mat-id

           _mat-copy

           _mat-extend

           _mat-print

           _mat-element

           _mat-dimension

           

           _mat-transpose  ; matrix operations

           _mat-scalar-product

           _mat-product

           _mat-sum

           _mat-diff

           _mat-inv

           _mat-det

           _mat-rank

           

           _pivot  ; gauss-jordan related functions

           _find-pivot-element

           _multiply-row-for-coeff

           _add-pivot-row*coeff-to-row

           _pivot-with-back

           _swap-rows

           



           _vet-element ; vector management operations

           _vet-new





           _same-plane  ; miscellanea

           _gen-submat-id

           

           ))



(import '(kernel:**epsilon))



;------------------------------------------------------

; The symbols that define a valid matrix or vector

; type are defined as a set of vars hoose value is

; the symbol of the var itself.

;------------------------------------------------------

(setf _array  '_array )

(setf _list   '_list  )

(setf _vector '_vector)





(defvar supported-matrix-types nil 

  "

The variable SUPPORTED-MATRIX-TYPES is the list of

supported types in matrix manipulations.

The first of the list is the default type

(see DEFAULT-MATRIX-TYPE).") 

(setf supported-matrix-types (list _array

                                   _list

                                   _vector))





(defvar default-matrix-type nil

  "

The variable DEFAULT-MATRIX-TYPE is set to

the default type used for matrix representatio.

(see SUPPORTED-MATRIX-TYPES)")

(setf default-matrix-type (first supported-matrix-types))



;-----------------------------------------------------------------------------

; FUNZIONI ELEMENTARI DI ACCESSO A MATRICI E VETTORI.

;-----------------------------------------------------------------------------



;-------------------------------------------------------------

; Selection of a matrix element.

; This macro can be used both for read and write 

; data onto a matrix element.

; INPUT : -the matrix (an array or a list)

;         -the first  indiex of the element

;         -the second indiex of the element

; OUTPUT: -the reference to the matrix element

;

; Example:   (setf a (make-array '(3 3)))  -> #2A((NIL NIL NIL) (NIL NIL NIL) (NIL NIL NIL))

;            (setf (_mat-element a 2 1) 12) -> 12

;            (print a)                     -> #2A((NIL NIL NIL) (NIL NIL NIL) (NIL 12 NIL))

;

;            (setf b '((1 2 3)(4 5 6) (7 8 9)))               -> ((1 2 3) (4 5 6) (7 8 9))

;            (setf (_mat-element b 2 1 :element-type list) 12) -> 12

;            (print b)                                        -> ((1 2 3) (4 5 6) (7 12 9))

;-------------------------------------------------------------

(defmacro _mat-element (matrix I1 i2 &key (element-type DEFAULT-MATRIX-TYPE))

  (declare (special _array _list))

  (cond 

   ((symbol-equal element-type _array)

    `(aref  ,matrix ,i1 ,i2 ))

   ((symbol-equal element-type _list)

    `(nth ,i2 (nth ,i1 ,matrix)))

   (t (kernel:error-handler (concatenate 'string 

                                         " Bad matrix type " (symbol-name element-type))

                            "_mat-element (macro)"))))



;-------------------------------------------------------------

; Selection of a vector element.

; This macro can be used both for read and write 

; data onto a vector element.

; INPUT : -the vector (an array or a list or a vector)

;         -the index of the element

; OUTPUT: -the reference to the vector element

;

; Example:   (setf a (make-array '(3)))   -> #(NIL NIL NIL)

;            (setf (_vect-element a 2) 12) -> 12

;            (print a)                    -> #(NIL NIL 12)

;

;            (setf b (vector 15 15 15))                        -> #(15 15 15)

;            (setf (_vect-element b 2 :element-type vector) 12) -> 12

;            (print b)                                         -> #(15 15 12)

;

;            (setf c '(7 8 9))                               -> (7 8 9)

;            (setf (_vect-element c 2 :element-type list) 12) -> 12

;            (print c)                                       -> (7 8 12)

;-------------------------------------------------------------

(defmacro _vet-element (vector I1 &key (element-type DEFAULT-MATRIX-TYPE))

  (declare (special _array _vector _list))

  (cond 

   ((or (symbol-equal element-type _array)

        (symbol-equal element-type _vector))

    `(aref  ,vector ,i1))

   ((symbol-equal element-type _list)

    `(nth ,i1 ,vector))))



;-----------------------------------------------------------------------------

; Create a vector of the specified SIZE and type specified

; by the element-type key . The initial element is specified by the key

; INITIAL-ELEMENT (default is 0).

;

; INPUT : - number of components

;         - initial elements value

;

; OUTPUT: - matrix

; 

; Example:

;

;   (_vet-new 3 :initial-element 0 :element-type _list)

;   -> (0 0 0)

;

;-----------------------------------------------------------------------------



(defmacro _vet-new (size &key (element-type DEFAULT-MATRIX-TYPE)

                         (initial-element 0))

  (declare (special _array _vector _list))

  (cond 

   ((or (symbol-equal element-type _array)

        (symbol-equal element-type _vector))

    `(make-array ,size :initial-element ,initial-element))

   ((symbol-equal element-type _list)

    `(make-list ,size :initial-element ,initial-element))))

  



;------------------------------------------------------------------------------

; Swaps two rows of a matrix.

; INPUT : - matrix

;         - first  row index

;         - second row index

; OUTPUT: - undefined

; Example: 

;          (setf a (make-array '(3 3) :initial-contents '((1 2 3)(4 5 6)(7 8 9)))) 

;                              -> #2A((1 2 3) (4 5 6) (7 8 9))

;          (_swap-rows a 0 2)  -> nil

;          (print a)           -> #2A((7 8 9) (4 5 6) (1 2 3)) 

;

;          (setf b  '((1 2 3)(4 5 6)(7 8 9)))    -> '((1 2 3)(4 5 6)(7 8 9))

;          (_swap-rows b 0 2 :element-type list) -> nil

;          (print b)                             -> ((7 8 9) (4 5 6) (1 2 3))

;

;------------------------------------------------------------------------------

(defmacro _swap-rows (matrix I1 i2 &key (element-type DEFAULT-MATRIX-TYPE))

  (declare (special _array _list))

  (cond

   ((symbol-equal element-type _array) 

    `(dotimes (i (array-dimension ,matrix 1))

       (swap (aref ,matrix ,i1 i)(aref ,matrix ,i2 i))))

   ((symbol-equal element-type _list) `(swap (nth ,i1 ,matrix)(nth ,i2 ,matrix)))))



;------------------------------------------------------------------------------

; Get the range of the DIMENSION-INDEX index of MATRIX. 

; INPUT : - matrix

;         - dimension index

;

; OUTPUT: - range of the DIMENSION-INDEX index of MATRIX.

;

; Example: 

;------------------------------------------------------------------------------

(defmacro _mat-dimension (matrix dimension-index

                                 &key (element-type DEFAULT-MATRIX-TYPE))

  (declare (special _array _list))

  (cond

   ((symbol-equal element-type _array)

    `(array-dimension ,matrix ,dimension-index))

   ((symbol-equal element-type _list)

    `(if (zerop ,dimension-index)

       (length ,matrix)

       (length (first ,matrix))))))



;-----------------------------------------------------------------------------

; (Destructive) Multiply the row of index ROW-INDEX of the matrix MATRIX

; for the coefficient COEFF.

; INPUT : - matrix

;         - row index

;         - coefficient

;

; OUTPUT: - undefined.

;

; Example: 

;          (setf a (make-array '(3 3) :initial-contents '((1 2 3)(4 5 6)(7 8 9))))

;                                          ->#2A((1 2 3) (4 5 6) (7 8 9))

;          (_multiply-row-for-coeff a 1 2) ->NIL

;          (print a)                       -> #2A((1 2 3) (8 10 12) (7 8 9))

;

;          (setf b  '((1 2 3)(4 5 6)(7 8 9)))

;                              ->((1 2 3) (4 5 6) (7 8 9))

;          (_multiply-row-for-coeff b 1 2 :element-type _list)

;                              ->NIL

;          (print b)           -> ((1 2 3) (8 10 12) (7 8 9)) 

;------------------------------------------------------------------------------

(defmacro _multiply-row-for-coeff (matrix row-index coeff

                                          &key (element-type DEFAULT-MATRIX-TYPE))

  `(dotimes (i (_mat-dimension ,matrix 1 :element-type  ,element-type))

     (setf (_mat-element ,matrix ,row-index i :element-type ,element-type)

           (* (_mat-element ,matrix ,row-index i :element-type ,element-type)

              ,coeff))))





;-----------------------------------------------------------------------------

; (Destructive) Add at the row of index ROW the row of index PIVOT-ROW multiplied

; for in the matrix MATRIX for the coefficient COEFF.

;

; INPUT : - matrix

;         - row index

;         - pivot row index

;         - coefficient

;

; OUTPUT: - undefined.

;

; Examples

;

;   (setf a (make-array '(3 4) :initial-contents '((1 2 3 5)(4 5 6 11)(7 8 9 2))))

;     -> #2A((1 2 3 5) (4 5 6 11) (7 8 9 2))

;   (_add-pivot-row*coeff-to-row a 0 2 3)

;     -> NIL

;   (pprint a)

;     -> #2A((22 26 30 11) (4 5 6 11) (7 8 9 2))

;

;-----------------------------------------------------------------------------



(defmacro _add-pivot-row*coeff-to-row (matrix row

                                              pivot-row

                                              coeff

                                              &key (element-type DEFAULT-MATRIX-TYPE))

  `(progn 

     (dotimes (_add_pivot_cnt (_mat-dimension ,matrix 1 :element-type ,element-type))

       (setf (_mat-element ,matrix ,row _add_pivot_cnt :element-type ,element-type)

             (+ (_mat-element ,matrix ,row _add_pivot_cnt :element-type ,element-type)

                (*  (_mat-element ,matrix ,pivot-row _add_pivot_cnt :element-type ,element-type)

                    ,coeff))))

     )

  )



;-----------------------------------------------------------------------------

; (Destructive) Execute pivot operation on the matrix MATRIX respect the element

; specified by the PIVOT-ROW and PIVOT-COLUMN indexes

;

; The static variables _pivot_coeff and _pivot_element*-1 are used for auxiliary 

; storage of values. This macro cannot be recursive !

;

; INPUT : - matrix

;         - row index

;         - column index

;

; OUTPUT: - undefined.

; 

; Example:

;          (setf a (make-array '(3 4) :initial-contents '((1 2 3 5)(4 5 6 11)(7 8 9 0))))

;                                          -> #2A((1 2 3 5) (4 5 6 11) (7 8 9 0))

;          (_PIVOT A 0 2)

;          -> NIL

;          (PRINT A)

;          -> #2A((1/3 2/3 1 5/3) (2 1 0 1) (4 2 0 -15))



;          (setf b  '((1 2 3 5)(4 5 6 11)(7 8 9 0)) )

;           -> ((1 2 3 5)(4 5 6 11)(7 8 9 0))

;          (_pivot b 0 2 :element-type _list)

;           -> NIL

;          (pprint b)

;           -> ((1/3 2/3 1 5/3) (2 1 0 1) (4 2 0 -15))

;

;------------------------------------------------------------------------------



(defvar _pivot_element*-1)

(defvar _pivot_coeff)



(defmacro _pivot (matrix pivot-row

                         pivot-column

                         &key (element-type DEFAULT-MATRIX-TYPE))

  `(progn 

     (setf _pivot-element*-1 (- (_mat-element ,matrix 

                                              ,pivot-row ,pivot-column 

                                              :element-type  ,element-type)))

     (dotimes (_pivot_cnt (_mat-dimension ,matrix 0 

                                          :element-type  ,element-type))

       (if (/= _pivot_cnt ,pivot-row)

         (progn

           (setf _pivot_coeff (/ (_mat-element ,matrix _pivot_cnt ,pivot-column 

                                               :element-type ,element-type)

                                 _pivot-element*-1))

           (_add-pivot-row*coeff-to-row ,matrix _pivot_cnt ,pivot-row _pivot_coeff 

                                        :element-type  ,element-type))))

     (setf _pivot_coeff (/ (- _pivot-element*-1)))

     (_multiply-row-for-coeff ,matrix ,pivot-row

                              _pivot_coeff

                              :element-type ,element-type)

     ))





;-----------------------------------------------------------------------------

; (Destructive) Execute a copy of a submatrix 

;

; The static variables _pivot_coeff and _pivot_element*-1 are used for auxiliary 

; storage of values. This macro cannot be recursive !

;

; INPUT : - matrix

;         - row index

;         - column index

;

; OUTPUT: - undefined.

; 

;-----------------------------------------------------------------------------



(defmacro _mat-copy (matrix1 matrix2 min-row1 max-row1

                             min-col1 max-col1

                             start-row2 start-col2

                             &key (element-type1 DEFAULT-MATRIX-TYPE)

                             (element-type2 DEFAULT-MATRIX-TYPE))

  `(dotimes (_mat_copy_row (1+ (- ,max-row1 ,min-row1)))

     (dotimes (_mat_copy_col (1+ (- ,max-col1 ,min-col1)))

       (setf (_mat-element ,matrix2 (+ ,start-row2 _mat_copy_row) (+ ,start-col2 _mat_copy_col)

                           :element-type ,element-type2)

             (_mat-element ,matrix1 (+ ,min-row1 _mat_copy_row) (+ ,min-col1 _mat_copy_col)

                           :element-type ,element-type1))

       ))

  )







;-----------------------------------------------------------------------------

;(Destructive) Extends the spefified MATRIX with the specifieds numbers

; of rows ADDED-ROWS and columns ADDED-COLS . The added elements are

; all zeroes .

;

; The static variables _mat_extend1 and _mat_extend2 are used for auxiliary 

; storage of values. This macro cannot be recursive !

;

; INPUT : - matrix

;         - added-rows

;         - added-cols

;

; OUTPUT: - undefined

;-----------------------------------------------------------------------------



(defvar _mat_extend1)

(defvar _mat_extend2)



(defmacro _mat-extend (matrix added-rows added-cols 

                                &key (element-type DEFAULT-MATRIX-TYPE))

  (declare (special _array _list))

  `(cond 

    ((symbol-equal ,element-type _array)

     (progn

       (setf _mat_extend1 (list (+ (array-dimension ,matrix 0) ,added-rows) 

                                (+ (array-dimension ,matrix 1) ,added-cols)))

       (setf _mat_extend2 (make-array _mat_extend1 :initial-element 0))

       (_mat-copy ,matrix _mat_extend2

                  0 (1- (array-dimension ,matrix 0))

                  0 (1- (array-dimension ,matrix 1))

                  0 0)

       (setf ,matrix _mat_extend2)

       )

     )

    ((symbol-equal ,element-type _list)

     (progn

       (setf _mat_extend1 (mapcar 

                           #'(lambda (x) (append x (make-list ,added-cols 

                                                              :initial-element 0)))

                           ,matrix))

       (setf _mat_extend2 (+ ,added-cols (length (car ,matrix))))

       (dotimes (i ,added-rows)

         (setf _mat_extend1 

               (append _mat_extend1 

                       (list (make-list _mat_extend2 :initial-element 0))))

         )

       (setf ,matrix _mat_extend1)

       )

     ))

  )



;-----------------------------------------------------------------------------

; Create a matrix of the specifieds sizes NROWS and NCOLS and type specified

; by the element-type key . The initial element is specified by the key

; INIT-ELEMENT (default is 0).

;

; INPUT : - number of rows

;         - number of columns

;         - initial elements value

;

; OUTPUT: - matrix

; 

; Example:

;

;   (_mat-new 2 3 :initial-element 0 :element-type _list)

;   -> ((0 0 0) (0 0 0))

;

;-----------------------------------------------------------------------------



(defmacro _mat-new (nrows ncols &key (initial-element 0) (element-type DEFAULT-MATRIX-TYPE))

  (declare (special _array _list))

  `(cond 

    ((symbol-equal ,element-type _array)

     (make-array (list ,nrows ,ncols) :initial-element ,initial-element )

     )

    ((symbol-equal ,element-type _list)

     (let ( _mat_new_auxmat )

       (dotimes (_mat_new_counter ,nrows _mat_new_auxmat)

         (setq _mat_new_auxmat 

               (append _mat_new_auxmat

                       (list (make-list ,ncols :initial-element ,initial-element))))))

     ))

  )



;-----------------------------------------------------------------------------

; (Destructive) Replaces a square specified submatrix of MATRIX with the 

; identity matrix of corrispondent size. The submatrix is specified by

; his hig-left corner element of START-ROW and START-COL indexs and

; by his SIZE.

;

; INPUT : - matrix

;         - start-row

;         - start-col

;         - size

;

; OUTPUT: - undefined.

; 

;     (setf b (make-array '(3 4) :initial-element nil))

;     -> #2A((NIL NIL NIL NIL) (NIL NIL NIL NIL) (NIL NIL NIL NIL))

;     (_gen-submat-id b 0 2 2)

;     -> NIL

;     (_mat-print b)

;     ->  NIL         NIL       1.E+0  0.00000E+0 

;         NIL         NIL  0.00000E+0       1.E+0 

;         NIL         NIL         NIL         NIL 

;

;-----------------------------------------------------------------------------



(defmacro _gen-submat-id (matrix start-row start-col size 

                          &key (element-type DEFAULT-MATRIX-TYPE))

  `(dotimes (_gen_subid_row ,size)

     (dotimes (_gen_subid_col ,size)

       (setf (_mat-element ,matrix (+ ,start-row _gen_subid_row) 

                           (+ ,start-col _gen_subid_col)

                           :element-type ,element-type)

             (if (= _gen_subid_row _gen_subid_col)

               1

              ;else

               0))

       ))

  )



;-----------------------------------------------------------------------------

; Create a new matrix, of type specified, wich is an identity matrix of SIZE

; rows x columns.

;

; INPUT : - size

;

; OUTPUT: - matrix

; 

; Examples

;

;  (_mat-id 3)

;  -> #2A((1 0 0) (0 1 0) (0 0 1))

;

;-----------------------------------------------------------------------------



(defmacro _mat-id (size &key (element-type DEFAULT-MATRIX-TYPE))

  `(let ( (mat (_mat-new ,size ,size :element-type ,element-type)) ) 

     (_gen-submat-id mat 0 0 ,size :element-type ,element-type)

     mat

     )

  )



;-----------------------------------------------------------------------------

; (Destructive) Calculate the inverse of MATRIX. The result matrix is the same 

; matrix in input.

;

; INPUT : - matrix

;

; OUTPUT: - inverse of matrix

; 

; Examples

;

;  (setf a (make-array '(3 3) :initial-contents '((1 2 3)(5 2 0)(1 -1 3))))

;  -> #2A((1 2 3) (5 2 0) (1 -1 3))

;  (_mat-inv a )

;  -> #2A((-2/15 1/5 2/15) (1/3 0 -1/3) (7/45 -1/15 8/45))

;

;-----------------------------------------------------------------------------



(defvar _mat_inv_size)

(defvar _mat_inv_pivotel)

(defvar _mat_inv_count)



(defmacro _mat-inv (matrix &key (element-type DEFAULT-MATRIX-TYPE))

  `(progn

     (let ( _mat_inv_auxmat )

       (setf _mat_inv_size

             (_mat-dimension ,matrix 0 :element-type ,element-type))

       

       (setf _mat_inv_auxmat (_mat-id _mat_inv_size :element-type ,element-type))

       

       (dotimes (_mat_inv_count _mat_inv_size)

         (setf _mat_inv_pivotel 

               (_find-pivot-element ,matrix _mat_inv_count (1- _mat_inv_size) 

                                    _mat_inv_count _mat_inv_count

                                    :element-type ,element-type))

         (if (/= (first _mat_inv_pivotel) _mat_inv_count)

           (progn

             (_swap-rows ,matrix _mat_inv_count (first _mat_inv_pivotel) 

                         :element-type ,element-type)

             (_swap-rows _mat_inv_auxmat _mat_inv_count (first _mat_inv_pivotel)

                         :element-type ,element-type)

             )

           )

         (_pivot-with-back ,matrix _mat_inv_count _mat_inv_count

                           _mat_inv_auxmat

                           :element-type ,element-type)

         )

       (setf ,matrix _mat_inv_auxmat)

       )

     )

  )



;-----------------------------------------------------------------------------

; (Destructive) Calculate the value of det(A) , where A is a matrix

;

; INPUT : - matrix

;

; OUTPUT: - determinant of the matrix

; 

; Examples

;

;  (setf a (make-array '(3 3) :initial-contents '((1 2 3)(5 2 0)(1 -1 3))))

;  -> #2A((1 2 3) (5 2 0) (1 -1 3))

;  (_mat-det a )

;  -> -45

;

;-----------------------------------------------------------------------------



(defmacro _mat-det (A &key (element-type DEFAULT-MATRIX-TYPE))

  `(do*

     ((_mat_det_n (_mat-dimension ,A 0 :element-type ,element-type))

      (_mat_det_n-1 (- _mat_det_n 1))

      (_mat_det_zero nil)

      (_mat_det_i (- _mat_det_n 2) (- _mat_det_i 1)))

     ((or _mat_det_zero 

          (< _mat_det_i 0)) 

      (if _mat_det_zero

        0

        (do ((_mat_det_j 0 (+ _mat_det_j 1))

             (_mat_det_res 1 (* _mat_det_res 

                                (_mat-element ,A _mat_det_j 

                                              _mat_det_j :element-type ,element-type))))

            ((= _mat_det_j _mat_det_n) _mat_det_res))))

     (do

       ((_mat_det_h _mat_det_n-1 (- _mat_det_h 1))

        (_mat_det_k 0))

       ((or _mat_det_zero (= _mat_det_h _mat_det_i)))

       (when (zerop (_mat-element ,A _mat_det_h _mat_det_h

                                  :element-type ,element-type))

         (setf _mat_det_k (do

                            ((_mat_det_j 0 (+ _mat_det_j 1)))

                            ((or (> _mat_det_j _mat_det_h) 

                                 (not (zerop (_mat-element ,A _mat_det_h _mat_det_j

                                                           :element-type ,element-type))))

                             _mat_det_j)))

         (if (> _mat_det_k _mat_det_h)

           (setf _mat_det_zero t)

           (dotimes (_mat_det_j _mat_det_n)

             (setf (_mat-element ,A _mat_det_j _mat_det_h :element-type ,element-type) 

                   (+ (_mat-element ,A _mat_det_j _mat_det_h :element-type ,element-type)

                      (_mat-element ,A _mat_det_j _mat_det_k :element-type ,element-type))))))

       (unless _mat_det_zero

         (do

           ((_mat_det_d (/ (_mat-element ,A _mat_det_i _mat_det_h :element-type ,element-type)

                           (_mat-element ,A _mat_det_h _mat_det_h :element-type ,element-type)))

            (_mat_det_j 0 (+ _mat_det_j 1)))

           ((= _mat_det_j _mat_det_n))

           (setf (_mat-element ,A _mat_det_i _mat_det_j :element-type ,element-type)

                 (- (_mat-element ,A _mat_det_i _mat_det_j :element-type ,element-type)

                    (* _mat_det_d 

                       (_mat-element ,A _mat_det_h _mat_det_j :element-type ,element-type)))

                 )))))

  )



;-----------------------------------------------------------------------------

; (Destructive) Calculate the rank of a submatrix in MATRIX.

; MIN-ROW, MAX-ROW, MIN-COL and MAX-COL individuates this submatrix.

; This macro had even a side effect on min-row value.

;

; INPUT : - matrix

;         - min-row index

;         - max-row index

;         - min-col index

;         - max-col index

;

; OUTPUT: - the matrix rank.

;

; EXAMPLES

;    (setf a (make-array '(3 3) :initial-contents '((1 2 3)(5 2 0)(1 -1 3)) ))

;    -> #2A((1 2 3) (5 2 0) (1 -1 3))

;    (_mat-rank a 0 2 0 2)

;    -> 3

;    (setf a '((1 2 3)(1 -1 3)(-2 2 -6)) )

;    -> ((1 2 3)(1 -1 3)(-2 2 -6))

;    (_mat-rank a 1 1 1 1 :element-type _list)

;    -> 1

;------------------------------------------------------------------------------



(defmacro _mat-rank (matrix min-row max-row min-col max-col 

                            &key (element-type DEFAULT-MATRIX-TYPE))

  (declare (special **epsilon))

  `(let ((_mat_rank_n (1+ (min (- ,max-row ,min-row) (- ,max-col ,min-col))))

         (_mat_rank_rank 0)

         (_mat_rank_min-row ,min-row))

     (dotimes (_mat_rank_i _mat_rank_n _mat_rank_rank)

       (let* ((_mat_rank_x-y (_find-pivot-element ,matrix _mat_rank_min-row ,max-row ,min-col

                                                 ,max-col :element-type ,element-type))

              (_mat_rank_row-index (first _mat_rank_x-y))

              (_mat_rank_col-index (second _mat_rank_x-y)))

         (if (< (abs (_mat-element ,matrix _mat_rank_row-index _mat_rank_col-index

                                   :element-type ,element-type)) **epsilon)

           (return _mat_rank_rank))

         (_pivot ,matrix _mat_rank_row-index _mat_rank_col-index :element-type ,element-type)

         (_swap-rows ,matrix _mat_rank_min-row _mat_rank_row-index :element-type ,element-type)

         (incf _mat_rank_min-row)

         (incf _mat_rank_rank)))))





;-----------------------------------------------------------------------------

; (Destructive) Calculate the rank of a submatrix in MATRIX.

; MIN-ROW, MAX-ROW, MIN-COL and MAX-COL individuates this submatrix.

; This macro had even a side effect on min-row value.

;

; INPUT : - matrix

;         - min-row index

;         - max-row index

;         - min-col index

;         - max-col index

;

; OUTPUT: - the matrix rank.

;

; EXAMPLES

;    (setf a (make-array '(3 3) :initial-contents '((1 2 3)(5 2 0)(1 -1 3)) ))

;    -> #2A((1 2 3) (5 2 0) (1 -1 3))

;    (_mat-rank a 0 2 0 2)

;    -> 3

;    (setf a '((1 2 3)(1 -1 3)(-2 2 -6)) )

;    -> ((1 2 3)(1 -1 3)(-2 2 -6))

;    (_mat-rank a 1 1 1 1 :element-type _list)

;    -> 1

;------------------------------------------------------------------------------



(defmacro _mat-transpose (matrix &key (element-type DEFAULT-MATRIX-TYPE))

  `(do* ((_mat_transp_n (_mat-dimension ,matrix 0 :element-type  ,element-type))

         (_mat_transp_m (_mat-dimension ,matrix 1 :element-type  ,element-type))

         (_mat_transp_outmat (_mat-new (list m n) :element-type  ,element-type))

         (_mat_transp_i 0 (+ _mat_transp_i 1)))

    ((= _mat_transp_i _mat_transp_n) (setf ,matrix _mat_transp_outmat)

    (do ((_mat_transp_j 0 (+ _mat_transp_j 1)))

        ((= _mat_transp_j _mat_transp_m))

      (setf (_mat-element _mat_transp_outmat _mat_transp_j _mat_transp_i 

                          :element-type  ,element-type) 

            (_mat-element ,matrix _mat_transp_i _mat_transp_j

                          :element-type  ,element-type)))))

  )







;-----------------------------------------------------------------------------

; (Destructive) Similar to pivot, but executes the same pivot transformation on

; another matrix BACK-MATRIX. This macro is been expressively written for use in

; matrix inversion macro. The BACK-MATRIX must be of the same type of the MATRIX.

;

; The static variables _pivot_wback_coeff and _pivot_wback_element*-1 are used

; for auxiliary storage of values. This macro cannot be recursive !

;

; INPUT : - matrix

;         - row index

;         - column index

;

; OUTPUT: - undefined.

;

;------------------------------------------------------------------------------



(defvar _pivot_wback_element*-1)

(defvar _pivot_wback_coeff)



(defmacro _pivot-with-back (matrix pivot-row

                                   pivot-column

                                   back-matrix

                                   &key (element-type DEFAULT-MATRIX-TYPE))

  `(progn 

     (setf _pivot_wback_element*-1 (- (_mat-element ,matrix 

                                                    ,pivot-row ,pivot-column 

                                                    :element-type  ,element-type)))

     (dotimes (_pivot_wback_cnt (_mat-dimension ,matrix 0 

                                                :element-type  ,element-type))

       (if (/= _pivot_wback_cnt ,pivot-row)

         (progn

           (setf _pivot_wback_coeff (/ (_mat-element ,matrix _pivot_wback_cnt 

                                                     ,pivot-column 

                                                     :element-type ,element-type)

                                       _pivot_wback_element*-1))

           (_add-pivot-row*coeff-to-row ,matrix _pivot_wback_cnt 

                                        ,pivot-row _pivot_wback_coeff 

                                        :element-type  ,element-type)

           (_add-pivot-row*coeff-to-row ,back-matrix _pivot_wback_cnt 

                                        ,pivot-row _pivot_wback_coeff 

                                        :element-type  ,element-type))))

     (setf _pivot_wback_coeff (/ (- _pivot_wback_element*-1)))

     (_multiply-row-for-coeff ,matrix ,pivot-row

                              _pivot_wback_coeff

                              :element-type ,element-type)

     (_multiply-row-for-coeff ,back-matrix ,pivot-row

                              _pivot_wback_coeff

                              :element-type ,element-type)

     ))



;-----------------------------------------------------------------------------

; (Destructive) Multiply MATRIX with the specified scalar value SCAL

;

; INPUT : - matrix

;         - scalar value

;

; OUTPUT: - undefined.

; 

;     (setf b (make-array '(3 4) :initial-element nil))

;     -> #2A((NIL NIL NIL NIL) (NIL NIL NIL NIL) (NIL NIL NIL NIL))

;     (_gen-submat-id b 0 2 2)

;     -> NIL

;     (_mat-print b)

;     ->  

;

;-----------------------------------------------------------------------------



(defmacro _mat-scalar-product (matrix scal

                                      &key (element-type DEFAULT-MATRIX-TYPE))

  `(let

     ((_mat_sprod_nrows (_mat-dimension ,matrix 0 :element-type ,element-type))

      (_mat_sprod_ncols (_mat-dimension ,matrix 1 :element-type ,element-type))

      )

     (dotimes (_mat_sprod_row _mat_sprod_nrows)

       (dotimes ( _mat_sprod_col _mat_sprod_ncols)

         (setf (_mat-element ,matrix _mat_sprod_row _mat_sprod_col

                             :element-type ,element-type)

               (* (_mat-element ,matrix _mat_sprod_row _mat_sprod_col

                                :element-type ,element-type)

                  ,scal)))))

  )



;-----------------------------------------------------------------------------

; Multiply matrix A for the matrix B, producing a new matrix of the same

; type of matrix A.

;

; INPUT : - matrix

;         - matrix

;

; OUTPUT: - matrix

; 

;   (setf a '((1 2 3)(5 2 0)(1 -1 3)))

;   -> ((1 2 3) (5 2 0) (1 -1 3))

;   (setf b '((0 1 1 0)(2 1 0 1)(3 0 1 0)))

;   -> ((0 1 1 0) (2 1 0 1) (3 0 1 0))

;   (_mat-product a b :element-type1 _list :element-type2 _list)

;   -> ((13.0 3.0 4.0 2.0) (4.0 7.0 5.0 2.0) (7.0 0.0 4.0 -1.0))

;

;-----------------------------------------------------------------------------



(defmacro _mat-product (A B &key (element-type1 DEFAULT-MATRIX-TYPE) 

                          (element-type2 DEFAULT-MATRIX-TYPE) )

  `(let* ((_mat_product_n (_mat-dimension ,A 0 :element-type ,element-type1))

          (_mat_product_p (_mat-dimension ,A 1 :element-type ,element-type1))

          (_mat_product_m (_mat-dimension ,B 1 :element-type ,element-type2))

          (_mat_product_C (_mat-new _mat_product_n _mat_product_m

                         :initial-element 0.0d0

                         :element-type ,element-type1)))

     (dotimes (_mat_product_i _mat_product_n _mat_product_C)

       (dotimes (_mat_product_j _mat_product_m)

         (dotimes (_mat_product_k _mat_product_p) 

           (setf (_mat-element _mat_product_C _mat_product_i _mat_product_j

                               :element-type ,element-type1) 

                (+ (_mat-element _mat_product_C _mat_product_i _mat_product_j

                                 :element-type ,element-type1)

                   (* (_mat-element ,A _mat_product_i _mat_product_k

                                    :element-type ,element-type1)

                      (_mat-element ,B _mat_product_k _mat_product_j

                                    :element-type ,element-type2)))))))))





;-----------------------------------------------------------------------------

; Adds matrix A with the matrix B, producing a new matrix of the same

; type of matrix A.

;

; INPUT : - matrix

;         - matrix

;

; OUTPUT: - matrix

; 

;   (setf a '((1 2 3)(5 2 0)(1 -1 3)))

;   -> ((1 2 3) (5 2 0) (1 -1 3))

;   (setf b '((0 1 1)(2 1 0)(3 0 1)))

;   -> ((0 1 1) (2 1 0) (3 0 1))

;   (_mat-sum a b :element-type1 _list :element-type2 _list)

;   -> ((1 3 4) (7 3 0) (4 -1 4))

;

;-----------------------------------------------------------------------------



(defmacro _mat-sum (A B &key (element-type1 DEFAULT-MATRIX-TYPE) 

                      (element-type2 DEFAULT-MATRIX-TYPE) )

  `(let* ((_mat_sum_n (_mat-dimension ,A 0 :element-type ,element-type1))

          (_mat_sum_m (_mat-dimension ,A 1 :element-type ,element-type1))

          (_mat_sum_C (_mat-new _mat_sum_n _mat_sum_m

                                 :element-type ,element-type1)))

     (dotimes (_mat_sum_i _mat_sum_n _mat_sum_C)

       (dotimes (_mat_sum_j _mat_sum_m)

         (setf (_mat-element _mat_sum_C _mat_sum_i

                             _mat_sum_j :element-type ,element-type1)

               (+ (_mat-element ,A _mat_sum_i _mat_sum_j

                                :element-type ,element-type1) 

                  (_mat-element ,B _mat_sum_i _mat_sum_j

                                :element-type ,element-type2)))))))



;-----------------------------------------------------------------------------

; Subtracts matrix A with the matrix B, producing a new matrix of the same

; type of matrix A.

;

; INPUT : - matrix

;         - matrix

;

; OUTPUT: - matrix

; 

;   (setf a '((1 2 3)(5 2 0)(1 -1 3)))

;   -> ((1 2 3) (5 2 0) (1 -1 3))

;   (setf b '((0 1 1)(2 1 0)(3 0 1)))

;   -> ((0 1 1) (2 1 0) (3 0 1))

;   (_mat-diff a b :element-type1 _list :element-type2 _list)

;   -> ((1 1 2) (3 1 0) (-2 -1 2))

;

;-----------------------------------------------------------------------------



(defmacro _mat-diff (A B &key (element-type1 DEFAULT-MATRIX-TYPE) 

                      (element-type2 DEFAULT-MATRIX-TYPE) )

  `(let* ((_mat_sum_n (_mat-dimension ,A 0 :element-type ,element-type1))

          (_mat_sum_m (_mat-dimension ,A 1 :element-type ,element-type1))

          (_mat_sum_C (_mat-new _mat_sum_n _mat_sum_m

                                 :element-type ,element-type1)))

     (dotimes (_mat_sum_i _mat_sum_n _mat_sum_C)

       (dotimes (_mat_sum_j _mat_sum_m)

         (setf (_mat-element _mat_sum_C _mat_sum_i

                             _mat_sum_j :element-type ,element-type1)

               (- (_mat-element ,A _mat_sum_i _mat_sum_j

                                :element-type ,element-type1) 

                  (_mat-element ,B _mat_sum_i _mat_sum_j

                                :element-type ,element-type2)))))))





;-----------------------------------------------------------------------------

; Formatted print the matrix MATRIX.

; INPUT : - matrix

;

; OUTPUT: - undefined

;

; Example: 

;          (_mat-print (make-array '(3 3) :initial-contents '((1 2 3)(4 5 6)(7 8 9))))

;            ->      1.E+0       2.E+0       3.E+0 

;                    4.E+0       5.E+0       6.E+0 

;                    7.E+0       8.E+0       9.E+0 

;          (_mat-print '((1 2 3)(4 5 6)) :element-type _list)

;            ->      1.E+0       2.E+0       3.E+0 

;                    4.E+0       5.E+0       6.E+0 

;------------------------------------------------------------------------------

(defmacro _mat-print (matrix  &key (element-type DEFAULT-MATRIX-TYPE))

  `(dotimes (i (_mat-dimension ,matrix 0 :element-type ,element-type) (format t "~%"))

     (format t "~%")

     (dotimes (j (_mat-dimension ,matrix 1 :element-type ,element-type))

       (format t " ~10E "  (_mat-element ,matrix i j :element-type ,element-type)))))



;------------------------------------------------------------------------------

; Tests if the plane defined by NEW-PARAMETERS is coincident with PLANE.

; If the test is successful returns the plane-identifier.

; INPUT : -

;

; OUTPUT: - plane-identifier or nil.

;

; Example: 

;              (_same-plane #(1 2 3) '(P1 #(1 2 3)))    -> P1

;              (_same-plane #(1 2 4) '(P1 #(1 2 3)))    -> NIL

;              (_same-plane '(1 2 3) '(P1 (1 2 3)))     -> P1

;              (_same-plane '(1 2 4) '(P1 (1 2 3)))     -> NIL

;------------------------------------------------------------------------------

(defun _same-plane (new-parameters plane)

  (if (every '= new-parameters (get-coord plane))

    (get-id plane)))



;------------------------------------------------------------------------------

; Searc in the submatrix defined by the indices MIN-ROW MAX-ROW MIN-COL MAX-COL

; for the best pivot element.

; The best pivot element is the element of the submatrix of MATRIX of higher

; absolute value.                   

; INPUT : - matrix

;         - min-row-index

;         - max-row-index

;         - min-col-index

;         - max-col-index

;

; OUTPUT: - list of the two indices of the best pivot element

;

; Example: 

;(_find-pivot-element (make-array '(3 3) :initial-contents '((1 2 3)(4 5 6)(7 8 9)))

;                                 0 2 0 2)

;                                                                      -> (2 2)

;(_find-pivot-element '((1 2 3)(4 5 6)(7 8 9))

;                     0 2 0 2 :element-type _list)

;                                                                      -> (2 2)

;(_find-pivot-element (make-array '(3 3) :initial-contents '((1 2 3)(4 5 6)(7 8 9)))

;                                 0 2 0 1)

;                                                                      -> (2 1)

;(_find-pivot-element '((1 2 3)(4 5 6)(7 8 9))

;                                 0 2 0 1 :element-type _list)

;                                                                      -> (2 1)

;------------------------------------------------------------------------------

(defmacro _find-pivot-element (matrix min-row

                                      max-row

                                      min-col

                                      max-col

                                      &key (element-type DEFAULT-MATRIX-TYPE))

  `(let ((delta-col (- ,max-col ,min-col -1))

         (row-index ,min-row)

         (col-index ,min-col)

         (pivot-row-index ,min-row)

         (pivot-col-index ,min-col)

         current-pivot-element-value

           i j)

     (setf current-pivot-element-value 

              (abs (_mat-element ,matrix pivot-row-index pivot-col-index

                               :element-type ,element-type)))                                  

     (dotimes (i (- ,max-row ,min-row -1)

                 (list pivot-row-index pivot-col-index)) ;result value

       (setf col-index ,min-col)

       (dotimes (j delta-col)

         (let ((current-element-value 

                (abs (_mat-element ,matrix row-index col-index

                                   :element-type ,element-type))))

           (if (< current-pivot-element-value

                  (abs current-element-value))

             (setf pivot-row-index row-index

                   pivot-col-index col-index)))

         (incf col-index))

       (incf row-index))))



;------------------------------------------------------------------------------

; INPUT : -

;

; OUTPUT: -

;

; Example: 

;------------------------------------------------------------------------------



(defmacro _ (matrix I1 i2 &key (element-type DEFAULT-MATRIX-TYPE))

  (declare (special _array _list matrix i1 i2))

  (cond

   ((symbol-equal element-type _array) )

   ((symbol-equal element-type _list) )))





(defun print-example( example )

  (format t   ";  ")

  (princ example)

  (format t "~%;  -> ")

  (princ (eval example)))



;

;(print-example '(_find-pivot-element '((1 2 3)(4 5 6)(7 8 9))

;                 0 2 0 1 :element-type _list))

;

